/*
 * DisTagger
 *Author: Matthieu Constant
 *
 * Copyright (C) 2010 Université Paris-Est, LIGM and CNRS
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.distagger;

import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.List;

public class AnneConfiguration {
       private File config;
       private List<ConfigListener> listeners = new ArrayList<ConfigListener>();
       private List<File> texts = new ArrayList<File>();
       private File euh;
       private File meta;
       private File classes;
       private File inserts;
       private File outDir;
       private boolean toBeSaved = false;
       private static final String EUH = "EUH";
       private static final String CLASSES = "CLASSES";
       private static final String INSERTS = "INSERTS";
       private static final String OUT = "OUT";
       private static final String TEXTS = "TEXTS";
       private static final String META = "META";
       
       public AnneConfiguration(){
    	   fireConfigCleaned();
       }
       
       
       public void load(String configFilename) throws IOException,AnneException{
    	        this.config = new File(configFilename);    	        
                loadConfigurationFile(configFilename);
                fireConfigLoaded();
       }

       private void loadConfigurationFile(String filename) throws IOException,AnneException{
    	   BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(filename),"UTF-8"));
    		  String s;
    		  while((s = reader.readLine()) != null){
    			   String[] tab = s.split(";");
    			   if(tab.length > 0){
    				   String name = tab[0];
    				   if(name.equals("TEXTS")){
    					   for(int i = 1;i<tab.length;i++){
    						   this.add(new File(tab[i]));
    					   }    					   
    				   }
    				   else{
    					   if(tab.length != 2){
    						   throw new AnneException("Configuration in "+ filename+ " is not in good format.");
    					   }
    					   File val = new File(tab[1]);
    					   if(name.equals(EUH)){
    						    this.setEuh(val);
    					   }
    					   if(name.equals(CLASSES)){
    						   this.setClasses(val);
    					   }
    					   if(name.equals(OUT)){
    						   this.setOutDir(val);
    					   }
    					   if(name.equals(INSERTS)){
    						   this.setInserts(val);
    					   }
    					   if(name.equals(META)){
    						   this.setMeta(val);
    					   }
    				   }
    			   }
    		  }
    		  reader.close();
       }
       
       public void addConfigListener(ConfigListener l){
    	   listeners.add(l);
       }
       
       
       protected void fireConfigChanged(){
    	   toBeSaved = true;
    	   for(ConfigListener l:listeners){
    		   l.configChanged();
    	   }
       }
       
       protected void fireConfigCleaned(){
    	   toBeSaved = false;
    	   for(ConfigListener l:listeners){
    		   l.configCleaned();
    	   }
       }
       
       protected void fireConfigLoaded(){
    	   toBeSaved = false;
    	   for(ConfigListener l:listeners){
    		   l.configLoaded();
    	   }
       }
       
       protected void fireConfigSaved(){
    	   toBeSaved = false;
    	   for(ConfigListener l:listeners){
    		   l.configSaved();
    	   }
       }
             
    public boolean isToBeSaved(){
    	return toBeSaved;
    }
       
	public File getClasses() {
		return classes;
	}

	public void setClasses(File classes) {
		this.classes = classes;
		fireConfigChanged();
	}

	public File getConfig() {
		return config;
	}

	public void setConfig(File config) {
		this.config = config;
	}

	public File getEuh() {
		return euh;
	}

	public void setEuh(File euh) {
		this.euh = euh;		
		fireConfigChanged();
	}

	public File getInserts() {
		return inserts;
		
	}

	public void setInserts(File inserts) {
		this.inserts = inserts;
		fireConfigChanged();
	}

	public File getMeta() {
		return meta;
		
	}

	public void setMeta(File meta) {
		this.meta = meta;
		fireConfigChanged();
	}
	
	
	public List<File> getTexts() {
		return texts;
	}


	public File getOutDir() {
		return outDir;
	}


	public void setOutDir(File outDir) {
		this.outDir = outDir;
		fireConfigChanged();
	}


	public void add(File e) {
		texts.add(e);
		fireConfigChanged();
	}

	
	public void remove(int index) {
		texts.remove(index);
		fireConfigChanged();
	}
     
	public void clear(){
		for(int i = texts.size() - 1 ; i >= 0; i--){
			texts.remove(i);
		}		
		fireConfigCleaned();
	}
	
	public void save(String filename) throws IOException{
			  write(filename);
			  setConfig(new File(filename));
			  fireConfigSaved();						
	}
	
  public void write(String filename) throws FileNotFoundException,IOException{
		write(new BufferedOutputStream(new FileOutputStream(filename)));
		fireConfigSaved();
 }

  public void write() throws FileNotFoundException,IOException{
		write(new BufferedOutputStream(System.out));
}
  
	private void write(OutputStream out) throws IOException{	  
		BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(out,"UTF-8"));
	      if(getEuh() != null ) writer.append(EUH+";"+getEuh()+"\n");
	      if(getClasses() != null ) writer.append(CLASSES+";"+getClasses()+"\n");
	      if(getOutDir() != null ) writer.append(OUT+";"+getOutDir()+"\n");
	      if(getInserts() != null ) writer.append(INSERTS+";"+getInserts()+"\n");
	      if(getMeta() != null ) writer.append(META+";"+getMeta()+"\n");
	      writer.append(TEXTS);
	      for(File f:getTexts()){
	    	  writer.append(";"+f);
	      }
	      writer.append("\n");
	      writer.close();
	}
	
	
       
}
