/**
 * 
 *
 * Software: LGTools
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.format;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import jxl.Workbook;
import jxl.WorkbookSettings;
import jxl.write.Label;
import jxl.write.WritableSheet;
import jxl.write.WritableWorkbook;
import jxl.write.WriteException;
import jxl.write.biff.RowsExceededException;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;

import fr.upemlv.lgtools.tagging.lex.LexicalEntry;
import fr.upemlv.lgtools.tagging.lex.LexicalEntryParser;
import fr.upemlv.lgtools.tagging.lex.ListDico;
import fr.upemlv.lgtools.tagging.lex.Tag;
import fr.upemlv.lgtools.tagging.lex.Tagset;
import fr.upemlv.lgtools.tagging.nlp.Tokenizer;



/**
 * Reads a dela dictionary and transform it into an Excel file
 * 
 * 
 * @author Matthieu Constant
 *
 */

public class Dela2Excel {

	final private static String DELA_INPUT = "dela";
	final private static String EXCEL_OUTPUT = "xls";
	final private static String TAGSET = "tagset";
	final private static String TAG_FILTER = "tag";
	final private static String DELA_ENCODING = "enc"; 
	final private static String LANGUAGE = "lang";
	final private static String DEFAULT_ENCODING = "UTF16-LE"; 
	final private static int MAX = 30000;

	
	
	private static JSAP createJSAP() throws JSAPException{
		JSAP jsap = new JSAP();
		FlaggedOption opt = new FlaggedOption(DELA_INPUT)
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(true)
        .setShortFlag(JSAP.NO_SHORTFLAG)
        .setLongFlag(DELA_INPUT)        
        ;
		opt.setHelp("Input dictionary in DELA format (.dic)");
		jsap.registerParameter(opt);
	    
		opt = new FlaggedOption(EXCEL_OUTPUT)
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(true)
        .setShortFlag(JSAP.NO_SHORTFLAG)
        .setLongFlag(EXCEL_OUTPUT)        
        ;
		opt.setHelp("Output Excel file (.xls)");
		jsap.registerParameter(opt);
		
		opt = new FlaggedOption(LANGUAGE)
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(true)
        .setShortFlag(JSAP.NO_SHORTFLAG)
        .setLongFlag(LANGUAGE)        
        ;
		opt.setHelp("Language of dictionary");
		jsap.registerParameter(opt);
		
		opt = new FlaggedOption(TAGSET)
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(true)
        .setShortFlag(JSAP.NO_SHORTFLAG)
        .setLongFlag(TAGSET)        
        ;
		opt.setHelp("Tagset configuration file");
		jsap.registerParameter(opt);
		
		opt = new FlaggedOption(DELA_ENCODING)
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault(DEFAULT_ENCODING)
        .setShortFlag(JSAP.NO_SHORTFLAG)
		.setLongFlag(DELA_ENCODING);
		opt.setHelp("dela encoding");
		jsap.registerParameter(opt);
		
		opt = new FlaggedOption(TAG_FILTER)
        .setStringParser(JSAP.STRING_PARSER)
        .setRequired(false)
        .setDefault((String)null)
        .setShortFlag(JSAP.NO_SHORTFLAG)
		.setLongFlag(TAG_FILTER);
		opt.setHelp("tag used to filter entries");
		jsap.registerParameter(opt);
		
		return jsap;
	}
	

	
	
	private static Dela2ExcelConfig parseArguments(String[] args) throws JSAPException{
		JSAP jsap = createJSAP();
		JSAPResult config = jsap.parse(args);
		if(!config.success()){
			System.err.print("Usage: ");
			System.err.print("java "+Dela2Excel.class.getCanonicalName()+" ");
            System.err.println(jsap.getUsage());
            System.err.print(jsap.getHelp());
            System.exit(1);
		}
		String encoding = config.getString(DELA_ENCODING);
		String dela = config.getString(DELA_INPUT);
		String xls = config.getString(EXCEL_OUTPUT);
		String tag = config.getString(TAG_FILTER);
		String tagset = config.getString(TAGSET);
		String lang = config.getString(LANGUAGE);
		return new Dela2ExcelConfig(dela, xls, encoding,tag,tagset,lang);
	}
	
	private static WritableSheet createNewSheet(WritableWorkbook wb,String name,Tagset tagset,int maxWordLength) throws RowsExceededException, WriteException{
		wb.createSheet(name, wb.getSheets().length);
		WritableSheet sheet = wb.getSheet(wb.getSheets().length - 1);
		int j = maxWordLength + 2;
		for(int i = 0 ; i < maxWordLength ; i++){
			sheet.addCell(new Label(i, 0, "FORM-"+i));	
		}
		
		sheet.addCell(new Label(maxWordLength, 0, "LEMMA"));
		sheet.addCell(new Label(maxWordLength + 1, 0, "POS"));
		for(String type:tagset.getTypes()){
			if(!type.equals("POS")){
				sheet.addCell(new Label(j, 0, type));
				j++;
			}
		}
		return sheet;
	}
	
	private static HashMap<String,List<LexicalEntry>> populateMap(Dela2ExcelConfig config,Tagset tagset) throws IOException{
		HashMap<String,List<LexicalEntry>> map = new HashMap<String, List<LexicalEntry>>();
		Tag tag = new Tag(tagset);
		   if(config.getTag() != null){
			   tag.addCat(config.getTag());
		   }		   
		   Reader reader = new InputStreamReader(new FileInputStream(config.getDelaFilename()),config.getEncoding());
		   ListDico dico = new ListDico(reader, tagset, LexicalEntryParser.UNITEX_LEXICAL_ENTRY_PARSER, config.getLang(),false);
		   for(LexicalEntry entry:dico){
				if(entry.getTag().contains(tag)){
					String cat = "??";
					for(String struct:entry.getTag().getCategories("STRUCT")){
						cat = struct;
						break;
					}
					List<LexicalEntry> entries = map.get(cat);
					if(entries == null){
						entries = new LinkedList<LexicalEntry>();
						map.put(cat, entries);
					}
					entries.add(entry);
				}
			}		
		return map;
	}
	
	private static String getValue(Collection<String> vals){
		String res = "";
		boolean first = true;
		for(String v:vals){
			if(first){
				first = false;
			}
			else{
				res += "+";
			}
			res += v;
		}
		return res;
	}
	
	private static void writeWord(WritableSheet sheet,int row,String word,Tokenizer tokenizer,int maxLength) throws RowsExceededException, WriteException{
		int col = 0;
		for(String tok:tokenizer.perform(word)){
			sheet.addCell(new Label(col, row, tok));
			col++;
		}
	}
	
	public static void saveExcelFile(Tagset tagset,HashMap<String,List<LexicalEntry>> map,Dela2ExcelConfig config, HashMap<String, Integer> lengths) throws IOException, RowsExceededException, WriteException{		
		WorkbookSettings ws = new WorkbookSettings();
		WritableWorkbook wb = Workbook.createWorkbook(new File(config.getExcelFilename()),ws);
		Tokenizer tokenizer = new Tokenizer(config.getLang());
		for(String struct:map.keySet()){
			int sheetCount = 0;
			int l = lengths.get(struct);
			WritableSheet sheet = createNewSheet(wb,struct,tagset,l);

			int i = 1;
			int j;
			List<LexicalEntry> entries = map.get(struct);
			Collections.sort(entries);
			for(LexicalEntry entry:entries){
				if(i % MAX == 0){
					sheetCount++;
					sheet = createNewSheet(wb,struct+"-"+sheetCount,tagset,l);
					i = 1;
				}				
				
				j = l+2;
				writeWord(sheet,i,entry.getWord(),tokenizer,l);
				//sheet.addCell(new Label(0, i, entry.getWord()));
				sheet.addCell(new Label(l, i, entry.getLemma()));
				sheet.addCell(new Label(l+1, i, entry.getTag().getCategories("POS").toString()));
				for(String type:tagset.getTypes()){
					if(!type.equals("POS")){
						sheet.addCell(new Label(j, i, getValue(entry.getTag().getCategories(type))));
						j++;
					}
				}
				i++;
				//System.err.println(i);
			}
		}
		wb.write();
		wb.close();
	}
	
	
	private static HashMap<String,Integer> computeMaxLengths(HashMap<String,List<LexicalEntry>> map,String lang){
		HashMap<String,Integer> lengths = new HashMap<String, Integer>();
		Tokenizer tokenizer = new Tokenizer(lang);
		for(String name:map.keySet()){
			int maxSize  = 0;
			for(LexicalEntry entry:map.get(name)){
				int size = tokenizer.perform(entry.getWord()).size();
				if(size > maxSize){
					maxSize = size;
				}
			}
			lengths.put(name, maxSize);
		}
		return lengths;
	}
	
	
	public static void main(String[] args) throws JSAPException, IOException, RowsExceededException, WriteException{
	   Dela2ExcelConfig config = parseArguments(args);
	   Tagset tagset = new Tagset(config.getTagset());
	   System.err.println("Populating...");
	   HashMap<String,List<LexicalEntry>> map = populateMap(config,tagset);
	   System.err.println("Computing maximum lengths...");
	   HashMap<String,Integer> lengths = computeMaxLengths(map,config.getLang());
	   System.err.println("Saving in file...");
	   saveExcelFile(tagset, map, config,lengths);
	   System.err.println("Done.");
	}
	
}
