/**
 * 
 *
 * Software: LGTools
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.parsing;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Switch;

import fr.upemlv.lgtools.parsing.mwe.CrfMweRecognizer;
import fr.upemlv.lgtools.parsing.mwe.FilteredGoldMweRecognizer;
import fr.upemlv.lgtools.parsing.mwe.MweRecognizer;
import fr.upemlv.lgtools.parsing.mwe.SimpleMweRecognizer;
import fr.upemlv.lgtools.tagging.LGTaggerMain;
import fr.upemlv.lgtools.tagging.lex.Tagset;
import fr.upemlv.lgtools.text.RawText;

/**
 * Program that parses a text (by including MWE information)
 */

public class LGParserMain {
	private static String OUTPUT = "output";
	
	private static String MWE_ANNOTATION = "mwe_annotation";
	private static String GOLD_MWE_RECOGNITION = "gold_recognition";
	private static String GOLD_SUBTREES = "gold";
	private static String FORCE_PRETERMINALS = "force";
   private static String TEST_CORPUS = "train";
   private static String CONFIG = "config";
   private static String PARSER_TYPE = "parser";
   private static String SIMPLE_PARSER = "simple";
   private static String CRF_PARSER = "crf";
   private static String BASELINE_PARSER = "baseline";
   
   
	
	private static  JSAP createJSAP() throws JSAPException{
		JSAP jsap = new JSAP();

		FlaggedOption opt;
		
		opt = new FlaggedOption(TEST_CORPUS)
		.setStringParser(JSAP.STRING_PARSER)
		.setRequired(true)
		.setShortFlag('t')
		.setRequired(true)
		.setLongFlag(TEST_CORPUS);
		opt.setHelp("Path of the test corpus (in CONLL format)");
		jsap.registerParameter(opt);

		opt = new FlaggedOption(OUTPUT)
		.setStringParser(JSAP.STRING_PARSER)
		.setShortFlag('o')
		.setRequired(true)
		.setLongFlag(OUTPUT);
		opt.setHelp("Output of the parsing");
		jsap.registerParameter(opt);
			
		
			opt = new FlaggedOption(CONFIG)
			.setStringParser(JSAP.STRING_PARSER)		
			.setShortFlag('c')
			.setLongFlag(CONFIG)
			.setRequired(true)
			;
			opt.setHelp("Path of configuration file");
			jsap.registerParameter(opt);

			opt = new FlaggedOption(PARSER_TYPE)
			.setStringParser(JSAP.STRING_PARSER)        
			.setShortFlag('p')
			.setLongFlag(PARSER_TYPE);
			opt.setHelp("Parser type: 'baseline', 'simple' or 'crf'");
			opt.setRequired(false);
			opt.setDefault("simple");
			jsap.registerParameter(opt);
		
			Switch sw = new Switch(GOLD_SUBTREES)
			.setShortFlag('s')
			.setLongFlag(GOLD_SUBTREES);
			sw.setHelp("Retrieves gold mwe subtrees in crf mode");
			jsap.registerParameter(sw);
			
			sw = new Switch(FORCE_PRETERMINALS)
			.setShortFlag('f')
			.setLongFlag(FORCE_PRETERMINALS);
			sw.setHelp("force mwe preterminals in crf mode");
			jsap.registerParameter(sw);
			
			sw = new Switch(GOLD_MWE_RECOGNITION)
			.setShortFlag('m')
			.setLongFlag(GOLD_MWE_RECOGNITION);
			sw.setHelp("use gold mwe recognition");
			jsap.registerParameter(sw);
			
			sw = new Switch(MWE_ANNOTATION)
			.setShortFlag('a')
			.setLongFlag(MWE_ANNOTATION);
			sw.setHelp("also print parses with MWE annotations");
			jsap.registerParameter(sw);
			
			return jsap;
	}
	
	private static void usage(JSAP jsap){
		System.err.print("Usage: ");
		System.err.print("java "+LGTaggerMain.class.getCanonicalName()+" ");
		System.err.println(jsap.getUsage());
		System.err.print(jsap.getHelp());
		System.exit(1);
	}
	
	
	public static void main(String[] argv) throws Exception{
		JSAP jsap = createJSAP();
		JSAPResult arguments = jsap.parse(argv);
		if(!arguments.success()){
			usage(jsap);
		}
		String testCorpusPath = arguments.getString(TEST_CORPUS);		
		String configPath = arguments.getString(CONFIG);
		String parserType = arguments.getString(PARSER_TYPE);
		String outputPath = arguments.getString(OUTPUT);
		boolean withMweAnnotation = arguments.getBoolean(MWE_ANNOTATION);
		
		Treebank testTreebank = new Treebank(testCorpusPath);				
		ParsingConfig config = new ParsingConfig(configPath);
		
		MatParser parser = null;
		Tagset tagset = config.getLgtaggerConfig().getTagset();
		RawText t = testTreebank.toRawText(tagset);		
		
		if(parserType.equals(BASELINE_PARSER)){
			  parser = new BaselineParser(config);
			  parser.parse(t, null).save(outputPath);;
			  return;
		}
		
		if(parserType.equals(SIMPLE_PARSER)){
				parser = new SimpleMweMatParser(config,withMweAnnotation);
				parser.parse(t, null).save(outputPath);
				if(parser.getIntermediateParses() != null){
				  parser.getIntermediateParses().save(outputPath+".mwe");
				}
				return;
		}
		
			if(parserType.equals(CRF_PARSER)){
				SimpleMweRecognizer smwer = new SimpleMweRecognizer(config);				
				MweRecognizer mwer;
				if(!arguments.getBoolean(GOLD_MWE_RECOGNITION)){
				   mwer = new CrfMweRecognizer(config, smwer);
				}
				else{
					mwer = new FilteredGoldMweRecognizer(tagset, smwer);
				}
				
				parser = new MweMatParser(config, withMweAnnotation,arguments.getBoolean(FORCE_PRETERMINALS), arguments.getBoolean(GOLD_SUBTREES));
				TextItems items = mwer.apply(testTreebank);
				Treebank res = parser.parse(t,items);
				res.save(outputPath);
				if(parser.getIntermediateParses() != null){
					  parser.getIntermediateParses().save(outputPath+".mwe");
				}
				return;
			}
			 usage(jsap);		
		
	}	
}
