/**
 * 
 *
 * Software: LGTools
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.parsing;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import edu.berkeley.nlp.ling.Tree;
import edu.berkeley.nlp.ling.Trees;
import fr.upemlv.lgtools.parsing.mwe.AutomaticMweRecognizer;
import fr.upemlv.lgtools.parsing.mwe.CrfMweRecognizer;
import fr.upemlv.lgtools.parsing.mwe.FilteredGoldMweRecognizer;
import fr.upemlv.lgtools.parsing.mwe.GoldMweSubtreeGetter;
import fr.upemlv.lgtools.parsing.mwe.MweRecognizer;
import fr.upemlv.lgtools.parsing.mwe.MweSubtreeGetter;
import fr.upemlv.lgtools.parsing.mwe.SimpleMweRecognizer;
import fr.upemlv.lgtools.parsing.mwe.SubtreeGetter;
import fr.upemlv.lgtools.text.Sentence;
import fr.upemlv.lgtools.utils.FileUtilities;

/**
 * Parser integrating MW recognition
 * 
 * 
 * @author Matthieu Constant
 *
 */

public class MweMatParser extends MatParser{
	private final static String MODIFIED_FTB_DIR = "mftb";
	private final static String MODIFIED_FTB = MODIFIED_FTB_DIR+File.separator+"mftb";	
	private final static String MWE_SUBTREES_DIR = "mwes";
	private final static String MWE_SUBTREES = MWE_SUBTREES_DIR+File.separator+"mwes";
	private final static String GENERAL_GRAMMAR = "grammar";
	private final static String MWE_GRAMMAR = "mwes";
	
	private BKYParser gParser = null;
	private BKYParser mweParser = null;
	private SubtreeGetter getter;
	private final boolean goldSubtrees;
	private final boolean forcePreterminal;
	
	
	public MweMatParser(ParsingConfig config,boolean withMweAnnotation,boolean forcePreterminal,boolean goldSubtrees) throws UnsupportedEncodingException, FileNotFoundException {
		this(config,withMweAnnotation,forcePreterminal,goldSubtrees,-1);
	}
	
	
	public MweMatParser(ParsingConfig config,boolean forcePreterminal,boolean goldSubtrees) throws UnsupportedEncodingException, FileNotFoundException {
		this(config,false,forcePreterminal,goldSubtrees,-1);
	}
	
	public MweMatParser(ParsingConfig config,int iterLearn) throws UnsupportedEncodingException, FileNotFoundException {
		this(config,false,false,false,iterLearn);
	}
	
	public MweMatParser(ParsingConfig config,boolean withMweAnnotation,boolean forcePreterminal,boolean goldSubtrees,int iterLearn) throws UnsupportedEncodingException, FileNotFoundException {
		super(config,iterLearn,withMweAnnotation);		
		this.forcePreterminal = goldSubtrees?true:forcePreterminal;
		this.goldSubtrees = goldSubtrees;
		loadParsers();
	}

	private void loadParsers() throws UnsupportedEncodingException, FileNotFoundException{
		String grammar =getConfig().getOutputDir()+File.separator+GENERAL_GRAMMAR+".gr"; 
		if(new File(grammar).exists()){
		  gParser = new BKYParser(grammar);
		}
		grammar = getConfig().getOutputDir()+File.separator+MWE_GRAMMAR+".gr";
		if(new File(grammar).exists()){
			  mweParser = new BKYParser(grammar);
			  if(!goldSubtrees){
			    getter = new MweSubtreeGetter(mweParser,gParser);
			  }
			  else{
				  String dirPath = getConfig().getOutputDir()+File.separator+MWE_SUBTREES_DIR;
				  getter = new GoldMweSubtreeGetter(dirPath);
			  }
		}	
	}
	
	
	private void modifyRawText(Sentence<String> sentence,PreprocessedItem item){		
		for(int i = item.getEndPosition() ; i >= item.getStartPosition() ; i--){
			sentence.getSequence().remove(i);
		}
		sentence.getSequence().add(item.getStartPosition(),item.getTerminal());
	}
	
	public Sentence<String> preprocess(Sentence<String> sentence,SentenceItems items){	
		if(items == null){
			return sentence;
		}
		items.sortReverse();
		for(PreprocessedItem item:items){
			modifyRawText(sentence, item);
		}
		return sentence;
	}
	
	private void addItemInParseTree(Tree<String> parent,int positionInChildren,PreprocessedItem item){
		Tree<String> newNode = new Tree<String>(item.getPreterminal());
		List<Tree<String>> c = new LinkedList<Tree<String>>();
		c.add(new Tree<String>(item.getTerminal()));
		newNode.setChildren(c);
		parent.getChildren().add(positionInChildren,newNode);
	}
	

	@SuppressWarnings("unchecked")
	private ParseTree modifyParseTree(ParseTree tree, PreprocessedItem item){
		List<Tree<String>> nodes;
		int start = item.getStartPosition();
		int end = item.getEndPosition();
		//System.out.println(tree);

		if(start == end){
			Tree<String> terminal = tree.getTerminals().get(start);
			Tree<String> parent = tree.getParent(terminal);
			nodes = new LinkedList<Tree<String>>();
			nodes.add(terminal);
			parent.setChildren(new LinkedList<Tree<String>>());
			addItemInParseTree(parent,0, item);
		}
		else{
			nodes = tree.getAncestorsInFirstCommonAncestorChildren(start, end);
			Tree<String> parent = tree.getFirstCommonAncestor(start, end); 
			int first = parent.getChildren().indexOf(nodes.get(0));
			parent.getChildren().removeAll(nodes);
			addItemInParseTree(parent, first, item);
		}
		
		Tree<String> mweRoot = new Tree<String>(Trees.PennTreeReader.ROOT_LABEL);
		Tree<String> preterminal = new Tree<String>(item.getPreterminal());
		mweRoot.setChildren(Arrays.asList(preterminal));
		preterminal.setChildren(nodes);
		return new ParseTree(mweRoot);
	}
	
	
	public List<ParseTree> preprocess(ParseTree tree,SentenceItems items){
		List<ParseTree> trees = new LinkedList<ParseTree>();
		for(PreprocessedItem item:items){
			trees.add(modifyParseTree(tree, item));
		}		
		return trees;
	}
	
	private void writeMweParseTrees(List<ParseTree> trees,Writer writer) throws IOException{
		for(ParseTree tree:trees){
			 Treebank.writeParseTree(tree, writer);
		}
	}
	
	private void preprocess(Treebank treebank,String section) throws Exception{
		AutomaticMweRecognizer simpleMweRecognizer = new SimpleMweRecognizer(getConfig());
		MweRecognizer goldRecognizer = new FilteredGoldMweRecognizer(getConfig().getLgtaggerConfig().getTagset(), simpleMweRecognizer);
		TextItems items = goldRecognizer.apply(treebank);
		Iterator<SentenceItems> it = items.iterator();
		
		FileUtilities.createDir(getConfig().getOutputDir()+File.separator+MODIFIED_FTB_DIR);
		String mftb = getConfig().getOutputDir()+File.separator+MODIFIED_FTB+"_"+section+".mrg";
		OutputStreamWriter out = new OutputStreamWriter(new FileOutputStream(mftb),"UTF-8");
		
		FileUtilities.createDir(getConfig().getOutputDir()+File.separator+MWE_SUBTREES_DIR);
		String mwftb = getConfig().getOutputDir()+File.separator+MWE_SUBTREES+"_"+section+".mrg";
		OutputStreamWriter mwout = new OutputStreamWriter(new FileOutputStream(mwftb),"UTF-8");
		
		for(ParseTree tree:treebank){
			 writeMweParseTrees(preprocess(tree,it.next()),mwout);
			 Treebank.writeParseTree(tree, out);
		}
		
		out.close();
		mwout.close();
	}
	
	@Override
	public void learn(Treebank trainTreebank,Treebank validTreebank,Treebank testTreebank) throws Exception{
		String outputDir = getConfig().getOutputDir();
		//learn CRF model if required
		SimpleMweRecognizer mwer = new SimpleMweRecognizer(getConfig());
		CrfMweRecognizer mwer2 = new CrfMweRecognizer(getConfig(), mwer);
		mwer2.learnModel(false, trainTreebank);
		
		//write training files in output directory
		preprocess(trainTreebank,"1");
		preprocess(validTreebank,"2");
		preprocess(testTreebank,"3");
		
		
		//learn global PCFG
		System.err.println("Learning global grammar....");
		BKYParserWrapper parser = new BKYParserWrapper(outputDir);
		parser.grammarTrain("grammar", outputDir+File.separator+MODIFIED_FTB_DIR, getIterNumber());
		//learn MWE specific PCFG
		System.err.println("Learning mwe-specific grammar....");
		BKYParserWrapper parser2 = new BKYParserWrapper(outputDir);
		parser2.grammarTrain("mwes", outputDir+File.separator+MWE_SUBTREES_DIR, getIterNumber());
		System.err.println("Done.");
	    loadParsers();		
	}

	private static List<String> getItemSequence(PreprocessedItem item,Sentence<String> sentence){
		List<String> res = new LinkedList<String>();
		for(int i = item.getStartPosition() ; i <= item.getEndPosition() ; i++){
			res.add(sentence.get(i));
		}
		return res;
	}
	
	private ParseTree postprocess(Sentence<String> sentence,ParseTree tree, PreprocessedItem item){
		//System.err.println(item);
		Tree<String> term = tree.getTerminals().get(item.getStartPosition());
		Tree<String> preterm = tree.getParent(term);
		Tree<String> pretermParent = tree.getParent(preterm);
		int index = pretermParent.getChildren().indexOf(preterm);		
		ParseTree subtree = getter.get(getItemSequence(item, sentence), forcePreterminal?item.getPreterminal():preterm.getLabel());		
		//System.err.println("==>"+subtree);
		pretermParent.getChildren().remove(index); //remove preterminal node
		//replace it by mwe subtree
		tree.addNodesAt(pretermParent, index, subtree.getTree().getChildren());
		//System.err.println("****"+new ParseTree(tree.getTree()));
		return new ParseTree(tree.getTree());
	}
	/*
	private ParseTree cleanTree(ParseTree tree){
		Tree<String> root = tree.getTree();
		for(Tree<String> t:root.getPostOrderTraversal()){
			String[] tab = t.getLabel().split("\\+");
			t.setLabel(tab[0]);
		}
		return new ParseTree(root);
	}
	*/
	public ParseTree postprocess(Sentence<String> sentence, ParseTree tree, SentenceItems items){
		if(items == null){
			return tree;
		}
		items.sort();
		//System.err.println("===");
		//System.err.println(items.size());
		if(isWithMweAnnotation()){
			addIntermediateParse(tree==null?null:tree.createCopy());
		}
		for(PreprocessedItem item:items){			
			tree = postprocess(sentence,tree,item);
			//tree = cleanTree(tree);
		}
		return tree;
	}
	
	@Override
	public ParseTree parse(Sentence<String> sentence, SentenceItems items) {		
		Sentence<String> modified = new Sentence<String>(new LinkedList<String>(sentence.getSequence())); 
		modified = preprocess(modified,items);
		ParseTree tree = gParser.parse(modified);		
		tree = postprocess(sentence,tree,items);
		return tree;
	}
	
	
}
