/**
 * 
 *
 * Software: LGTools
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.parsing;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import edu.berkeley.nlp.ling.Tree;

/**
 * Wrapper of a Tree<String> from Berkeley Parser
 * 
 * @author Matthieu Constant
 *
 */
public class ParseTree {
	final private Tree<String> tree;
	final private List<Tree<String>> terminals;
	final private Map<Tree<String>,Integer> positions = new HashMap<Tree<String>, Integer>();
	final private Map<Tree<String>,Tree<String>> parents = new HashMap<Tree<String>, Tree<String>>();
	
	
	
	
	public ParseTree(Tree<String> tree){
		this.tree = tree;
		fillParents();
		terminals = tree.getTerminals();
		int pos = 0;
		for(Tree<String> node:terminals){
			positions.put(node, pos);
			pos++;
		}
	}
	
	
	private void fillParents(){
		fillParents(tree);		
	}
	
	private void fillParents(Tree<String> parent){
		for(Tree<String> c:parent.getChildren()){
			parents.put(c, parent);
			fillParents(c);
		}
	}
	
	public List<Tree<String>> getTerminals(){
		return terminals;
	}
	
	/**
	 * get node parent
	 * @param node tree node
	 * @return the node parent
	 */
	
	public Tree<String> getParent(Tree<String> node){
		return parents.get(node);
	}
	
	
	/**
	 * get all ancestors of a node
	 * 
	 * @param node tree node
	 * @return all ancestors of node
	 */
	
	public List<Tree<String>> getAncestors(Tree<String> node){
		List<Tree<String>> ancestors = new LinkedList<Tree<String>>();
		Tree<String> t;
		while((t = getParent(node)) != null){			
			ancestors.add(t);
			node = t;
		}
		return ancestors;
	}
	
	
	public Tree<String> getFirstCommonAncestor(int startPos,int endPos){
		return getFirstCommonAncestor(getTerminals().get(startPos),getTerminals().get(endPos));
	}
	
	/**
	 * get first common ancestor of two nodes
	 * 
	 * @param node1 tree node
	 * @param node2 tree node
	 * @return the first common ancestor of the two nodes
	 */
	
	public Tree<String> getFirstCommonAncestor(Tree<String> node1,Tree<String> node2){
		List<Tree<String>> ancestors1 = getAncestors(node1);
		List<Tree<String>> ancestors2 = getAncestors(node2);
		for(Tree<String> n1:ancestors1){
			for(Tree<String> n2:ancestors2){
				if(n1 == n2){
					return n1;
				}
			}
		}
		
		return tree;
	}
	
	public int getTerminalPosition(Tree<String> node){
		Integer pos = positions.get(node);
		if(pos == null){
			throw new RuntimeException("Node is not a terminal node!!\n"+node);
		}
		return pos;
	}
	
	public List<Tree<String>> getAncestorsInFirstCommonAncestorChildren(int startTerminalNode,int endTerminalNode){
		Tree<String> parent = getFirstCommonAncestor(terminals.get(startTerminalNode), terminals.get(endTerminalNode));
		LinkedHashSet<Tree<String>> res = new LinkedHashSet<Tree<String>>();
		Set<Tree<String>> children = new HashSet<Tree<String>>(parent.getChildren());
		for(int pos = startTerminalNode ; pos <= endTerminalNode ; pos++){
			for(Tree<String> node:getAncestors(terminals.get(pos))){
				if(children.contains(node)){
					res.add(node);
				}
			}						
		}
		List<Tree<String>> res2 = new LinkedList<Tree<String>>(res);
		return res2;
	}
	
	
	/**
	 * 
	 * @param node tree node
	 * @return the left-most terminal symbol of the subtree of the node 
	 */
	
	public Tree<String> getLeftMostTerminal(Tree<String> node){
		if(node.isLeaf()){
			return node;
		}
		return getLeftMostTerminal(node.getChildren().get(0));
	}
	
	/**
	 * 
	 * @param node tree node
	 * @return the right-most terminal symbol of the subtree of the node 
	 */
	
	public Tree<String> getRightMostTerminal(Tree<String> node){
		if(node.isLeaf()){
			return node;
		}
		return getRightMostTerminal(node.getChildren().get(node.getChildren().size() - 1));
	}
	
	
	/**
	 * test whether considering the sequence of terminals from a start position to an end position, as a Multi-word expression
	 * breaks the syntactic structure of sentence 
	 * 
	 * @param startTerminalPosition start position of multiword expression
	 * @param endTerminalPosition end position of multiword expression
	 * @return whether the multiword expression breaks the syntactic structure of the sentence
	 */
	
	public boolean isBreakingSyntacticStructure(int startTerminalPosition,int endTerminalPosition){
		if(startTerminalPosition == endTerminalPosition) {
			return false;
		}
		List<Tree<String>> c = getAncestorsInFirstCommonAncestorChildren(startTerminalPosition,endTerminalPosition);		
		int lpos = getTerminalPosition(getLeftMostTerminal(c.get(0)));
		int rpos = getTerminalPosition(getRightMostTerminal(c.get(c.size() - 1)));		
		return (lpos != startTerminalPosition) || (rpos != endTerminalPosition);
	}
	
	/**
	 * 
	 * @return the terminal symbols in sentence order
	 */
	
	public List<String> getTerminalSymbols(){
		List<String> symbols = new LinkedList<String>();
		for(Tree<String> s:getTerminals()){
			symbols.add(s.getLabel());
		}		
		return symbols;
	}
	
	/**
	 * 
	 * @return the preterminal symbols in sentence order
	 */
	
	public List<String> getPreTerminalSymbols(){
		List<String> symbols = new LinkedList<String>();
		for(Tree<String> s:getTerminals()){
			symbols.add(getParent(s).getLabel());
		}		
		return symbols;
	}
	
	
	
	
	/**
	 * @return the tree
	 */
	public Tree<String> getTree() {
		return tree;
	}

	@Override
	public String toString(){
		return tree.toString();
	}
	
	public void addNodesAt(Tree<String> parent,int index,List<Tree<String>> nodes){
		List<Tree<String>> children = new LinkedList<Tree<String>>(parent.getChildren());
		Collections.reverse(nodes);
		for(Tree<String> n:nodes){
			children.add(index, n);
		}
		parent.setChildren(children);
	}
	
	
	/*
	 * Create copy of a subtree
	 */
	private static Tree<String> copy(Tree<String> tree){
		Tree<String> t = new Tree<String>(tree.getLabel());
		if(tree.isLeaf()){
			return t;
		}
		List<Tree<String>> children = new LinkedList<Tree<String>>();
		for(Tree<String> c:tree.getChildren()){
			children.add(copy(c));
		}
		t.setChildren(children);		
		return t;
	}
	
	public ParseTree createCopy(){
		return new ParseTree(copy(this.getTree()));
	}
	
}
