/**
 *
 * Software: LGTools
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.parsing;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;

import fr.upemlv.lgtools.tagging.LGTaggerConfig;
import fr.upemlv.lgtools.tagging.impl.Utils;
import fr.upemlv.lgtools.text.PropertyConstants;

/**
 * 
 * @author Matthieu Constant
 *
 */

public class ParsingConfig {
   private final static String LANG = "LANG"; //language
   private final static String OUTPUT_DIR = "OUTPUT_DIR"; // directory path containing output
   private final static String LGTAGGER = "LGTAGGER"; //config file path
   public final static String LGTAGGER_MODEL_NAME = "lgtagger-model"; 
   private final static String MWE_TERMINAL = "MWE_TERMINAL"; //representation of terminal for MWEs
   private final static String MWE_PRETERMINAL = "MWE_PRETERMINAL"; //representation of preterminal for MWEs
   private final static String SIMPLIFICATION_LIMITATIONS = "SIMPLIFICATION_LIMITATIONS";
   
   private String lang = "French";
   private String mweTerminal = "$"+PropertyConstants.TOKEN+"$";
   private String mwePreTerminal = "$POS$";
   private String outputDir = null;
   private LGTaggerConfig lgtagger = null;
   private String lgtaggerConfigPath = null;
   
   
   public ParsingConfig() throws IOException{
	   this.outputDir = "/tmp";
	   this.lgtaggerConfigPath = Utils.getAbsolutePath("collocations-mwe", LANG, "config")+".cfg";
	   this.lgtagger = new LGTaggerConfig(lgtaggerConfigPath,true);
   }
   
   
   public ParsingConfig(String filename) throws IOException{
	   BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(filename),"UTF-8"));
	   String line;
	   while((line = reader.readLine()) != null){
		   processLine(line);   
	   }
	   reader.close();
	   if(lgtagger == null || outputDir == null){
		   throw new RuntimeException("Parsing configfile is incomplete! (cf. doc)");
	   }
   }
   
   private void processLine(String line) throws IOException{
	   line = line.trim();
	   if(line.startsWith("#") || line.isEmpty()){
		   return;
	   }
	   String[] tab = line.split("=");
	   if(tab.length != 2){
		   throw new RuntimeException("A line in the Parsing configfile is not in the good format! (cf. doc)");
	   }
	   String prop = tab[0];
	   String value = tab[1];
	   if(prop.equals(LANG)){
		   this.lang = value;
	   }
	   if(prop.equals(OUTPUT_DIR)){
		   this.outputDir = value;
		   File dir = new File(outputDir);
		   if(!dir.exists()){
			   System.err.println("Creating output directory "+dir+"...");
			   dir.mkdirs();
		   }
	   }
	   if(prop.equals(LGTAGGER)){
		   this.lgtaggerConfigPath = Utils.getAbsolutePath(value, this.lang, "config")+".cfg";
		   this.lgtagger = new LGTaggerConfig(this.lgtaggerConfigPath, true);
		   
	   }
	   if(prop.equals(MWE_PRETERMINAL)){
		   this.mwePreTerminal = value;
	   }
	   if(prop.equals(MWE_TERMINAL)){
		   this.mweTerminal = value;
	   }
   }

/**
 * @return the lang
 */
public String getLang() {
	return lang;
}


/**
 * @return the mweTerminal format
 */
public String getMweTerminalFormat() {
	return mweTerminal;
}



/**
 * @return the mwePreTerminal format
 */
public String getMwePreTerminalFormat() {
	return mwePreTerminal;
}



/**
 * @return the outputDir
 */
public String getOutputDir() {
	return outputDir;
}



/**
 * @return the lgtagger
 */
public LGTaggerConfig getLgtaggerConfig() {
	return lgtagger;
}

/**
 * 
 * @return the model path
 */
public String getModelPath(){
	return outputDir + File.separator + LGTAGGER_MODEL_NAME;
}


/**
 * @return the lgtaggerConfigPath
 */
public String getLgtaggerConfigPath() {
	return lgtaggerConfigPath;
}

   
   
}
