/**
 * 
 *
 * Software: LGTools
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.parsing;

import java.io.File;
import java.io.IOException;
import java.util.Iterator;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Switch;

import fr.upemlv.lgtools.parsing.mwe.AutomaticMweRecognizer;
import fr.upemlv.lgtools.parsing.mwe.CrfMweRecognizer;
import fr.upemlv.lgtools.parsing.mwe.FilteredGoldMweRecognizer;
import fr.upemlv.lgtools.parsing.mwe.MweRecognizer;
import fr.upemlv.lgtools.parsing.mwe.SimpleMweRecognizer;
import fr.upemlv.lgtools.text.FullRawText;
import fr.upemlv.lgtools.text.LightRawText;
import fr.upemlv.lgtools.text.LightTaggedText;
import fr.upemlv.lgtools.text.RawText;
import fr.upemlv.lgtools.text.Sentence;
import fr.upemlv.lgtools.utils.FileUtilities;

/**
 * @author Matthieu Constant
 *
 */
public class ParsingProcesses {
	private static String OUTPUT = "output";
	private static String PROCESS_TYPE = "type";
	private static String CONFIG = "config";
	private static String INPUT_TREEBANK_PATH_OR_NAME = "input_treebank";
	private static String INPUT_TEXT_PATH = "input_text";
	private static String MWE_ANNOTATION = "mwe_annotation";
	private static String GOLD_MWE_RECOGNITION = "gold_recognition";
	private static String GOLD_SUBTREES = "gold";
	private static String FORCE_PRETERMINALS = "force";
	private static String GRAMMAR = "grammar";
	
	private static  JSAP createJSAP() throws JSAPException{
		JSAP jsap = new JSAP();

		FlaggedOption opt;
		
		opt = new FlaggedOption(PROCESS_TYPE)
		.setStringParser(JSAP.STRING_PARSER)
		.setRequired(true)
		.setShortFlag('p')
		.setLongFlag(PROCESS_TYPE);
		opt.setHelp("Type of process: tcrf (train CRF model), tpre (preprocess at learning stage), tpcfg (train PCFG), ppre (preprocess at parsing stage), parse (parse text), ppost (postprocess at parsing stage)");
		jsap.registerParameter(opt);

		opt = new FlaggedOption(CONFIG)
		.setStringParser(JSAP.STRING_PARSER)
		.setShortFlag('c')
		.setRequired(true)
		.setLongFlag(CONFIG);
		opt.setHelp("Parsing configuration file path");
		jsap.registerParameter(opt);
			
		
			opt = new FlaggedOption(INPUT_TREEBANK_PATH_OR_NAME)
			.setStringParser(JSAP.STRING_PARSER)		
			.setShortFlag('i')
			.setLongFlag(INPUT_TREEBANK_PATH_OR_NAME)
			;
			//opt.setRequired(true);
			opt.setHelp("Path of input treebank");
			jsap.registerParameter(opt);
			
			opt = new FlaggedOption(OUTPUT)
			.setStringParser(JSAP.STRING_PARSER)		
			.setShortFlag('o')
			.setLongFlag(OUTPUT)
			;
			opt.setRequired(false);
			opt.setHelp("Name of output");
			jsap.registerParameter(opt);

			opt = new FlaggedOption(INPUT_TEXT_PATH)
			.setStringParser(JSAP.STRING_PARSER)        
			.setShortFlag('t')
			.setLongFlag(INPUT_TEXT_PATH);
			opt.setHelp("Input raw text path");
			opt.setRequired(false);
			jsap.registerParameter(opt);
			
			opt = new FlaggedOption(GRAMMAR)
			.setStringParser(JSAP.STRING_PARSER)        
			.setShortFlag('g')
			.setLongFlag(GRAMMAR);
			opt.setHelp("Grammar name");
			opt.setRequired(false);
			jsap.registerParameter(opt);
		
			Switch sw = new Switch(GOLD_SUBTREES)
			.setShortFlag('s')
			.setLongFlag(GOLD_SUBTREES);
			sw.setHelp("Retrieves gold mwe subtrees in crf mode");
			jsap.registerParameter(sw);
			
			sw = new Switch(FORCE_PRETERMINALS)
			.setShortFlag('f')
			.setLongFlag(FORCE_PRETERMINALS);
			sw.setHelp("force mwe preterminals in crf mode");
			jsap.registerParameter(sw);
			
			sw = new Switch(GOLD_MWE_RECOGNITION)
			.setShortFlag('m')
			.setLongFlag(GOLD_MWE_RECOGNITION);
			sw.setHelp("use gold mwe recognition");
			jsap.registerParameter(sw);
			
			sw = new Switch(MWE_ANNOTATION)
			.setShortFlag('a')
			.setLongFlag(MWE_ANNOTATION);
			sw.setHelp("also print parses with MWE annotations");
			jsap.registerParameter(sw);
			
			return jsap;
	}
	
	private static void usage(JSAP jsap){
		System.err.print("Usage: ");
		System.err.print("java "+ParsingProcesses.class.getCanonicalName()+" ");
		System.err.println(jsap.getUsage());
		System.err.print(jsap.getHelp());
		System.exit(1);
	}
	
	
	
	private static void parsingPostProcessing(ParsingConfig config,String referenceRawTextPath,String treebankPath,String itemsPath,String outputTreebankPath,boolean forcePreterminal,boolean goldSubtrees) throws IOException, InterruptedException{
		MweMatParser parser = new MweMatParser(config, true,forcePreterminal, goldSubtrees);
		parser.instanciateIntermediateParses();
		Treebank treebank = new Treebank(treebankPath);
		RawText text = new FullRawText(referenceRawTextPath, "UTF-8", config.getLang(), false);
		TextItems items = new TextItems(itemsPath);
		Iterator<SentenceItems> it = items.iterator();
		Iterator<Sentence<String>> it2 = text.iterator(); 
		Treebank res = new Treebank();
		for(ParseTree tree:treebank){
			res.addTree(parser.postprocess(it2.next(), tree, it.next()));
		}
		res.save(outputTreebankPath);
	}
	
	private static void parsingPreprocessing(ParsingConfig config,String textPath,String goldCorpusPath,boolean isGold,String outputPrefix) throws Exception{
		Treebank gold = null;
		RawText text = new FullRawText(textPath, "UTF-8", config.getLang(), false);
		SimpleMweRecognizer smwer = new SimpleMweRecognizer(config);				
		MweRecognizer mwer;
		if(!isGold){
			mwer = new CrfMweRecognizer(config, smwer);
		}
		else{
			gold = new Treebank(goldCorpusPath);
			mwer = new FilteredGoldMweRecognizer(config.getLgtaggerConfig().getTagset(), smwer);
		}		
		MweMatParser parser = new MweMatParser(config, false,false, false);
		TextItems items = mwer.apply(text,gold);
		Iterator<SentenceItems> it = items.iterator();
		FullRawText res = new FullRawText();
		for(Sentence<String> sentence:text){
		  res.addSentence(parser.preprocess(sentence, it.next()));
		}		
		items.save(outputPrefix+".items");
		res.saveSegmented(outputPrefix+".txt");
	}
	
	
	private static void trainingPreprocessing(ParsingConfig config, String trainingCorpusPath,String outputName) throws Exception{
		Treebank treebank = new Treebank(trainingCorpusPath);
		MweMatParser parser = new MweMatParser(config, 1);
		AutomaticMweRecognizer simpleMweRecognizer = new SimpleMweRecognizer(parser.getConfig());
		MweRecognizer goldRecognizer = new FilteredGoldMweRecognizer(parser.getConfig().getLgtaggerConfig().getTagset(), simpleMweRecognizer);
		
		TextItems items = goldRecognizer.apply(treebank);
		Iterator<SentenceItems> it = items.iterator();
		Treebank mwes = new Treebank();
		for(ParseTree tree:treebank){
			mwes.addTrees(parser.preprocess(tree, it.next()));			
		}
		String dir = config.getOutputDir()+File.separator+"mftb";
		if(!new File(dir).exists()){new File(dir).mkdirs();}
		treebank.save(dir+File.separator+outputName+".mrg");
		dir = config.getOutputDir()+File.separator+"mwes";
		if(!new File(dir).exists()){new File(dir).mkdirs();}
		mwes.save(dir+File.separator+outputName+".mrg");
	}

	private static void learnCRFModel(ParsingConfig config,String trainingCorpusPath) throws Exception{
		
		MweMatParser parser = new MweMatParser(config, 1);
		Treebank trainTreebank = new Treebank(trainingCorpusPath);

		//learn CRF model if required
		SimpleMweRecognizer mwer = new SimpleMweRecognizer(parser.getConfig());
		CrfMweRecognizer mwer2 = new CrfMweRecognizer(parser.getConfig(), mwer);
		mwer2.learnModel(false, trainTreebank);
	}
	
	private static String getOutput(String output,ParsingConfig config){
		if(output == null){
			System.err.println("Output name is required");
			System.exit(1);
		}
		return config.getOutputDir()+File.separator+output;
	}
	
	private static String getPath(String filename,ParsingConfig config){
		if(filename == null){
			System.err.println("Filename is required");
			System.exit(1);
		}
		if(new File(filename).exists()){
			return filename;
		}
		return config.getOutputDir()+File.separator+filename;
	}
	
	
	public static void main(String[] args) throws Exception{
		JSAP jsap = createJSAP();
		JSAPResult arguments = jsap.parse(args);
		if(!arguments.success()){
			usage(jsap);
		}
		String configPath = arguments.getString(CONFIG);
		ParsingConfig config = new ParsingConfig(configPath);
		String process = arguments.getString(PROCESS_TYPE);
		String input = arguments.getString(INPUT_TREEBANK_PATH_OR_NAME);
		String tinput = arguments.getString(INPUT_TEXT_PATH);
		String output = arguments.getString(OUTPUT);
		String grammar = arguments.getString(GRAMMAR);
		boolean goldMWER = arguments.getBoolean(GOLD_MWE_RECOGNITION);
		boolean forcePreterminals = arguments.getBoolean(FORCE_PRETERMINALS);
		boolean  goldSubtrees = arguments.getBoolean(GOLD_SUBTREES);
		
		if(process.equals("parse")){
			if(grammar == null){
				System.err.println("There should be a grammar name");
				usage(jsap);
			}
			BKYParserWrapper bkyparser = new BKYParserWrapper(config.getOutputDir());
			bkyparser.parse(grammar, tinput+".txt", getPath(output,config)+".mrg");
		    FileUtilities.copy(tinput+".items", getPath(output,config)+".items");
		}
		
		if(input == null){
			System.err.println("There should be an input treebank");
			usage(jsap);
		}
		
		if(process.equals("tcrf")){			
			learnCRFModel(config, input);
		}
		if(process.equals("tpre")){			
			trainingPreprocessing(config, input, output);
		}
		
		if(process.equals("tpcfg")){
			
			BKYParserWrapper bkyparser = new BKYParserWrapper(config.getOutputDir());
			bkyparser.grammarTrain(output, getPath(input,config), 5);			
		}
		
		if(process.equals("mrg2txt")){
			Treebank ftb = new Treebank(input);
			RawText text = new LightRawText(new LightTaggedText(config.getLgtaggerConfig().getTagset(), ftb));
			text.saveAsIs(getPath(output,config)+".txt");			
		}
		
		if(process.equals("ppre")){
			parsingPreprocessing(config, tinput, input, goldMWER, getOutput(output,config));
		}
		
		if(process.equals("ppost")){
			if(tinput == null){
				System.err.println("There should be the initial raw text in the arguments (-t)");
				usage(jsap);
			}
			parsingPostProcessing(config, getPath(tinput,config), getPath(input,config)+".mrg", getPath(input,config)+".items", getOutput(output, config)+".mrg", forcePreterminals, goldSubtrees);
		}
		
		
	}
	

}
