/**
 * 
 *
 * Software: LGTools
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.parsing;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import edu.berkeley.nlp.ling.Tree;
import fr.upemlv.lgtools.parsing.mwe.AutomaticMweRecognizer;
import fr.upemlv.lgtools.parsing.mwe.FilteredGoldMweRecognizer;
import fr.upemlv.lgtools.parsing.mwe.MweRecognizer;
import fr.upemlv.lgtools.parsing.mwe.SimpleMweRecognizer;
import fr.upemlv.lgtools.text.RawText;
import fr.upemlv.lgtools.text.Sentence;
import fr.upemlv.lgtools.utils.FileUtilities;

/**
 * Parser integrating MW temporary nodes at training preprocessing 
 * and removing them at test postprocessing
 * 
 * @author Matthieu Constant
 *
 */

public class SimpleMweMatParser extends MatParser{
	private final static String MODIFIED_FTB2_DIR = "mft2";
	private final static String MODIFIED_FTB2 = MODIFIED_FTB2_DIR+File.separator+"mftb2";
	private final static String GRAMMAR = "tmpGrammar";
	private final static String SPECIAL_CHARACTER = "*";
	private BKYParser parser = null;
	
	
	public SimpleMweMatParser(ParsingConfig config,int iterLearn) {
		this(config,false,iterLearn);
	}
	
	public SimpleMweMatParser(ParsingConfig config) {
		this(config,false,-1);
	}
		
	public SimpleMweMatParser(ParsingConfig config,boolean withMweAnnotation) {
		this(config,withMweAnnotation,-1);
	}
	
		
	public SimpleMweMatParser(ParsingConfig config,boolean withMweAnnotation,int iterLearn) {
		super(config,iterLearn,withMweAnnotation);
		String grammarPath = config.getOutputDir()+File.separator+GRAMMAR+".gr";
		if(new File(grammarPath).exists()){
			this.parser = new BKYParser(grammarPath);
		}
	}

	private void preprocess(ParseTree tree,PreprocessedItem item){
		List<Tree<String>> nodes;
		int start = item.getStartPosition();
		int end = item.getEndPosition();
		
		nodes = tree.getAncestorsInFirstCommonAncestorChildren(start, end);
		Tree<String> parent = tree.getFirstCommonAncestor(start, end); 
		List<Tree<String>> children = new LinkedList<Tree<String>>(parent.getChildren());
		int index = children.indexOf(nodes.get(0));		
		children.removeAll(nodes);
		Tree<String> tempNode = new Tree<String>(SPECIAL_CHARACTER+item.getPreterminal()+SPECIAL_CHARACTER);
		tempNode.setChildren(nodes);
		children.add(index,tempNode);
		parent.setChildren(children);
	}
	
	private void preprocess(ParseTree tree,SentenceItems items){
		for(PreprocessedItem item:items){
			preprocess(tree,item);
		}
	}
	
	private void preprocess(Treebank treebank,String section) throws Exception{
		AutomaticMweRecognizer simpleMweRecognizer = new SimpleMweRecognizer(getConfig());
		MweRecognizer goldRecognizer = new FilteredGoldMweRecognizer(getConfig().getLgtaggerConfig().getTagset(), simpleMweRecognizer);
		TextItems items = goldRecognizer.apply(treebank);
		Iterator<SentenceItems> it = items.iterator();
		
		FileUtilities.createDir(getConfig().getOutputDir()+File.separator+MODIFIED_FTB2_DIR);
		String mftb = getConfig().getOutputDir()+File.separator+MODIFIED_FTB2+"_"+section+".mrg";
		OutputStreamWriter out = new OutputStreamWriter(new FileOutputStream(mftb),"UTF-8");
		
		for(ParseTree tree:treebank){
			 preprocess(tree,it.next());
			 Treebank.writeParseTree(tree, out);
		}
		
		out.close();
	}
	
	@Override
	public void learn(Treebank trainTreebank, Treebank validtreebank,
			Treebank testtreebank) throws Exception {
		preprocess(trainTreebank,"1");
		preprocess(validtreebank,"2");
		preprocess(testtreebank,"3");
		System.err.println("Learning grammar....");
		BKYParserWrapper parser = new BKYParserWrapper(getConfig().getOutputDir());
		parser.grammarTrain(GRAMMAR, getConfig().getOutputDir()+File.separator+MODIFIED_FTB2_DIR, getIterNumber());
		
	}


	private static boolean isTmpNode(Tree<String> node){
		return node.getLabel().startsWith(SPECIAL_CHARACTER) && node.getLabel().endsWith(SPECIAL_CHARACTER); 
	}
	
	private void modifyTree(Tree<String> parent,Tree<String> node){
		List<Tree<String>> pchildren = new LinkedList<Tree<String>>(parent.getChildren());
		parent.setChildren(pchildren);
		List<Tree<String>> children = new LinkedList<Tree<String>>(node.getChildren());
		node.setChildren(children);
		Tree<String> newParent = node;
		if(isTmpNode(node)){
			int index = pchildren.indexOf(node);
			pchildren.remove(index);
			int i = index;
			for(Tree<String> c:children){
				pchildren.add(i, c);
				i++;
			}
			newParent = parent;
		}
		for(Tree<String> c:children){
			modifyTree(newParent, c);
		}
	}
	
	private ParseTree postprocess(ParseTree tree){
		if(tree == null){
			return null;
		}
		Tree<String> root = tree.getTree();
	    modifyTree(root,root.getChildren().get(0));		
	    return new ParseTree(root);
	}
	
	@Override
	public ParseTree parse(Sentence<String> sentence, SentenceItems items) {
		ParseTree tree = parser.parse(sentence);
		if(isWithMweAnnotation()){
			addIntermediateParse(tree==null?null:tree.createCopy());
		}
		return postprocess(tree);
	}

	
	
	public Treebank parse(RawText text){
		return parse(text,null);		
	}
	
	
}
