/**
 * 
 *
 * Software: LGTools
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.parsing.mwe;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import fr.upemlv.lgtools.parsing.ParsingConfig;
import fr.upemlv.lgtools.parsing.PreprocessedItem;
import fr.upemlv.lgtools.parsing.SentenceItems;
import fr.upemlv.lgtools.parsing.TextItems;
import fr.upemlv.lgtools.parsing.Treebank;
import fr.upemlv.lgtools.tagging.LGTaggerConfig;
import fr.upemlv.lgtools.tagging.LGTaggerConstants;
import fr.upemlv.lgtools.tagging.lex.Tagset;
import fr.upemlv.lgtools.text.LightRawText;
import fr.upemlv.lgtools.text.LightTaggedText;
import fr.upemlv.lgtools.text.MultiFeaturedElement;
import fr.upemlv.lgtools.text.PropertyConstants;
import fr.upemlv.lgtools.text.PropertyExtractor;
import fr.upemlv.lgtools.text.PropertyExtractorFactory;
import fr.upemlv.lgtools.text.RawText;
import fr.upemlv.lgtools.text.Sentence;

/**
 * @author Matthieu Constant
 *
 */
public abstract class AutomaticMweRecognizer extends MweRecognizer{

	//final protected ParsingConfig parserConfig;
	final protected LGTaggerConfig taggerConfig;
	final String mweTerminalFormat;
	final String mwePreTerminalFormat;
	final private Map<String,Integer> extractorIndex = new HashMap<String,Integer>();
	final protected List<PropertyExtractor> extractors = new ArrayList<PropertyExtractor>();
	private int currentStart = -1;
	private int currentEnd = -1;
	private List<String> propertyValues = null;

	
	protected AutomaticMweRecognizer(ParsingConfig config){
		//this.parserConfig = config;
		this.taggerConfig = config.getLgtaggerConfig();
		this.extractorIndex.put(PropertyConstants.MWE_IOB, 0);
		this.extractors.add(PropertyExtractorFactory.createMWEIOBPropertyExtractor());
		extractorIndex.put("POS",1);					  
		extractors.add(PropertyExtractorFactory.createMWEPropertyExtractor("POS"));
		this.mweTerminalFormat = config.getMweTerminalFormat();
		this.mwePreTerminalFormat = config.getMwePreTerminalFormat();
	    loadPropertyExtractors(mweTerminalFormat);
	    loadPropertyExtractors(mwePreTerminalFormat);	    
	}
	
	protected AutomaticMweRecognizer(LGTaggerConfig lgtConfig){
		//this.parserConfig = null;
		this.taggerConfig = lgtConfig;
		this.mweTerminalFormat = "$"+PropertyConstants.TOKEN+"$";
		this.mwePreTerminalFormat = "$POS$";
		
		this.extractorIndex.put(PropertyConstants.MWE_IOB, 0);
		this.extractors.add(PropertyExtractorFactory.createMWEIOBPropertyExtractor());
		loadPropertyExtractors(mweTerminalFormat);
	    loadPropertyExtractors(mwePreTerminalFormat);
	}
	
	
	private void loadPropertyExtractors(String format){		
		Pattern pattern = Pattern.compile("\\$([^$]+)\\$");
		Matcher m = pattern.matcher(format);
		while(m.find()){
			String prop = m.group(1);
			if(PropertyConstants.TOKEN.equals(prop)){
				if(!extractorIndex.containsKey(PropertyConstants.TOKEN)){
				  extractorIndex.put(prop,extractorIndex.size());				  
				  extractors.add(PropertyExtractorFactory.createTokenPropertyExtractor());
				}
			}
			else{
				if(PropertyConstants.SIMPLIFICATION.equals(prop)){
					if(!extractorIndex.containsKey(PropertyConstants.SIMPLIFICATION)){
						extractorIndex.put(prop,extractorIndex.size());				  
						  extractors.add(PropertyExtractorFactory.createSimplificationPropertyExtractor());
					}
				}else{

					if(!extractorIndex.containsKey(prop)){
						extractorIndex.put(prop,extractorIndex.size());					  
						extractors.add(PropertyExtractorFactory.createMWEPropertyExtractor(prop));
					}
				}
			}
		}		
	}
	
	private String instanciateFormatTemplate(String format){
		String res = format;
		for(String prop:extractorIndex.keySet()){
			res = res.replaceAll("\\$"+prop+"\\$", propertyValues.get(extractorIndex.get(prop)));
		}
		return res;
	}
	
	private PreprocessedItem createPreprocessedItem(){
		String preterminal = instanciateFormatTemplate(mwePreTerminalFormat);
		String terminal = instanciateFormatTemplate(mweTerminalFormat);
		String partOfSpeech = propertyValues.get(extractorIndex.get("POS"));
		return new PreprocessedItem(currentStart, currentEnd, terminal, preterminal,partOfSpeech);
	}
	
	
	private void init(){
		currentStart = -1;
		currentEnd = -1;
		propertyValues = null;
	}
	
	private void updateB(int currentPos, MultiFeaturedElement mfe){
		currentStart = currentPos;
		currentEnd = currentPos;
		propertyValues = new ArrayList<String>();
		for(PropertyExtractor prop:extractors){
			propertyValues.add(mfe.get(prop.getPropertyName()));
		}		
	}
	
	private void updateI(int currentPos, MultiFeaturedElement mfe){
		currentEnd = currentPos;
		Integer index = extractorIndex.get(PropertyConstants.TOKEN);
		if(index != null){
		   propertyValues.set(index,propertyValues.get(index)+LGTaggerConstants.TOKEN_SEPARATOR+mfe.get(PropertyConstants.TOKEN));
		}
	}
	
	
	protected SentenceItems extractSentenceMWEs(Sentence<MultiFeaturedElement> sentence){
		SentenceItems items = new SentenceItems();
		int currentPos = 0;
		for(MultiFeaturedElement mfe:sentence){
						
			if(isB(mfe)){
				if(currentStart != -1){
					items.addItem(createPreprocessedItem());					
				}
			    updateB(currentPos,mfe);
			}
			if(isI(mfe)){
				updateI(currentPos,mfe);
			}
			if(isO(mfe)){
				if(currentStart != -1){
					items.addItem(createPreprocessedItem());					
				}
			    init();
			}
			currentPos++;
		}
		if(currentStart != -1){
			items.addItem(createPreprocessedItem());			
		}
		init();
		return items;
	}
	
	protected abstract boolean isB(MultiFeaturedElement mfe);
	protected abstract boolean isI(MultiFeaturedElement mfe);
	protected abstract boolean isO(MultiFeaturedElement mfe);
	
	@Override
	public TextItems apply(Treebank treebank)
			throws Exception {
		//we transform treebank in raw text
		return apply(new LightRawText(new LightTaggedText(taggerConfig.getTagset(), treebank)));
	}
	
	@Override
	public TextItems apply(RawText text,Treebank treebank) throws Exception{
		return apply(text);
	}
	
}
