/**
 * 
 *
 * Software: LGTools
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */
package fr.upemlv.lgtools.parsing.mwe;

import java.io.File;
import java.util.Iterator;
import java.util.List;

import fr.upemlv.lgtools.parsing.ParseTree;
import fr.upemlv.lgtools.parsing.ParsingConfig;
import fr.upemlv.lgtools.parsing.PreprocessedItem;
import fr.upemlv.lgtools.parsing.SentenceItems;
import fr.upemlv.lgtools.parsing.TextItems;
import fr.upemlv.lgtools.parsing.Treebank;
import fr.upemlv.lgtools.tagging.LGTaggerParameters;
import fr.upemlv.lgtools.tagging.TaggingEnvironment;
import fr.upemlv.lgtools.tagging.TrainEnvironment;
import fr.upemlv.lgtools.tagging.impl.LGTagger;
import fr.upemlv.lgtools.tagging.impl.LGTaggerFactory;
import fr.upemlv.lgtools.tagging.impl.LGTaggerLearner;
import fr.upemlv.lgtools.tagging.impl.LGTaggerPreprocessor;
import fr.upemlv.lgtools.tagging.impl.PreprocessingContext;
import fr.upemlv.lgtools.tagging.lex.InternalTagset;
import fr.upemlv.lgtools.tagging.lex.Tag;
import fr.upemlv.lgtools.tagging.lex.Tagset;
import fr.upemlv.lgtools.text.FullTaggedText;
import fr.upemlv.lgtools.text.LightRawText;
import fr.upemlv.lgtools.text.LightTaggedText;
import fr.upemlv.lgtools.text.MultiFeaturedElement;
import fr.upemlv.lgtools.text.MultiFeaturedText;
import fr.upemlv.lgtools.text.RawText;
import fr.upemlv.lgtools.text.Sentence;
import fr.upemlv.lgtools.text.TaggedElement;
import fr.upemlv.lgtools.text.TaggedText;

/**
 * CRF-based MWE Recognizer
 * 
 * @author Matthieu Constant
 *
 */

public class CrfMweRecognizer extends AutomaticMweRecognizer{

	private final AutomaticMweRecognizer prerecognizer;
	private final String modelPath;
	private final String configPath;
    
	
	/**
	 * 
	 * 
	 * @param config parsing config file
	 * @param prerecognizer MWE pre-recognizer
	 * @param forceTraining force training eventhough CRF model already exists
	 * @param trainingTreebank treebank for building the training corpus
	 */
    
	public CrfMweRecognizer(ParsingConfig config, AutomaticMweRecognizer prerecognizer) {
		super(config);
		this.modelPath = config.getModelPath();
		this.prerecognizer = prerecognizer;
		this.configPath = config.getLgtaggerConfigPath();
	}
	
	
	private Tag createEmptyTag(){
		Tag t = new Tag(taggerConfig.getTagset());
		t.addCat(InternalTagset.EMPTY);
		t.addCat(InternalTagset.B);
		return t;
	}
	
	private Tag createTag(String cat,boolean isB){
		Tag t = new Tag(taggerConfig.getTagset());
		t.addCat(cat);
		t.addCat(isB?InternalTagset.B:InternalTagset.I);
		return t;
	}
	
	
	private TaggedElement processWord(String word,int pos, PreprocessedItem nextItem){
		if(nextItem == null){ //EMPTY	    			
			return new TaggedElement(word, createEmptyTag());
		}
		else{			
		   if((pos < nextItem.getStartPosition())){ //EMPTY			  
			   return new TaggedElement(word, createEmptyTag());
		   }
		   else{
			  if(pos == nextItem.getStartPosition()){
				  return  new TaggedElement(word,createTag(nextItem.getPartOfSpeech(), true));
			  }
			  else{
				  return  new TaggedElement(word,createTag(nextItem.getPartOfSpeech(), false));
			  }
		   }
		}
	}
	
	
	/* build the training corpus to learn CRF model
	 */
	
	private TaggedText buildTrainingCorpus(Treebank trainingTreebank) throws Exception{
		Tagset tagset = taggerConfig.getTagset();
	    RawText text = new LightRawText(new LightTaggedText(tagset, trainingTreebank));
	    TextItems rawMwes = prerecognizer.apply(text);
	    Iterator<SentenceItems> listIt = rawMwes.iterator();
	    TaggedText res = new FullTaggedText(tagset);
	    int cntBefore = 0;
	    int cntAfter = 0;
	    for(ParseTree sentenceTree:trainingTreebank){
	    	PreprocessedItem nextItem = null;
	    	List<String> sentenceText = sentenceTree.getTerminalSymbols();
	    	SentenceItems items = listIt.next();
	    	cntBefore += items.size();
	    	items = FilteredGoldMweRecognizer.filterItems(items, sentenceTree,taggerConfig.getProjectedCats().isEmpty()?null:taggerConfig.getProjectedCats());
	    	cntAfter += items.size();
	    	Iterator<PreprocessedItem> itemIt = items.iterator();
	    	if(itemIt.hasNext()){
	    		nextItem = itemIt.next();
	    	}	    	
	    	Sentence<TaggedElement> sentence = new Sentence<TaggedElement>();
	    	res.addSentence(sentence);
	    	for(int pos = 0 ; pos < sentenceText.size() ; pos++){
	    		sentence.add(processWord(sentenceText.get(pos),pos,nextItem));
	    		if((nextItem != null) && (pos == nextItem.getEndPosition())){
	    			nextItem = null;
	    			if(itemIt.hasNext()){
	    	    		nextItem = itemIt.next();
	    	    	}
	    		}
	    	}
	    }
	    //System.err.println(cntBefore+"=>"+cntAfter);
	    return res;
	}

	
	private boolean modelExists(){
		return new File(modelPath+".wap").exists();
	}
	
	
	public void learnModel(boolean forceTraining, Treebank trainingTreebank) throws Exception{
		if(trainingTreebank == null){
			throw new RuntimeException("Cannot learn model because treebank is missing...");
		}
		if((!forceTraining) & (modelExists()) ){
			System.err.println("Skip MWE Recognizer training...");
			return;
		}
		System.err.println("## Learning CRF-based MWE Recognizer ##");
		System.err.println("Constructing training corpus...");
		TaggedText trainCorpus = buildTrainingCorpus(trainingTreebank);
		//trainCorpus.save("/tmp/save.model.txt");
		LGTaggerParameters parameters = new LGTaggerParameters(modelPath, true, true);
		parameters.setConfig(configPath);
		parameters.setDecomposedText(true);
		TrainEnvironment env = new TrainEnvironment(parameters);
		PreprocessingContext context = new PreprocessingContext(env);
		LGTaggerPreprocessor preprocessor = new LGTaggerPreprocessor(context); 
		LGTaggerLearner crfLearner = new LGTaggerLearner(env, preprocessor);
		System.err.println("Learning model...");
		crfLearner.perform(trainCorpus, null);
		System.err.println("Done.");
	}
	
	
	@Override
	public TextItems apply(RawText text) throws Exception {		
		if(!modelExists()){
			throw new RuntimeException("Cannot apply MWE Recognizer as the CRF model has not been learnt!!");		
		}
		LGTaggerParameters parameters = new LGTaggerParameters(modelPath,true,false);
		parameters.setDecomposedText(true);
		TaggingEnvironment env = new TaggingEnvironment(parameters);		
		LGTagger lgtagger = LGTaggerFactory.createStandardSingleModelTagger(env);
		TaggedText t = lgtagger.perform(text);
		PreprocessingContext context = new PreprocessingContext(taggerConfig.getLang(), taggerConfig.getResources(), extractors, taggerConfig.getTagset());
		LGTaggerPreprocessor processor = new LGTaggerPreprocessor(context);
		MultiFeaturedText mtext = processor.preprocess(t);
		
		TextItems list = new TextItems();
		for(Sentence<MultiFeaturedElement> sentence: mtext){
			list.addSentence(extractSentenceMWEs(sentence));
		}
		return list;
		
	}

	@Override
	protected boolean isB(MultiFeaturedElement mfe) {
		if(isO(mfe)){
			return false;
		}
		Tag tag = mfe.getTag();
		String v = tag.getValue(InternalTagset.IOB);		
		return InternalTagset.B.equals(v);
	}

	@Override
	protected boolean isI(MultiFeaturedElement mfe) {
		Tag tag = mfe.getTag();
		String v = tag.getValue(InternalTagset.IOB);		
		return InternalTagset.I.equals(v);		
	}

	@Override
	protected boolean isO(MultiFeaturedElement mfe) {
		Tag tag = mfe.getTag();
		String v = tag.getValue("POS");
		return InternalTagset.EMPTY.equals(v);
	}


	
	
	
	

}
