/**
 * 
 *
 * Software: LGTagger
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.tagging;

import java.io.IOException;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import fr.upemlv.lgtools.tagging.impl.Utils;
import fr.upemlv.lgtools.tagging.lex.InternalTagset;
import fr.upemlv.lgtools.tagging.nlp.CRFProcess;
import fr.upemlv.lgtools.tagging.nlp.UnitexProcess;
import fr.upemlv.lgtools.text.PropertyExtractor;

/**
 * Environment class containing all loaded data with respect to the text
 * 
 * @author Matthieu Constant
 *
 */

public class Environment {	
	private final LGTaggerConfig config;	
	private final String tempPathPrefix = "/tmp/tmp"+new Date().getTime();
	private final CRFProcess crfProcess = Utils.createWapitiProcess();
	private final UnitexProcess unitexProcess = Utils.createUnitexProcess();
	protected final LGTaggerParameters parameters;
	
    
    
    public Environment(LGTaggerParameters parameters) throws IOException, InterruptedException{
    	LGTaggerConstants.isLGTaggerPathDefined();    	
    	this.parameters = parameters;     	    	
    	this.config = new LGTaggerConfig(getConfigFilename(),parameters.isTaggingWithMWSegmentation());    	    	
    }
   
    
    public Environment(LGTaggerArguments arguments) throws IOException, InterruptedException{
    	this(arguments.getParameters());
    }

    
    
    
    /**
     * Generates in a file the feature templates in CRF++ format
     * 
     * @param filename filename
     * @param withBigrams whether bigrams have to be included
     * @throws IOException 
     */
    
    
    
    public void generateCRFTemplates(String filename,boolean withBigrams) throws IOException{
    	config.generateCRFTemplates(filename, withBigrams);
    }
    
    
    
	/**
	 * @return the parameters
	 */
	public LGTaggerParameters getParameters() {
		return parameters;
	}


	/**
	 * @return the lang
	 * @see fr.upemlv.tagging.config.Config#getLang()
	 */
	public String getLang() {
		return config.getLang();
	}

	
	/**
	 * @return the propertyExtractors
	 */
	public List<PropertyExtractor> getPropertyExtractors() {
		return config.getPropertyExtractors();
	}


	/**
	 * @return the resources
	 */
	public List<String> getResources() {
		List<String> resources = new LinkedList<String>(config.getResources());
		List<String> additional = parameters.getAdditionalResources();
		if(additional != null){
		   resources.addAll(Utils.formatResourcePaths(additional,getLang()));
		}
		return resources;
	}


	/**
	 * @return
	 * @see fr.upemlv.lgtools.tagging.LGTaggerArguments#isTaggingWithMWSegmentation()
	 */
	public boolean isTaggingWithMWSegmentation() {
		return parameters.isTaggingWithMWSegmentation();
	}

	/**
	 * @return the modelPath
	 */
	public String getModelPath() {
		return parameters.getModel();
	}

	/**
	 * @return the config filename
	 */
	 
	public String getConfigFilename() {
		return parameters.getConfig()==null?this.parameters.getModel()+".cfg":parameters.getConfig();
	}

	/**
	 * @return
	 * @see fr.upemlv.lgtools.tagging.LGTaggerArguments#isSegmented()
	 */
	public boolean isSegmented() {
		return parameters.isSegmented();
	}



	/**
	 * @return
	 * @see fr.upemlv.lgtools.tagging.LGTaggerArguments#getNthreads()
	 */
	public int getNthreads() {
		return parameters.getNthreads();
	}

	/**
	 * @return
	 * @see fr.upemlv.lgtools.tagging.LGTaggerArguments#getDevPath()
	 */
	public String getDevPath() {
		return parameters.getDevPath();
	}
	
	
	public InternalTagset getInternalTagset() throws IOException{
		return config.getTagset();
	}


	/**
	 * @return the crfProcess
	 */
	public CRFProcess getCrfProcess() {
		return crfProcess;
	}






	/**
	 * @return the unitexProcess
	 */
	public UnitexProcess getUnitexProcess() {
		return unitexProcess;
	}






	/**
	 * @return
	 * @see fr.upemlv.lgtools.tagging.LGTaggerConfig#getTagset()
	 */
	public InternalTagset getTagset() {
		return config.getTagset();
	}


	/**
	 * @return the tempPathPrefix
	 */
	public String getTempPathPrefix() {
		return tempPathPrefix;
	}
	
/**
 * 
 * @return whether the input (for training) or output (for tagging) is decomposed
 */
	
	public boolean isDecomposed(){
		return parameters.isDecomposedText();
	}
	
    
}
