/** 
 *
 * Software: LGTagger
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */


package fr.upemlv.lgtools.tagging;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Switch;
import com.martiansoftware.jsap.UnflaggedOption;

/**
 * 
 * Includes all arguments used in LGTagger either for training or for tagging
 * 
 * @author Matthieu Constant
 *
 */

public class LGTaggerArguments {
	final public static String TRAIN_LEXICON_SUFFIX = "-train-lexicon.txt";
	final private static String LEX_CONFIG_OPT = "config";
	final private static String TEXT_OPT = "text";
	final private static String MODEL_OPT = "model";
	final private static String MODEL2_OPT = "model2";
	final private static String WITH_SEG_OPT = "with_seg";
	final private static String IS_TOKENIZED_OPT = "is_tokenized";
	final private static String CORPUS_REF_OPT = "corpus_ref";
	final private static String TRAINING_ALGO_OPT = "algo";
	final private static String TAGGING_RESOURCES_OPT = "additional_resources";
	final private static String EVAL_OPT = "eval";
	final private static String NTHREADS_OPT = "nthreads";
	final private static String DEV_OPT = "dev";
	final private static String L1_OPT = "l1";
	final private static String LEXICON_PROJECTION ="lexicon_projection";
	final private static String TREEBANK ="treebank";
	final private static String TAGGING_METHOD_OPT = "method";
	final private static String IS_DECOMPOSED_TEXT_OPT = "is_decomposed";
	final private static String WITH_LEMMA_OPT = "with_lemma";	
	
	private final LGTaggerParameters parameters = new LGTaggerParameters();
	



	public LGTaggerArguments(boolean isTraining,String[] args) throws JSAPException{
		parameters.setTraining(isTraining);
		JSAP jsap = createJSAP();
		JSAPResult config = jsap.parse(args);
		if(!config.success()){
			System.err.print("Usage: ");
			System.err.print("java "+LGTaggerMain.class.getCanonicalName()+" ");
			System.err.println(jsap.getUsage());
			System.err.print(jsap.getHelp());
			System.exit(1);
		}
		
		this.parameters.setTaggingWithMWSegmentation(config.getBoolean(WITH_SEG_OPT));
		this.parameters.setModel(config.getString(MODEL_OPT));
		this.parameters.setDecomposedText(config.getBoolean(IS_DECOMPOSED_TEXT_OPT));
		
		if(!isTraining){ //tagging
		  this.parameters.setEval(config.getBoolean(EVAL_OPT));
		  this.parameters.setModel2(config.getString(MODEL2_OPT));
		  this.parameters.setSegmented(config.getBoolean(IS_TOKENIZED_OPT));
		  this.parameters.setTexts(config.getStringArray(TEXT_OPT));
		  this.parameters.setAdditionalResources(config.getString(TAGGING_RESOURCES_OPT));
		  this.parameters.setMethod(config.getString(TAGGING_METHOD_OPT));
		  this.parameters.setWithLemma(config.getBoolean(WITH_LEMMA_OPT));
		}
		else{ //training
			this.parameters.setConfig(config.getString(LEX_CONFIG_OPT));
			this.parameters.setRef(config.getString(CORPUS_REF_OPT));
			this.parameters.setNthreads(config.getInt(NTHREADS_OPT));
			this.parameters.setDevPath(config.getString(DEV_OPT));
			this.parameters.setL1(config.getDouble(L1_OPT));
			this.parameters.setAlgo(config.getString(TRAINING_ALGO_OPT));
			this.parameters.setTreebank(config.getBoolean(TREEBANK));
			this.parameters.setWithProjection(config.getBoolean(LEXICON_PROJECTION));
		}
		
		check();
	}

	private void check(){
		String method = parameters.getMethod();
		
		if(parameters.isWithProjection() && !parameters.isTreebank()){
			throw new RuntimeException("You should activate --treebank option when projecting lexicon!!");
		}
		
		
		if(!parameters.isTaggingWithMWSegmentation()){
			if(LGTaggerConstants.DOUBLE_MODEL_METHOD.equals(method)
					|| LGTaggerConstants.BASELINE_METHOD.equals(method)){
				throw new RuntimeException("To apply the "+method+" method, you should be in segmentation mode!");
			}
		}
		if(LGTaggerConstants.DOUBLE_MODEL_METHOD.equals(method)){
			if(parameters.getModel2() == null){
				throw new RuntimeException("The double model method requires a second model in the arguments!!");
			}
		}
		
		if(!LGTaggerConstants.DOUBLE_MODEL_METHOD.equals(method)){
			if(parameters.getModel2() != null){
				throw new RuntimeException("You should not define a second model if you don't use the double model method!!");
			}
		}
		
		
	}
	


	private JSAP createJSAP() throws JSAPException{
		JSAP jsap = new JSAP();

		FlaggedOption opt;
		Switch sw;
		
		opt = new FlaggedOption(MODEL_OPT)
		.setStringParser(JSAP.STRING_PARSER)
		.setRequired(true)
		.setShortFlag('m')
		.setLongFlag(MODEL_OPT);
		opt.setHelp("Path of the CRF model (without extension)");
		jsap.registerParameter(opt);

		sw = new Switch(WITH_SEG_OPT)
		.setShortFlag(JSAP.NO_SHORTFLAG)
		.setLongFlag(WITH_SEG_OPT);
		sw.setHelp("Text will be segmented in multiword units.");
		jsap.registerParameter(sw);

		sw = new Switch(IS_DECOMPOSED_TEXT_OPT)
		.setShortFlag(JSAP.NO_SHORTFLAG)
		.setLongFlag(IS_DECOMPOSED_TEXT_OPT);
		sw.setHelp("At learning step, training corpus is already decomposed. At tagging step, the output text will be decomposed.");
		jsap.registerParameter(sw);
		
		if(parameters.isTraining()){
			
			
			
			opt = new FlaggedOption(LEX_CONFIG_OPT)
			.setStringParser(JSAP.STRING_PARSER)
			.setRequired(true)
			.setShortFlag('c')
			.setLongFlag(LEX_CONFIG_OPT)        
			;
			opt.setHelp("Path of configuration file on lexical resources used");
			jsap.registerParameter(opt);

			opt = new FlaggedOption(DEV_OPT)
			.setStringParser(JSAP.STRING_PARSER)		
			.setShortFlag(JSAP.NO_SHORTFLAG)
			.setLongFlag(DEV_OPT)        
			;
			opt.setHelp("Path of development corpus for training");
			jsap.registerParameter(opt);
			
			opt = new FlaggedOption(CORPUS_REF_OPT)
			.setStringParser(JSAP.STRING_PARSER)        
			.setShortFlag(JSAP.NO_SHORTFLAG)
			.setLongFlag(CORPUS_REF_OPT);
			opt.setHelp("Path of the reference annotated corpus (for training and evaluation).");
			opt.setRequired(true);
			jsap.registerParameter(opt);



			opt = new FlaggedOption(NTHREADS_OPT)
			.setStringParser(JSAP.INTEGER_PARSER)		
			.setShortFlag(JSAP.NO_SHORTFLAG)
			.setLongFlag(NTHREADS_OPT)
			.setDefault("1")
			;
			opt.setHelp("Number of threads");
			jsap.registerParameter(opt);

			opt = new FlaggedOption(L1_OPT)
			.setStringParser(JSAP.DOUBLE_PARSER)		
			.setShortFlag(JSAP.NO_SHORTFLAG)
			.setLongFlag(L1_OPT)
			.setDefault("0.5")
			;
			opt.setHelp("L1 penalty");
			jsap.registerParameter(opt);

			opt = new FlaggedOption(TRAINING_ALGO_OPT)
			.setStringParser(JSAP.STRING_PARSER)        
			.setShortFlag(JSAP.NO_SHORTFLAG)
			.setLongFlag(TRAINING_ALGO_OPT);
			opt.setHelp("Algorithm to be used for training with Wapiti: either 'rprop' or 'l-bfgs'");
			opt.setRequired(false);
			opt.setDefault(LGTaggerConstants.RPROP_ALGO);
			jsap.registerParameter(opt);
			
			sw = new Switch(LEXICON_PROJECTION)
			.setShortFlag(JSAP.NO_SHORTFLAG)
			.setLongFlag(LEXICON_PROJECTION);
			sw.setHelp("The lexicons in the configuration are projected on the corpus (option --treebank must be activated).");
			jsap.registerParameter(sw);
			
			sw = new Switch(TREEBANK)
			.setShortFlag(JSAP.NO_SHORTFLAG)
			.setLongFlag(TREEBANK);
			sw.setHelp("The input reference corpus must be a treebank.");
			jsap.registerParameter(sw);
			
		}
		else{
			sw = new Switch(WITH_LEMMA_OPT)
			.setShortFlag(JSAP.NO_SHORTFLAG)
			.setLongFlag(WITH_LEMMA_OPT);
			sw.setHelp("Outputs the lemma of each lexical unit.");
			jsap.registerParameter(sw);
			
			sw = new Switch(IS_TOKENIZED_OPT)
			.setShortFlag(JSAP.NO_SHORTFLAG)
			.setLongFlag(IS_TOKENIZED_OPT);
			sw.setHelp("Text has to be tokenized (one token per line) and segmented in sentences (delimiter: empty line).");
			jsap.registerParameter(sw);

			sw = new Switch(EVAL_OPT)		
			.setShortFlag(JSAP.NO_SHORTFLAG)
			.setLongFlag(EVAL_OPT);		
			sw.setHelp("The evaluation mode is set. Input texts should have been assigned their gold tags (line format: word\ttag).");
			jsap.registerParameter(sw);
			
			opt = new FlaggedOption(TAGGING_RESOURCES_OPT)
			.setStringParser(JSAP.STRING_PARSER)        
			.setShortFlag(JSAP.NO_SHORTFLAG)
			.setLongFlag(TAGGING_RESOURCES_OPT);
			opt.setHelp("Path of file containing lexical resources used for tagging (format: one resource per line).");
			opt.setRequired(false);
			jsap.registerParameter(opt);

			opt = new FlaggedOption(MODEL2_OPT)
			.setStringParser(JSAP.STRING_PARSER)
			.setRequired(false)
			.setLongFlag(MODEL2_OPT);
			opt.setHelp("Path of the second CRF model for 'double' method (without extension). This model has to be devoted to tagging only!");
			jsap.registerParameter(opt);
			
			opt = new FlaggedOption(TAGGING_METHOD_OPT)
			.setStringParser(JSAP.STRING_PARSER)        
			.setShortFlag(JSAP.NO_SHORTFLAG)
			.setLongFlag(TAGGING_METHOD_OPT);
			opt.setHelp("Method used to tag the text: either 'baseline' (shortest path seg + tag), either 'single' (one seg+tag model) or 'double' (seg model + tag model)");
			opt.setRequired(false);
			opt.setDefault(LGTaggerConstants.SINGLE_MODEL_METHOD);
			jsap.registerParameter(opt);
			
			
			UnflaggedOption opt1 = new UnflaggedOption(TEXT_OPT)
			.setStringParser(JSAP.STRING_PARSER)
			.setGreedy(true)
			.setRequired(true);
			opt1.setHelp("Text(s) to be tagged.");
			jsap.registerParameter(opt1);
			
		}
		return jsap;
	}


	/**
	 * 
	 * @return the parameters of process
	 */
    public LGTaggerParameters getParameters(){
    	return parameters;
    }
	
   
}
