/**
 *
 * Software: LGTools
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.tagging;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import fr.upemlv.lgtools.tagging.impl.Utils;
import fr.upemlv.lgtools.tagging.lex.InternalTagset;
import fr.upemlv.lgtools.tagging.lex.Tagset;
import fr.upemlv.lgtools.text.FeatureTemplate;
import fr.upemlv.lgtools.text.PropertyConstants;
import fr.upemlv.lgtools.text.PropertyExtractor;
import fr.upemlv.lgtools.text.PropertyExtractorFactory;
import fr.upemlv.lgtools.text.SetFeature;
import fr.upemlv.lgtools.text.TemplateFeature;
import fr.upemlv.lgtools.utils.IntCounter;

/**
 * Contains general configuration (list of lexical resources and crf templates) 
 * 
 * @author Matthieu Constant
 *
 */

public class LGTaggerConfig {
	private final static String INSIDE_TEMPLATE_SEP="/";
	private final static String PROPERTY_ARGUMENT_SEP=",";
	private final static String LPAR_REGEX="\\(";
	private final static String RPAR_REGEX="\\)";
	private final static SetFeature[] SET_FEATURES = new SetFeature[]{SetFeature.BPOS,SetFeature.BPOSZ};
	private final StringBuilder lang = new StringBuilder();
	private final StringBuilder tagsetTmp = new StringBuilder();
	private final List<String> resourcesTmp = new LinkedList<String>();
	private final Set<String> projectedCats = new HashSet<String>();
	private final InternalTagset tagset;
	private final boolean withSeg;
	private final List<String> resources;
	
	private final List<FeatureTemplate> templates = new LinkedList<FeatureTemplate>();
	private final List<PropertyExtractor> propertyExtractors = new LinkedList<PropertyExtractor>();
	private final Map<String,Integer> propertyIndexes = new HashMap<String, Integer>();
	
	
	
	
	private enum Field{
		LANG{

			@Override
			public void addInformation(String line,LGTaggerConfig config) {
				config.lang.delete(0,config.lang.length());
				config.lang.append(line);
			}
			
		},TAGSET{

			@Override
			public void addInformation(String line,LGTaggerConfig config) {
				config.tagsetTmp.delete(0,config.lang.length());
				config.tagsetTmp.append(line);
			}
			
		},RESOURCES{

			@Override
			public void addInformation(String line,LGTaggerConfig config) {
				config.resourcesTmp.add(line);
			}
			
		},TEMPLATES{

			@Override
			public void addInformation(String line,LGTaggerConfig config) {
				config.templates.add(createTemplate(line));
			}
			
		},PROJECTIONS{

			@Override
			public void addInformation(String line,LGTaggerConfig config) {
				config.projectedCats.add(line);
			}
			
		};
		
		abstract public void addInformation(String line,LGTaggerConfig config);
	};
	private Field field = null; 
	
	
	private static FeatureTemplate createTemplate(String line){
		Pattern pattern = Pattern.compile("([^(]+)\\((.+)\\)");
		String[] tpl = line.split(INSIDE_TEMPLATE_SEP);
		FeatureTemplate template = new FeatureTemplate();
		
		for(String item:tpl){			
		   Matcher m = pattern.matcher(item);
		   if(m.matches()){
			   String propertyType = m.group(1);
			   String[] tmp = m.group(2).split(RPAR_REGEX+LPAR_REGEX);
			   if(tmp.length > 2){
				   throw new RuntimeException("Template "+line+" is not well defined!");
			   }
			   List<String> args = new LinkedList<String>();
			   if(tmp.length == 2){
				   for(String arg:tmp[0].split(PROPERTY_ARGUMENT_SEP)){
					   args.add(arg);
				   }
			   }			   
			   int relativePosition = Integer.parseInt(tmp[tmp.length - 1]);			   
			   template.addFeature(new TemplateFeature(propertyType, args, relativePosition));
			   
		   }
		   else{
			   throw new RuntimeException("Template "+line+" is not well defined!");
		   }
		  
		}
		return template;
	}
	
	
	
	
	public LGTaggerConfig(String filename,boolean withSeg) throws IOException{
		BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(filename),"UTF-8"));
		 String line;
		 while((line = reader.readLine()) != null){
			 line = line.trim();
			 if(line.equals("") || line.startsWith("#")){
				 continue;
			 }
			 if(checkField(line)){
				 continue;
			 }
			 field.addInformation(line,this);
		 }
		 this.resources = Utils.formatResourcePaths(resourcesTmp, getLang());
		 tagset = loadTagset();
		 this.withSeg = withSeg;
		 createPropertyExtractors();
	}
	
	public LGTaggerConfig(String lang, List<String> resources, List<String> templates,boolean withSeg,InternalTagset tagset){
		LGTaggerConstants.isLGTaggerPathDefined();
		this.resources = Utils.formatResourcePaths(resources, lang);
		this.tagset = tagset;
		this.withSeg = withSeg;
		for(String t:templates){
			field.addInformation(t,this);
		}
		createPropertyExtractors();
	}
	

	private boolean checkField(String line){
		if(!line.endsWith(":")){
			return false;
		}
		line = line.substring(0,line.length() - 1);
		if(line.equals(Field.LANG.name())){
			field = Field.LANG;
			return true;
		}
		if(line.equals(Field.RESOURCES.name())){
			field = Field.RESOURCES;
			return true;
		}
		if(line.equals(Field.TEMPLATES.name())){
			field = Field.TEMPLATES;
			return true;
		}
		if(line.equals(Field.TAGSET.name())){
			field = Field.TAGSET;
			return true;
		}
		if(line.equals(Field.PROJECTIONS.name())){
			field = Field.PROJECTIONS;
			return true;
		}
		return false;
	}

	private InternalTagset loadTagset() throws IOException{    	
    	String name = tagsetTmp.toString();
    	return new InternalTagset(new Tagset(Utils.getAbsolutePathForLex(name,getLang())));
    	
    }
	
	
	
	private void createPropertyExtractors(){
    	HashSet<String> propertyNames = new HashSet<String>();
    	//we add by default the token property as first property
    	propertyExtractors.addAll(PropertyExtractorFactory.createInstances(PropertyConstants.TOKEN, null,withSeg,tagset));
    	propertyNames.add(PropertyConstants.TOKEN);
    	propertyIndexes.put(PropertyConstants.TOKEN, propertyIndexes.size());
    	
    	
    	for(FeatureTemplate template:getTemplates()){    		
    		for(TemplateFeature feature:template.getFeatures()){    			
    			List<PropertyExtractor> extractors = PropertyExtractorFactory.createInstances(feature.getPropertyType(),feature.getArguments(),withSeg,tagset);
    			for(PropertyExtractor extractor:extractors){
    				if(!propertyNames.contains(extractor.getPropertyName())){    					
    					propertyExtractors.add(extractor);
    					propertyNames.add(extractor.getPropertyName());
    					propertyIndexes.put(extractor.getPropertyName(), propertyIndexes.size());
    				}    			
    			}
    		}
    	}
    }
	
	

	/**
	 * @return the lang
	 */
	public String getLang() {
		return "".equals(lang.toString())?null:lang.toString();
	}

	

	/**
	 * @return the resources
	 */
	public List<String> getResources() {
		return resources;
	}

	
	
	private SetFeature nextSetFeature(FeatureTemplate template){
		for(TemplateFeature f:template.getFeatures()){
			String name = f.getPropertyName();			
			for(SetFeature sf:SET_FEATURES){
				if(name.equals(sf.getName())){					
					return sf;
				}
			}
			
		}
		return null;
	}
	
	private void generateSimpleTemplate(FeatureTemplate template,IntCounter id,BufferedWriter writer) throws IOException{
		String line = "U"+id.get()+":";
		for(TemplateFeature f:template.getFeatures()){
			//System.err.println(f);
			//System.err.println(f.getPropertyName());
			//System.err.println(propertyIndexes);			
			int propId = propertyIndexes.get(f.getPropertyName());
			line += "/%x["+f.getRelativePosition()+","+propId+"]";
		}
		line = line.replace(":/", ":");
		writer.write(line);
		writer.newLine();
		id.increment();
	}
	
	
	private void generateTemplate(FeatureTemplate template,IntCounter id,BufferedWriter writer) throws IOException{
		SetFeature f = nextSetFeature(template);
		if(f == null){
			generateSimpleTemplate(template, id, writer);
		}
		else{
			for(String cat:f.getElements(tagset, withSeg)){
				FeatureTemplate tmp = template.replace(f.getName(),cat);
				generateTemplate(tmp,id,writer);
			}
			
		}
	}
	
	
	/**
     * Generates in a file the feature templates in CRF++ format
     * 
     * @param filename filename
     * @param withBigrams whether bigrams have to be included
     * @throws IOException 
     */
    
    
    
    public void generateCRFTemplates(String filename,boolean withBigrams) throws IOException{
    	BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(filename),Utils.UTF8));
    	
    	IntCounter id = new IntCounter(1);    	
    	for(FeatureTemplate template:getTemplates()){
    		generateTemplate(template,id,writer);
    		
    	}
    	if(withBigrams){
    		writer.write("B");
    		writer.newLine();
    	}
    	
    	writer.close();
    }
	

	/**
	 * @return the templates
	 */
	public List<FeatureTemplate> getTemplates() {
		return templates;
	}




	/**
	 * @return the tagset
	 */
	public InternalTagset getTagset() {
		return tagset;
	}




	/**
	 * @return the propertyExtractors
	 */
	public List<PropertyExtractor> getPropertyExtractors() {
		return propertyExtractors;
	}




	/**
	 * @return the projectedCats
	 */
	public Set<String> getProjectedCats() {
		return projectedCats;
	}
	
	
	
}
