/**
 * 
 *
 * Software: LGTools
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.tagging;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.Date;
import java.util.HashMap;
import java.util.List;

import fr.upemlv.lgtools.tagging.impl.LGTagger;
import fr.upemlv.lgtools.tagging.impl.LGTaggerLearner;
import fr.upemlv.lgtools.tagging.lex.StandardTagParser;
import fr.upemlv.lgtools.tagging.nlp.ExecInfo;
import fr.upemlv.lgtools.tagging.nlp.MultiAnnotationScore;
import fr.upemlv.lgtools.tagging.validation.LGTaggerValidateConfig;
import fr.upemlv.lgtools.tagging.validation.ValidationModel;
import fr.upemlv.lgtools.tagging.validation.ValidationTest;
import fr.upemlv.lgtools.text.FullTaggedText;
import fr.upemlv.lgtools.text.TaggedTextFactory;

/**
 * @author Matthieu Constant
 *
 */
public class LGTaggerValidator {
	private final String dirPath;
	private final LGTaggerValidateConfig config;
	private final String tmpConfigPath = "/tmp/config"+new Date().getTime()+".txt";
	private final FullTaggedText corpus;
	private final List<FullTaggedText> corpusParts; 
	
	
	/**
	 * @param dirPath
	 * @param configPath
	 * @param corpusPath
	 * @throws IOException 
	 */
	public LGTaggerValidator(String dirPath, String configPath,
			String corpusPath) throws IOException {
		this.dirPath = dirPath;
		
		this.config = new LGTaggerValidateConfig(configPath);
		this.corpus = new FullTaggedText(corpusPath, "utf-8", this.config.getTagset(), new StandardTagParser(),FullTaggedText.ONE_WORD_PER_LINE);
		this.corpusParts = TaggedTextFactory.split(corpus);		
	}
	
	private void createDir(){
		File dir = new File(dirPath);
		if(!dir.exists()){
			if(dir.isFile()){
				throw new RuntimeException(dir + " already exists and is a file! You should indicate a directory!!");
			}
			if(!dir.mkdir()){
				throw new RuntimeException("Could not create directory "+dir);
			}
		}
	}
	
	
	private void trainPart(int i,HashMap<String,ExecInfo> infos) throws Exception{
		System.err.println("++++++++++++++ PART "+i+" ++++++++++++++++++++++");
		int n = i == corpusParts.size() -1 ? 0: i + 1;		
		FullTaggedText train = TaggedTextFactory.constructValidationTrainCorpus(corpusParts, corpus.getTagset(),i,n);
		FullTaggedText dev = corpusParts.get(i);		
		for(ValidationModel model:config.getModels()){
			String name = model.getName();
			System.err.println("---------------- MODEL:"+name+"------------------");			
			if(i == 0){
				infos.put(name, new ExecInfo());
			}
			String modelPathWithoutExt = model.getPath(dirPath, i);
			String modelPath = modelPathWithoutExt+".wap";
			String modelDirPath = new File(modelPath).getParent();
			File modelDir = new File(modelDirPath);
			if(!modelDir.exists()){
				modelDir.mkdir();
			}				

			File modelFile = new File(modelPath);
			if(!modelFile.exists()){
				model.generateConfig(tmpConfigPath);
				String addCorpus = model.getAdditionnalCorpus();
				LGTaggerLearner learner = model.createLearner(tmpConfigPath, modelPathWithoutExt,addCorpus);
				model.setInfo(learner.perform(train, dev));				
				infos.put(name, ExecInfo.add(infos.get(name), model.getInfo()));				
			}
		}

	} 
	
	private void saveInfo(String name, ExecInfo info) throws IOException{
		String filename= dirPath + File.separator + name + File.separator + name+"-infos.txt";
		if(new File(filename).exists()){
			return;
		}
		OutputStreamWriter out = new OutputStreamWriter(new FileOutputStream(filename),"UTF-8");
		out.write("TIME="+info.getTime()+"\n");
		out.write("FEATURES="+info.getFeatureCount()+"\n");
		out.write("LABELS="+info.getLabelCount()+"\n");
		out.close();
		
	}
	
	private void trainAllParts() throws Exception{
		HashMap<String, ExecInfo> infos = new HashMap<String, ExecInfo>();
		for(int i = 0 ; i < corpusParts.size() ; i++){
			trainPart(i,infos);
		}
		for(String name:infos.keySet()){			
			saveInfo(name,infos.get(name));
		}
	}
	
	public void train() throws Exception{
		System.err.println("========= TRAIN ==========");
		createDir();		
		trainAllParts();		
		new File(tmpConfigPath).deleteOnExit();
	}
	
	public HashMap<ValidationTest,MultiAnnotationScore>  test() throws Exception{
		System.err.println("========= TEST ==========");
		HashMap<ValidationTest,MultiAnnotationScore> map = new HashMap<ValidationTest, MultiAnnotationScore>();
		for(ValidationTest mtest:config.getTests()){
		   map.put(mtest, new MultiAnnotationScore());	
		}
		
		for(int i = 0 ; i < corpusParts.size() ; i++){
			int n = i == corpusParts.size() - 1 ?0:i + 1;			
			FullTaggedText test = corpusParts.get(n);
			for(ValidationTest mtest:config.getTests()){				
				LGTagger tagger = mtest.createTestTagger(dirPath,i);			
				MultiAnnotationScore msc = tagger.performAndEvaluate(test, mtest.isSeg(),true,false);
				MultiAnnotationScore msc1 = map.get(mtest);
				msc1.updateScore(msc);
			}
		}
		return map;
	}
	
	
	private void print(HashMap<ValidationTest,MultiAnnotationScore> map){
		System.err.println("============== RESULTS ================");
		for(ValidationTest test:map.keySet()){
			System.err.println(test.getName());
			System.err.println(map.get(test));
		}
	}
	
	
	public void validate() throws Exception{
		train();
		print(test());
	}

}
