/**
 * 
 *
 * Software: LGTools
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.tagging.impl;

import java.io.File;
import java.io.IOException;
import java.util.Date;
import java.util.List;

import fr.upemlv.lgtools.tagging.LGTaggerConstants;
import fr.upemlv.lgtools.tagging.LGTaggerParameters;
import fr.upemlv.lgtools.tagging.TaggingEnvironment;
import fr.upemlv.lgtools.tagging.TaggingEnvironmentAdaptor;
import fr.upemlv.lgtools.tagging.lex.StandardTagParser;
import fr.upemlv.lgtools.text.FullTaggedText;
import fr.upemlv.lgtools.text.LightRawText;
import fr.upemlv.lgtools.text.LightRawTextFromTFST;
import fr.upemlv.lgtools.text.MultiFeaturedText;
import fr.upemlv.lgtools.text.RawText;
import fr.upemlv.lgtools.text.TaggedText;
import fr.upemlv.lgtools.text.TaggedTextFactory;
import fr.upemlv.lgtools.text.TextFst;

/**
 * 
 * @author Matthieu Constant
 *
 */


public class LGTaggerFactory {
   
	

	
   
	
    
    /**
     * Create a lgtagger using a standard CRF model for both MW segmenting and tagging
     * 
     * @param env tagging environment
     * @return the lgtagger
     * @throws IOException
     */
    
	public static LGTagger createStandardSingleModelTagger(final TaggingEnvironment env) throws IOException{
		
		return new LGTagger() {	
			private final LGTaggerPreprocessor processor = new LGTaggerPreprocessor(new PreprocessingContext(env));
			
			private TaggedText tagging(MultiFeaturedText mcorpus) throws IOException, InterruptedException{				
				long id = new Date().getTime();
				String MULTICORPUS = "/tmp/multitext"+id+".txt";
				String OUT = "/tmp/outtaggedtext"+id+".txt";
				mcorpus.save(MULTICORPUS);
				//mcorpus.save("/tmp/couscous2");
				
				System.err.println("Tagging with "+env.getCrfProcess().getToolName()+"...");
				String modelPath = env.getModelPath()+".wap";			    			  
			    env.getCrfProcess().label(modelPath, MULTICORPUS,OUT);
			    
			    //BUILD RESULT			
			    int featCnt = mcorpus.featureCount() + 1;
			    		    		   
			    TaggedText res = new FullTaggedText(OUT, featCnt, "utf-8", env.getTagset(), new StandardTagParser());		    
			    if(env.isTaggingWithMWSegmentation()&&!(env.isDecomposed())){
			      res = TaggedTextFactory.recomposeTaggedText(res,false);
			    }
			    new File(MULTICORPUS).deleteOnExit();
			    new File(OUT).deleteOnExit();
			    return res;
			}
			
			
			@Override
			public TaggedText perform(RawText text) throws Exception {								
				return tagging(processor.preprocess(text));
			}
			
		};
		
	}
	
	/**
	 * Create a standard single model tagger from LGTagger parameters
	 * @param param LGTagger parameters
	 * @return
	 * @throws IOException
	 * @throws InterruptedException
	 */
	
	public static LGTagger createStandardSingleModelTagger(final LGTaggerParameters param) throws IOException, InterruptedException{
		TaggingEnvironment env = new TaggingEnvironment(param);
		return createStandardSingleModelTagger(env);	
	}
	
	public static LGTagger createBaselineTagger(final TaggingEnvironment env, final LGTagger simpleTagger,final List<String> resources){
			return new LGTagger() {
			
				private final LGTaggerPreprocessor preprocessor = new LGTaggerPreprocessor(new PreprocessingContext(env));
				
			@Override
			public TaggedText perform(RawText text) throws Exception {
				System.err.println(resources);
				TextFst tfst = preprocessor.lexicalAnalysis(text,resources);
				System.err.println("Segmenting with Shortest Path...");
				RawText t2 = new LightRawTextFromTFST(tfst,env.getTagset());
				//System.err.println(text);
				//System.err.println(t2);
				return simpleTagger.perform(t2);
				//return null;
			}
		};
	}
	
	public static LGTagger createTwoModelTagger(final LGTagger mweTagger, final LGTagger simpleTagger){
		return new LGTagger() {
		
		@Override
		public TaggedText perform(RawText text) throws Exception {
			System.err.println("MW segmentation with MW-based tagging CRF model...");
			TaggedText tt = mweTagger.perform(text);
			System.err.println("Tagging with CRF model...");
			return simpleTagger.perform(new LightRawText(tt));
			//return null;
		}
	};
}
	
	
	/**
	 * Create the default tagger without MW segmentation
	 * @param lang language
	 * @return
	 * @throws IOException
	 * @throws InterruptedException
	 */
	
	public static LGTagger createDefaultTagger(String lang) throws IOException, InterruptedException{
		return createDefaultTagger(lang, false);
	}
	
	/**
	 * Create the default tagger (with or without MW segmentation)
	 * 
	 * @param lang
	 * @param withSeg
	 * @return
	 * @throws IOException
	 * @throws InterruptedException
	 */
	
	public static LGTagger createDefaultTagger(String lang,boolean withSeg) throws IOException, InterruptedException{		
		String path = LGTaggerConstants.LGTAGGER_PATH;
		String SEP = File.separator;
		String modelPath = path+SEP+lang+SEP+"models"+SEP+"default"+(withSeg?"-mwe":"");		
		LGTaggerParameters parameters = new LGTaggerParameters(modelPath,withSeg,false);
		TaggingEnvironment env = new TaggingEnvironment(parameters);
		
		
		return createStandardSingleModelTagger(env);
	}
	
	
	public static LGTagger createTagger(TaggingEnvironment env,List<String> resources) throws IOException, InterruptedException{
		TaggingEnvironmentAdaptor adaptor = new TaggingEnvironmentAdaptor(env);
		if(env.isTaggingWithMWSegmentation()){
			if(LGTaggerConstants.BASELINE_METHOD.equals(env.getMethod())){				
				LGTaggerParameters parameters = new LGTaggerParameters(env.getModelPath(), false, false);				
				LGTagger simpleTagger =  LGTaggerFactory.createStandardSingleModelTagger(new TaggingEnvironment(parameters));
				return  LGTaggerFactory.createBaselineTagger(env, simpleTagger,resources);
			}
			if(LGTaggerConstants.DOUBLE_MODEL_METHOD.equals(env.getMethod())){
				LGTagger segTagger = LGTaggerFactory.createStandardSingleModelTagger(env);
				LGTagger simpleTagger = LGTaggerFactory.createStandardSingleModelTagger(adaptor.createEnvironementForSecondTagger());
				return  LGTaggerFactory.createTwoModelTagger(segTagger, simpleTagger);
			}
		}
		return LGTaggerFactory.createStandardSingleModelTagger(env);	
	}
	
}
