/**
 * 
 *
 * Software: LGTagger
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */
package fr.upemlv.lgtools.tagging.impl;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import edu.berkeley.nlp.ling.Tree;
import fr.upemlv.lgtools.parsing.ParseTree;
import fr.upemlv.lgtools.parsing.PreprocessedItem;
import fr.upemlv.lgtools.parsing.SentenceItems;
import fr.upemlv.lgtools.parsing.TextItems;
import fr.upemlv.lgtools.parsing.Treebank;
import fr.upemlv.lgtools.parsing.mwe.FilteredGoldMweRecognizer;
import fr.upemlv.lgtools.parsing.mwe.SimpleMweRecognizer;
import fr.upemlv.lgtools.tagging.LGTaggerConfig;
import fr.upemlv.lgtools.tagging.LGTaggerConstants;
import fr.upemlv.lgtools.tagging.TrainEnvironment;
import fr.upemlv.lgtools.tagging.lex.InternalTagset;
import fr.upemlv.lgtools.tagging.lex.StandardTagParser;
import fr.upemlv.lgtools.tagging.nlp.ExecInfo;
import fr.upemlv.lgtools.text.FullTaggedText;
import fr.upemlv.lgtools.text.LightTaggedText;
import fr.upemlv.lgtools.text.MultiFeaturedText;
import fr.upemlv.lgtools.text.Sentence;
import fr.upemlv.lgtools.text.TaggedElement;
import fr.upemlv.lgtools.text.TaggedText;
import fr.upemlv.lgtools.text.TaggedTextFactory;

public class LGTaggerLearner {	
	
	private final LGTaggerPreprocessor preprocessor;
	
	private final TrainEnvironment env;
	

	/**
	 * @param resources
	 * @param algo
	 * @param withSeg
	 * @param templates
	 * @param tagset
	 * @throws IOException 
	 */
	public LGTaggerLearner(TrainEnvironment env,LGTaggerPreprocessor preprocessor) throws IOException {		
		this.env = env;		
		this.preprocessor = preprocessor;
		
	}

	
	private TaggedText incorporateLexiconCorpus(TaggedText initialCorpus,TaggedText additionalTrainCorpus) throws IOException, InterruptedException{
	
		if(additionalTrainCorpus == null){
			return initialCorpus;
		}
		for(Sentence<TaggedElement> s:additionalTrainCorpus){
			initialCorpus.addSentence(s);
		}
		return initialCorpus;
	}
	
	

	private MultiFeaturedText prepareText(TaggedText initialCorpus,TaggedText additionalCorpus) throws Exception{
		TaggedText tcorpus = incorporateLexiconCorpus(initialCorpus,additionalCorpus);
		if(env.isTaggingWithMWSegmentation()&&!env.isDecomposed()){
			tcorpus = TaggedTextFactory.createDecomposedTaggedText(initialCorpus, env.getTagset());
		}
		return preprocessor.preprocess(tcorpus);		
	}
	
	private static String getConcatenatedWord(Sentence<String> sentence,int start,int end){
		String res = "";
		for(int pos = start; pos <= end ; pos++){
			res += LGTaggerConstants.TOKEN_SEPARATOR+sentence.get(pos);
		}
		return res.substring(1);
	}
	
	@SuppressWarnings("unchecked")
	private Treebank modifyTreebank(Treebank treebank,LGTaggerConfig config) throws Exception{		
		Treebank res = new Treebank();
		
		FilteredGoldMweRecognizer recognizer = new FilteredGoldMweRecognizer(env.getTagset(), new SimpleMweRecognizer(config),config.getProjectedCats());
		TextItems textItems = recognizer.apply(treebank);
		Iterator<SentenceItems> it = textItems.iterator();
		for(ParseTree t:treebank){
			SentenceItems sentenceItems = it.next();
			sentenceItems.sortReverse();
			Sentence<String> sentence = new Sentence<String>(t.getTerminalSymbols());
			for(PreprocessedItem item:sentenceItems){
				int start = item.getStartPosition();
				int end = item.getEndPosition();
				Tree<String> parent = t.getFirstCommonAncestor(start,end);
				List<Tree<String>> children = new LinkedList<Tree<String>>(parent.getChildren());
				parent.setChildren(children);
				List<Tree<String>> nodes = t.getAncestorsInFirstCommonAncestorChildren(start,end);
				int firstPos = children.indexOf(nodes.get(0));
				children.removeAll(nodes);
				Tree<String> preterminal = new Tree<String>(item.getPartOfSpeech());
				children.add(firstPos,preterminal);
				Tree<String> terminal = new Tree<String>(getConcatenatedWord(sentence,start,end));
				preterminal.setChildren(new LinkedList<Tree<String>>(Arrays.asList(terminal)));
			}
			res.addTree(new ParseTree(t.getTree()));
			
		}				
		return res;
	}
	
	/**
	 * perform CRF training from reference corpus filenames
	 * 
	 * @param trainCorpusPath
	 * @param devCorpusPath
	 * @throws Exception
	 */
	

	public void perform(String trainCorpusPath, String devCorpusPath,String additionalCorpusPath) throws Exception{
		
		TaggedText trainCorpus;
		TaggedText devCorpus;
		TaggedText addCorpus;
		if(!env.isWithTreebank()){			
			trainCorpus = new FullTaggedText(trainCorpusPath, "utf-8", env.getTagset(),new StandardTagParser(), FullTaggedText.ONE_WORD_PER_LINE);		
			devCorpus = devCorpusPath == null? null:new FullTaggedText(devCorpusPath, "utf-8", env.getTagset(),new StandardTagParser(), FullTaggedText.ONE_WORD_PER_LINE);
			addCorpus = additionalCorpusPath == null? null:new FullTaggedText(additionalCorpusPath, "utf-8", env.getTagset(),new StandardTagParser(), FullTaggedText.ONE_WORD_PER_LINE);
		}
		else{
			Treebank train = new Treebank(trainCorpusPath);
			Treebank dev = devCorpusPath == null? null:new Treebank(devCorpusPath);
			Treebank add = additionalCorpusPath == null? null:new Treebank(additionalCorpusPath);
			if(env.isWithLexiconProjection()){
				//System.err.println(env.getConfigFilename());
				LGTaggerConfig config = new LGTaggerConfig(env.getConfigFilename(), true);
				//System.err.println(config.getProjectedCats());
					train = modifyTreebank(train,config);
					if(dev != null){
						dev = modifyTreebank(dev,config);
					}
					if(add != null){
						add = modifyTreebank(add,config);
					}
			}
			trainCorpus = new LightTaggedText(env.getTagset(), train);
			trainCorpus.save("/tmp/here.txt");
			devCorpus = devCorpusPath == null? null:new LightTaggedText(env.getTagset(), dev);
			addCorpus = additionalCorpusPath == null? null:new LightTaggedText(env.getTagset(), add);
		}		
		perform(trainCorpus, devCorpus,addCorpus);		
	}
	
	
	public void perform(String trainCorpusPath, String devCorpusPath) throws Exception{
		perform(trainCorpusPath, devCorpusPath, null);
	}
	
	private void createMWTrainingLexicon(TaggedText text) throws IOException, InterruptedException{
		Utils.constructLexiconFromTaggedText(text, "+"+InternalTagset.INTERN, env.getUnitexProcess(), env.getModelPath(), true);		
	}
	
	public ExecInfo perform(TaggedText trainCorpus,TaggedText devCorpus) throws Exception{
		return perform(trainCorpus, devCorpus, null);
	}
	
	
	/**
	 * perform CRF training from reference tagged texts
	 * 
	 * @param trainCorpus
	 * @param devCorpus
	 * @throws Exception
	 */
	
	public ExecInfo perform(TaggedText trainCorpus,TaggedText devCorpus,TaggedText additionalCorpus) throws Exception{		
		System.err.println("Preparing training corpus...");
		MultiFeaturedText train = prepareText(trainCorpus,additionalCorpus);
		String tmpTrainPath = env.getTempPathPrefix()+"-train.txt";
		train.save(tmpTrainPath);
		//train.save("/tmp/couscous");
		new File(tmpTrainPath).deleteOnExit();
		System.err.println("Creating training lexicon...");	
		createMWTrainingLexicon(trainCorpus);
		String tmpDevPath = null;		
		if(devCorpus != null){
		   System.err.println("Preparing dev corpus...");
		   MultiFeaturedText dev = prepareText(devCorpus,null);
		   tmpDevPath = env.getTempPathPrefix()+"-dev.txt";
		   dev.save(tmpDevPath);
		   //dev.save("/tmp/ccc");
		   new File(tmpDevPath).deleteOnExit();
		}
		String tmpTemplatePath = env.getTempPathPrefix()+"-templ.txt";
		String tmpModelPath = env.getTempPathPrefix()+"-model.txt";
		System.err.println("Generating CRF feature templates...");
		env.generateCRFTemplates(tmpTemplatePath, true);
		System.err.println("Learning CRF model...");
		ExecInfo info = env.getCrfProcess().learn(tmpTemplatePath, tmpTrainPath,tmpModelPath ,env,tmpDevPath);
		//System.err.println("======>"+info);
		new File(tmpTemplatePath).deleteOnExit();
		new File(tmpModelPath).deleteOnExit();
		System.err.println("Saving CRF model...");
		Utils.copy(tmpModelPath, env.getModelPath()+".wap");
		Utils.copy(env.getConfigFilename(), env.getModelPath()+".cfg");
		System.err.println("Done.");
		return info;
	}


}
