/**
 * 
 *
 * Software: LGTools
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.tagging.lex;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

/**
 * Class represents a tagset transform that is defined in a configuration file
 * 
 * @author Matthieu Constant
 *
 */

public class ConfigurableTransform implements TagsetTransform {	
	private final List<ConfigurableTransformRule> rules = new LinkedList<ConfigurableTransformRule>(); 
	private final String KEEP = "__KEEP__";
	private List<String> keep;
	private final boolean isFlexible;
	
	/**
	 * Constructs transform from a configuration file
	 * 
	 * @param filename file path
	 * @param sourceTagset the source tagset
	 * @param destinationTagset the destination tagset
	 * @throws IOException 
	 */
	public ConfigurableTransform(String filename,Tagset sourceTagset,Tagset destinationTagset,TagParser sourceParser,TagParser destParser,boolean isFlexible) throws IOException{
		 BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(filename),"UTF-8"));
		 keep = new LinkedList<String>();
		 this.isFlexible = isFlexible;
		 String line;
		 while((line = reader.readLine())!=null){
			 line = line.trim();
			 if(line.equals("") || line.startsWith("#")){
				 continue;
			 }
			 if(line.startsWith(KEEP)){
				 String[] tab = line.split("=");
				 if(tab.length != 2){
					 throw new RuntimeException(KEEP+" not in good format in configurable transform file");
				 }
				 keep = Arrays.asList(tab[1].split(","));
			 }
			 else{
			   rules.add(new ConfigurableTransformRule(line,sourceTagset,destinationTagset,sourceParser,destParser));
			 }
		 }
		 reader.close();
	}
	
	
	@Override
	public Collection<Tag> convert(Tag tag) {
		for(ConfigurableTransformRule rule:rules){
			Collection<Tag> list;
			Collection<Tag> res = new LinkedList<Tag>();
			if((list =rule.apply(tag)) != null){
				for(Tag t:list){
					Tag t1 = new Tag(t);
					res.add(t1);
					for(String type:keep){
						t1.addCats(tag.getCategories(type));
					}
				}
				return res;
			}
		}
		String msg ="Tag "+tag+" cannot be converted by transformation!";
		if(!isFlexible){
		   throw new RuntimeException(msg);
		}
		else{
			System.err.println(msg);
			return new LinkedList<Tag>();
		}
	}

}
