/**
 * 
 *
 * Software: LGTagger
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.tagging.lex;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;

import fr.upemlv.lgtools.tagging.impl.Utils;

/**
 * 
 * 
 * @author Matthieu Constant
 *
 */

public class Tag {
	private final Set<String> categories = new HashSet<String>();
	private final Tagset tagset;
	
	public Tag(Tagset tagset){
		this.tagset = tagset;
	}
	
	public Tag(String cat,Tagset tagset){
		this.tagset = tagset;
		addCat(cat);
	}
	
	public Tag(Tag t){
		this.tagset = t.tagset;
		this.categories.addAll(t.categories);
	}
	
	
	
	/**
	 * Add a category in the tag. If category does not belong to the tagset, a runtime exception is thrown.
	 *
	 * @param cat category
	 * @param isFlexible indicates whether it is a flexible addition od not (if not, throw an exception each time a cat does not exist in the tag tagset)
	 */
	
	public void addCat(String cat,boolean isFlexible){		
		if(!tagset.containsCat(cat)){
			if(!isFlexible){
			  throw new RuntimeException("Error: category "+cat +" is not defined in tagset!");
			}
			return;
		}
		categories.add(cat);
	}
	
	
	/**
	 * adds a category in a non-flexible way (i.e. throw an exception if the category is not in tag tagset)
	 * 
	 * @param cat category
	 */
	
	public void addCat(String cat){
		addCat(cat, false);
	}
	
	
	
	/**
	 * Adds a collection of categories in the tag
	 * 
	 * @param cats
	 */
	
	public void addCats(Collection<String> cats){
		for(String c:cats){
			addCat(c);
		}
	}
	
	/**
	 * 
	 * @param t tag
	 * @return whether the current tag (this) contains all categories of the tag t 
	 */
	
	public boolean contains(Tag t){
		return categories.containsAll(t.categories);
	}
	
	
	/**
	 * 
	 * @param type type
	 * @return the concatenated value of cats of type type
	 */
	public String getValue(String type){
		String value="";
		boolean start = true;
		for(String cat:getCategories(type)){
			if(start){
				start = false;
			}
			else{
			   value+=Utils.INSIDE_TAG_SEPARATOR;	
			}
			value+=cat;
		}
		
		return value;
	}
	
	/**
	 *
	 * @return the textual value of the tag
	 */
	
	public String getValue(){
	   LinkedList<String> list = new LinkedList<String>(categories);
	   Collections.sort(list);
	   String res = "";
	   for(String c:list){
		   res += Utils.INSIDE_TAG_SEPARATOR + c;
	   }	   
	   return tagset.getPrintable(res.substring(Utils.INSIDE_TAG_SEPARATOR.length()));
	}
	
	public Tagset getTagset(){
	    return tagset;	
	}
	
    @Override
	public boolean equals(Object o){
		if(o instanceof Tag){
			Tag t = (Tag)o;
			return t.categories.containsAll(categories) && categories.containsAll(t.categories); 
		}
		return false;
	}
    
    /**
     * 
     * @param type a category type
     * @return the categories of the given type
     */
    
    public Collection<String> getCategories(String type){    	
    	Set<String> set = tagset.getCats(type);
    	HashSet<String> res = new HashSet<String>(categories);
    	res.retainAll(set);
    	return res;
    }
    
    /**
     * 
     * @return the categories
     */
    
    public Collection<String> getCategories(){    	    	
    	return categories;
    }
    
    /**
     * 
     * @return whether the tag is conform (it contains a cat of all mandatory types)
     */
    
    public boolean isConform(){
    	for(String type:tagset.getMandatoryTypes()){
    		if("".equals(getValue(type))){
    			throw new RuntimeException("Tag "+this+" should contain a "+type+" category!!");
    		}
    	}
    	return true;
    }
    
    
    /**
     * 
     * @return an exact copy of the current tag
     */
    
    public Tag copy(){
    	Tag tag = new Tag(this.getTagset());
    	tag.addCats(getCategories());
    	return tag;
    }
    
    @Override
    public String toString(){
    	//return getCategories().toString();
    	return getValue();
    }
    
}
