/**
 * 
 *
 * Software: LGTagger
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.tagging.lex;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * Represents a tagset
 * 
 * @author Matthieu Constant
 *
 */
public class Tagset {
	final public static String POS = "POS";
	final protected Map<String,Set<String>> map = new HashMap<String, Set<String>>();
	final protected Map<String,String> cats = new HashMap<String, String>();
	final protected Map<String,Boolean> mandatories = new HashMap<String, Boolean>();
	final protected Set<String> mandatoryTypes = new HashSet<String>();
	final protected Map<String,String> print = new HashMap<String, String>();	

	


	protected Tagset(){
		
	}
	
	/**
	 * Constructs a tagset from a configuration file
	 * 
	 * @param tagsetFilename path of the configuration file
	 * @throws IOException
	 */

	public Tagset(String tagsetFilename) throws IOException{
		BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(tagsetFilename),"UTF-8"));
		String line;
		String current = null;
		while((line = reader.readLine())!= null){
			line = line.trim();
			if(line.equals("")){
				continue;
			}
			if(line.endsWith(":yes") || line.endsWith(":no")){
				String[] tab = line.split(":");
				current = tab[0];
				mandatories.put(current, line.endsWith(":yes"));
				if(line.endsWith(":yes")){
					mandatoryTypes.add(current);
				}
				continue;
			}
			if(current == null){
				continue;
			}
			processComplexTag(line,current);
	
			Set<String> set = map.get(current);
			if(set == null){
				set = new HashSet<String>();
				map.put(current, set);
			}
			set.add(line);
			cats.put(line, current);
		}
		reader.close();
	}
	
	private void processComplexTag(String line,String currentType){
		if(line.contains("+")){
			String []tab = line.split("\\+");
			Set<String> set = map.get(currentType);
			for(String c:tab){
				if(set != null){
					if(set.contains(c)){						
					  continue;
					}
				}
				throw new RuntimeException("Complex tag "+line+" is invalid: members should belong to the same type and should be in the tagset");
			}
			Arrays.sort(tab);
			String t="";
			for(String s:tab){
				t+="+"+s;
			}
			t = t.substring(1);
			print.put(t, line);
		}	
	}

	
	/**
	 * 
	 * @param type category type
	 * @return the set of categories of this type
	 */

	public Set<String> getCats(String type){
		Set<String> res = map.get(type);
		if(res == null){
			throw new RuntimeException("Error: category type "+type +" is not defined in the tagset configuration file !");
		}
		return res;
	}

	/**
	 * 
	 * @param cat category
	 * @return whether the category belongs to the tagset
	 */

	public boolean containsCat(String cat){
		return getType(cat) != null;
	}

	/**
	 * get the type of a given category
	 * 
	 * @param cat
	 * @return type of the category; null if does not exist in tagset
	 */

	public String getType(String cat){
		return cats.get(cat);  
	}

	/***
	 * 
	 * @return set of category types in the tagset
	 */

	public Set<String> getTypes(){
		return map.keySet();
	}

	/**
	 * 
	 * @param type category type
	 * @return whether the category is mandatory or not
	 */

	public boolean isMandatory(String type){
		Boolean res = mandatories.get(type);
		if(res == null){
			throw new RuntimeException("Error: category type "+type +" is not defined in the tagset configuration file !");
		}
		return res;	   
	}

	/**
	 * 
	 * @return mandatory types
	 */

	public Set<String> getMandatoryTypes(){
		return mandatoryTypes;
	}


    /**
     * get the printable value of a tag
     * 
     * @param t a complex tag value
     * @return
     */
	
	public String getPrintable(String t){
		return print.get(t) == null?t:print.get(t);
	}
	


	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "Tagset [map=" + cats + "]";
	}
}
