/**
 * 
 *
 * Software: LGTagger
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.tagging.lex;

import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import fr.upemlv.lgtools.tagging.impl.Utils;

/**
 * Parses Unitex lexical entry
 * 
 * @author Matthieu Constant
 *
 */

public class UnitexLexicalEntryParser implements LexicalEntryParser {
    private final boolean hasFlexibleAddition;
	
	public UnitexLexicalEntryParser(boolean hasFlexibleAddition){
		this.hasFlexibleAddition = hasFlexibleAddition;
	}
	
	@Override
	public Collection<LexicalEntry> parse(String s, Tagset tagset) {
		return loadEntry(s,tagset);
	}

	private List<LexicalEntry> loadEntry(String entry,Tagset tagset){
		entry = entry.trim();
		LexicalEntry t = getInitialLexicalEntry(entry,tagset);
		List<LexicalEntry> tags = appendMorphoFeatures(t,entry,tagset);
		tags = checkMandatoryTypes(tags,tagset,entry); 
		return tags;		
	}


	/*
	 * get a partial lexical entry without morpho features
	 */

	private LexicalEntry getInitialLexicalEntry(String entry,Tagset tagset){
		Pattern pat = Pattern.compile("^([^,]+),([^.]*)\\.");
		//System.err.println("=>"+entry);
		entry = entry.replaceAll("\\\\,", "_");
		//System.err.println("<<"+entry);
		//Pattern pat = Pattern.compile("^((\\,)|[^,])+");
		Matcher m = pat.matcher(entry);
		m.find();
		String w = m.group(1).replaceAll("\\\\", "");
		w = w.replace('_', ',');
		String lemma = m.group(2).replaceAll("\\\\", "");
		lemma = lemma.replace('_', ',');
		if(lemma.equals("")){
			lemma = w;
		}
		Tag t = new Tag(tagset);
		int end = entry.lastIndexOf('.');
		String tag = entry.substring(end + 1);
		
		String[] tags = tag.split(":");

		
		String tab[] = tags[0].split("\\"+Utils.INSIDE_TAG_SEPARATOR);		
		//System.err.println(Arrays.asList(tab));
		//
		for(String cat:tab){
			t.addCat(cat,hasFlexibleAddition);				
		}
		return new LexicalEntry(w,lemma,t);
	}


	/*
	 * Filter entries with mandatory types empty 
	 */

	private List<LexicalEntry> checkMandatoryTypes(List<LexicalEntry> entries,Tagset tagset,String entry){
		List<LexicalEntry> keptEntries = new LinkedList<LexicalEntry>();
		for(LexicalEntry e:entries){
			boolean toKeep = true;
			for(String type:tagset.getMandatoryTypes()){
				if(e.getTag().getCategories(type).isEmpty()){
					toKeep = false;
					System.err.println("Warning: entry "+ entry +" misses feature "+type);
				}
			}	
			if(toKeep){
				keptEntries.add(e);
			}
		}	  	
		return keptEntries;
	}



	private List<LexicalEntry> appendMorphoFeatures(LexicalEntry e,String entry,Tagset tagset){

		int end = entry.lastIndexOf('.');
		String tag = entry.substring(end + 1);
		String[] tags = tag.split(":");


		List<LexicalEntry> list = new LinkedList<LexicalEntry>();
		for(int i = 1 ; i < tags.length ; i++){
			String mph = tags[i];
			HashSet<String> hs = new HashSet<String>();
			for(int j=0 ; j < mph.length() ; j++){
				char c = mph.charAt(j);
				String type = tagset.getType(""+c);
				if(type != null){
					hs.add(""+c);
				}
			}
			if(hs.size() != 0){
				LexicalEntry res = e.copy();
				for(String cat:hs){
					res.addTagFeature(cat);
				}
				list.add(res);
			}		  	
		}
		if(list.size() == 0){
			list.add(e);	
		}	  	
		return list;
	}

}