/**
 * 
 *
 * Software: LGTagger
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.tagging.nlp;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Contains a text tokenizer
 * 
 * @author Matthieu Constant
 *
 */
public class Tokenizer {
	 
	/**
	 * Simple pattern for tokenization (i.e. a sequence of letters) 
	 */
	
	public final static Pattern SIMPLE_PATTERN = Pattern.compile("\\p{L}+");
	
	/**
	 * Complex pattern for tokenization 
	 */
	
	public final static Pattern FRENCH_PATTERN = Pattern.compile("("+
			"(\\bMr*\\.)"+
			"|(\\p{L}+['])"+
			"|(\\d+,\\d+)"+
			"|(\\p{P}+)"+
			"|((\\p{L}|\\d)+)"+			
			")"			
			);
	
	private final Pattern pattern; 
	private final PostTokenizer postTokenizer;
	
	final private static Map<String,Pattern> patterns = new HashMap<String, Pattern>();
	{
		patterns.put("french", Tokenizer.FRENCH_PATTERN);
	}
	
	final private static Map<String,PostTokenizer> postTokenizers = new HashMap<String, PostTokenizer>();
	{
		postTokenizers.put("french", PostTokenizer.WESTERN_LANGUAGE);
	}
	
	
	/**
	 * Constructs a tokenizer with SIMPLE_PATTERN
	 */
	
	public Tokenizer(){
	    this.pattern = SIMPLE_PATTERN;
	    this.postTokenizer = null;
	}
	
	
	/**
	 * Constructs a tokenizer with a regex pattern
	 * 
	 * @param pattern the regex pattern
	 */
	
	public Tokenizer(Pattern pattern){
	    this.pattern = pattern;
	    this.postTokenizer = null;
	}
	
	/**
	 * Constructs a tokenizer with the recorded regex pattern dedicated to a given language
	 * 
	 * @param lang language
	 */
	
	public Tokenizer(String lang){
		lang = lang.toLowerCase();
		Pattern pat = patterns.get(lang);
		PostTokenizer tok = postTokenizers.get(lang);
		
		if(pat == null){
			this.pattern = SIMPLE_PATTERN;
		}
		else{
			this.pattern = pat;
		}
		
		if(tok == null){
			this.postTokenizer = PostTokenizer.DEFAULT;
		}
		else{
			this.postTokenizer = tok;
		}
		
	}
	
	/**
	 * perform the tokenization of a text
	 * 
	 * @param text text
	 * @return the list of the text tokens
	 */
	
	public List<String> perform(String text){
		List<String> list = new ArrayList<String>();
		Matcher m = pattern.matcher(text);
		while(m.find()){			
			list.addAll(postTokenizer.perform(m.group()));
		}
		
		return list;
	}

}
