/**
 * 
 *
 * Software: LGTagger
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.tagging.nlp;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.Collection;
import java.util.Date;

import fr.upemlv.lgtools.tagging.impl.Utils;
import fr.upemlv.lgtools.tagging.lex.Dico;
import fr.upemlv.lgtools.tagging.lex.Tagset;
import fr.upemlv.lgtools.tagging.lex.TreeDico;
import fr.upemlv.lgtools.tagging.lex.UnitexLexicalEntryParser;

/**
 * Contains Unitex functionalities 
 * 
 * @author Matthieu Constant
 *
 */
public class UnitexProcess {
	final private long id = new Date().getTime();
	final public String TMP = "/tmp";
	final private String TMP_SENTENCE = TMP+"/Sentence";
	final private String SENTENCE_GRF = TMP_SENTENCE+".grf";
	final private String SENTENCE_FST2 = TMP_SENTENCE+".fst2";
	final private String TMP_FILE = TMP+"/text"+id;
	final private String TXT = TMP_FILE+".txt";
	final private String SNT = TMP_FILE+".snt";
	final private String SNT_DIR = TMP_FILE+"_snt";
	final private String unitexToolPath;
	final private String unitexPath;
	
	public UnitexProcess(String unitexPath){
		this.unitexPath = unitexPath;
		this.unitexToolPath = unitexPath+File.separator+"App"+File.separator+"UnitexTool";
		//deleteTmpFilesOnExit();
	}		
	
	
	
	public void deleteTmpFiles(){
		new File(TXT).deleteOnExit();
		new File(SNT).deleteOnExit();
		String[] names = new File(SNT_DIR).list();		
		if(names != null){		
		  for(String n:names){			
			new File(SNT_DIR+"/"+n).delete();
		  }		
		}
		new File(SNT_DIR).delete();
		
	}
	
	
	/*
	 * execute an external command
	 * throws Runtime
	 */
	
	private void exec(String[] command) throws IOException, InterruptedException{
		//System.err.println(Arrays.asList(command));
		Process process = Runtime.getRuntime().exec(command);
		BufferedReader stdErr = new BufferedReader(new InputStreamReader(process.getErrorStream()));
		String line;
        while ((line = stdErr.readLine()) != null) {
           System.err.println(line);
        }
        stdErr.close();	
			
		int exit = process.waitFor();
		
		if(exit != 0){
			throw new RuntimeException("Unitex process "+command[1]+" finished with unexpected error!");
		}		
	}
	
	
	
	
	/**
	 * Convert a text in an source encoding to another encoding
	 * @param textFilename text path
	 * @param srcEncoding source encoding
	 * @param destEncoding destination encoding
	 * @throws IOException
	 * @throws InterruptedException 
	 */
	
	public void convert(String textFilename,String srcEncoding,String destEncoding) throws IOException, InterruptedException{
		String[] command = new String[]{
				unitexToolPath, "Convert", "-r"
				, "-s",srcEncoding
				, "-d",destEncoding
				,textFilename
		};
		exec(command);
	}
	
	
	
	/**
	 * Uncompress a unitex dictionary (.bin)
	 * 
	 * @param dicoFilename dico to uncompress
	 * @param outputFilename output filename
	 * @throws IOException
	 * @throws InterruptedException
	 */
	
	public void uncompress(String dicoFilename,String outputFilename) throws IOException, InterruptedException{
		String[] command = new String[]{
				unitexToolPath, "Uncompress"
				, "-o",outputFilename
				,dicoFilename
		};
		exec(command);
	}
	
	/**
	 * Compress a binary unitex dictionary
	 * 
	 * @param dicoFilename textual dictionary to compress
	 * @throws IOException
	 * @throws InterruptedException
	 */
	
	public void compress(String dicoFilename) throws IOException, InterruptedException{
		String[] command = new String[]{
				unitexToolPath, "Compress"
				,dicoFilename
		};
		exec(command);
	}
	
	
	/**
	 * Compile a grf graph in fst2
	 * 
	 * @param grf graph path
	 * @param lang language
	 * @throws IOException 
	 * @throws InterruptedException 
	 */
	
	public void compileGrf(String grf,String lang) throws IOException, InterruptedException{		
		String alphabet = unitexPath+"/"+lang+"/Alphabet.txt";
		String[] command = new String[]{ 
			unitexToolPath, "Grf2Fst2"
			,grf
			,"-y"
			,"--alphabet="+alphabet
		};
		
		exec(command);		
	}
	
	
	/**
	 * Flattens a fst2 graph into an equivalent FST
	 * 
	 * @param fst2 fst2 graph path
	 * @throws IOException
	 * @throws InterruptedException 
	 */
	
	public void flatten(String fst2) throws IOException, InterruptedException{
		String[] command = new String[]{ 
			unitexToolPath, "Flatten"
					,fst2
		};
		exec(command);
	}
	
	
	/**
	 * Flatten a fst2 graph into an equivalent RTN with a maximum depth
	 * 
	 * @param fst2 fst2 graph path
	 * @param depth maximum depth
	 * @throws IOException 
	 * @throws InterruptedException 
	 */
	
	public void flatten(String fst2,int depth) throws IOException, InterruptedException{
		String[] command = new String[]{ 
			unitexToolPath, "Flatten"
			,fst2
			, "--rtn", "-d",""+depth
		};
		exec(command);
	}
	
	/**
	 * 
	 * applies a transducer on text at preprocessing stage
	 * 
	 * @param textFilename text path with .snt extension
	 * @param fst2 fst2 graph path
	 * @param lang language
	 * @throws IOException
	 * @throws InterruptedException 
	 */
	
	public void preprocessingGraphApplication(String textFilename,String fst2,String lang) throws IOException, InterruptedException{
		String alphabet = unitexPath+"/"+lang+"/Alphabet.txt";
		String[] command = new String[]{ 
			unitexToolPath, "Fst2Txt"
			,"-t",textFilename
			,fst2
			,"-a",alphabet
			,"-M"
		};
		exec(command); 
	}
	
	/**
	 * Normalizes a text and creates a snt file
	 * 
	 * @param textFilename text path
	 * @throws InterruptedException 
	 * @throws IOException 
	 */
	
	public void normalize(String textFilename) throws IOException, InterruptedException{
		String[] command = new String[]{
				unitexToolPath, "Normalize"
				,textFilename
		};		
		exec(command);
	}
	
	
	
	
	
	
	/*
	 * Copies all graph in a given directory to another
	 * 
	 */
	
	private static void copyAllGraphs(String src,String dest) throws IOException{
		String[] files = new File(src).list(new FilenameFilter() {
			
			@Override
			public boolean accept(File dir, String name) {				
				return name.endsWith(".grf");
			}
		});
		//System.err.println(src);
		for(String f:files){
			Utils.copy(src+"/"+f, dest+"/"+f);
		}
		
	}	
	
	
	/**
	 * Segment a text in sentences (i.e. insert {S} at the end of each sentence)
	 * 
	 * @param textFilename path of the text to be segmented
	 * @param encoding text encoding
	 * @param lang language of the text
	 * @throws IOException
	 * @throws InterruptedException
	 * @return the stream of the segmented text 
	 */
	
	public Reader sentenceSegmenter(String textFilename,String encoding,String lang) throws IOException, InterruptedException{
		Utils.copy(textFilename, TXT);		
		convert(TXT, encoding, "LITTLE-ENDIAN");
		//System.err.println("Normalize");
		normalize(TXT);
		new File(SNT_DIR).mkdir();
		String sentenceGrfDir = unitexPath+"/"+lang+"/Graphs/Preprocessing/Sentence";
		//System.out.println(sentenceGrfDir);
		copyAllGraphs(sentenceGrfDir, TMP);
		compileGrf(SENTENCE_GRF, lang);
		flatten(SENTENCE_FST2,5);
		preprocessingGraphApplication(SNT, SENTENCE_FST2, lang);
		deleteTmpFiles();
		return new InputStreamReader(new FileInputStream(SNT),"UTF-16LE");
		//convert(SNT, "LITTLE-ENDIAN",encoding);
		
	}
	
	/**
	 * Segment a UTF-8 text in sentences (i.e. insert {S} at the end of each sentence)
	 * 
	 * @param textFilename path of the UTF-8 text to be segmented
	 * @param lang language of the text
	 * @throws IOException
	 * @throws InterruptedException
	 * @return the stream of the segmented text 
	 */
	
	public Reader sentenceSegmenter(String textFilename,String lang) throws IOException, InterruptedException{
		return sentenceSegmenter(textFilename,"UTF-8",lang);
	}
	
	/**
	 * Tokenize text
	 * 
	 * @param textFilename path of the text to be tokenized
	 * @param lang language
	 * @throws InterruptedException 
	 * @throws IOException 
	 */
	
	public void tokenize(String textFilename,String lang) throws IOException, InterruptedException{
		String alphabet = unitexPath+"/"+lang+"/Alphabet.txt";
		String[] command = new String[]{ 
				unitexToolPath, "Tokenize"
				,textFilename
				,"-a",alphabet
			};
		exec(command);
		 
	}
	
	public void lookup(String textFilename, String encoding,String lang, Collection<String> resources, String tfstPath) throws IOException, InterruptedException{
		Utils.copy(textFilename, TXT);
		///Utils.copy(textFilename, "/tmp/here");
		convert(TXT, encoding, "LITTLE-ENDIAN");
		normalize(TXT);
		new File(SNT_DIR).mkdir();
		tokenize(SNT, lang);		
		dicLookup(SNT, lang, resources);
		//Utils.copy(SNT_DIR+"/dlc", "/tmp/here");
		constructTFST(SNT,lang);		
		Utils.copy(SNT_DIR+"/text.tfst", tfstPath);
		
		deleteTmpFiles();
	}
	
	public Reader lookup(String textFilename, String encoding,String lang, Collection<String> resources) throws IOException, InterruptedException{		
		Utils.copy(textFilename, TXT);		
		convert(TXT, encoding, "LITTLE-ENDIAN");
		normalize(TXT);
		new File(SNT_DIR).mkdir();
		tokenize(SNT, lang);		
		dicLookup(SNT, lang, resources);
		constructTFST(SNT,lang);
		InputStreamReader r = new InputStreamReader(new FileInputStream(SNT_DIR+"/text.tfst"),"UTF-16LE");
		deleteTmpFiles();
		return r;
	/*	LinkedList<InputStream> streams = new LinkedList<InputStream>();
		streams.add(new FileInputStream(SNT_DIR+"/dlf"));
		streams.add(new FileInputStream(SNT_DIR+"/dlc"));
		Enumeration<InputStream> e = Collections.enumeration(streams);
		return new InputStreamReader(new SequenceInputStream(e),"UTF-16LE");*/
	}
	
	
	private void constructTFST(String textFilename,String lang) throws IOException, InterruptedException{
		String alphabet = unitexPath+"/"+lang+"/Alphabet.txt";
		String[] command = new String[5];
		command[0] = unitexToolPath;
		command[1] = "Txt2Tfst";
		command[2] = "-a";
		command[3] = alphabet;
		command[4] = textFilename;
		exec(command);
	}
	
	
	
	private void dicLookup(String textFilename, String lang, Collection<String> resources) throws IOException, InterruptedException{
		String alphabet = unitexPath+"/"+lang+"/Alphabet.txt";
		String[] command = new String[6+resources.size()];
		command[0] = unitexToolPath;
		command[1] = "Dico";
		command[2] = "-t";
		command[3] = textFilename;
		command[4] = "-a";
		command[5] = alphabet;
		int i = 6;
		for(String dic:resources){
			//System.err.println("ccc");
			command[i] = dic;			
			i++;
		}
		/*
		for(String s:command){
			System.err.println(s);
		}*/
		exec(command);		
	}
	
	/**
	 * 
	 * @param filename text on which to apply resources 
	 * @param lang language
	 * @param resources list of resources
	 * @param type type of dico to extract (i.e. dlf, dlc or err)
	 * @param encoding encoding
	 * @return the text dico
	 * @throws IOException
	 * @throws InterruptedException
	 */
	
	
	public Dico getTextDico(String filename,String lang,Collection<String> resources,String type,String encoding,Tagset tagset) throws IOException, InterruptedException{
		Utils.copy(filename, TXT);
		///Utils.copy(textFilename, "/tmp/here");
		convert(TXT, encoding, "LITTLE-ENDIAN");
		normalize(TXT);
		new File(SNT_DIR).mkdir();
		tokenize(SNT, lang);		
		dicLookup(SNT, lang, resources);
		TreeDico dic = new TreeDico(SNT_DIR+"/"+type,tagset,new UnitexLexicalEntryParser(true),lang,true);
		//Utils.copy(SNT_DIR+"/dlc", "/tmp/here");
		deleteTmpFiles();
		return dic;
	}
	
	
}
