/**
 * 
 *
 * Software: LGTools
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.tagging.validation;

import java.io.IOException;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import fr.upemlv.lgtools.tagging.impl.Utils;
import fr.upemlv.lgtools.tagging.lex.InternalTagset;
import fr.upemlv.lgtools.tagging.lex.Tagset;
import fr.upemlv.lgtools.utils.FileUtilities;

/**
 * @author Matthieu Constant
 *
 */
public class LGTaggerValidateConfig {
  private static String TMPL = "TMPL";
  private static String RES = "RES";
  private static String LANG = "LANG";
  private static String TAGSET = "TAGSET";
  private static String MODEL = "MODEL";
  private static String TEST = "TEST";
  private static String SEGM = "seg";
  private static String RESM = "res";
  private static String TMPLM = "tmpl";
  private static String ADDCORP = "addcorp";
  private static String ALGOM = "algo";
  private static String METHOD = "method";
  private static String TAGMODEL = "tagmodel";
  private static String SEGMODEL = "segmodel";
    
    
  private String language;
  private InternalTagset tagset;
  private String tagsetFilename;
  private final Map<String,List<String>> templates = new HashMap<String, List<String>>();
  private final Map<String,List<String>> resources = new HashMap<String, List<String>>();
  private final Map<String,ValidationModel> models = new HashMap<String, ValidationModel>();
  private final List<ValidationTest> tests = new LinkedList<ValidationTest>();
  
  
  
  private static boolean isLanguage(String type){
	  return LANG.equals(type);
  }
  
  private static boolean isResource(String type){
	  return RES.equals(type);
  }
  
  private static boolean isTemplates(String type){
	  return TMPL.equals(type);
  }
  
  private static boolean isTagset(String type){
	  return TAGSET.equals(type);
  }
  
  private static boolean isModel(String type){
	  return MODEL.equals(type);
  }
  
  private static boolean isTest(String type){
	  return TEST.equals(type);
  }
  
  private static boolean isAlgo(String type){
	  return ALGOM.equals(type);
  }
  
  private static boolean isSeg(String type){
	  return SEGM.equals(type);
  }
  
  private static boolean isRes(String type){
	  return RESM.equals(type);
  }
  
  private static boolean isAddCorpus(String type){
	  return ADDCORP.equals(type);
  }
  
  private static boolean isTmpl(String type){
	  return TMPLM.equals(type);
  }
  
  private static boolean isMethod(String type){
	  return METHOD.equals(type);
  }
  
  private static boolean isSegModel(String type){
	  return SEGMODEL.equals(type);
  }
  
  private static boolean isTagModel(String type){
	  return TAGMODEL.equals(type);
  }
  
  
  private static void addContent(Map<String,List<String>> map,String name,String[] content){
	  List<String> list = map.get(name);
	  if(list == null){
		  list = new LinkedList<String>();
	  }
	  list.addAll(Arrays.asList(content));
	  map.put(name, list);
  }
  
  
  
  private ValidationModel createValidationModel(String name,String[] content){
	  RuntimeException e = new RuntimeException("Missing information in model"+name+" defined in validation config file!");
	  String a = null;
	  List<String> r = null;	  
	  List<String> t = null;
	  Boolean s = null;
	  String add = null;
	  
	  for(String c : content){
		  String[]info = c.split(":");		  
		  if(info.length != 2){
			  //throw e;
			  continue;
		  }
		  String type = info[0];
		  String val = info[1];
		  if(isAlgo(type)){
			  a = val;			 
		  }
		  if(isRes(type)){
			 r = getResources(val); 
		  }
		  
		  if(isAddCorpus(type)){
			  add = val;
		  }
		  
		  if(isTmpl(type)){
				 t = getTemplates(val); 
		  }
		  if(isSeg(type)){
				 s = Boolean.parseBoolean(val); 
		  }
		  
	  }
	  
	  
	  if(a == null || t == null || s == null){
		  throw e;
	  }
	  if(r == null){
		  r = new LinkedList<String>();
	  }
	  
	  return new ValidationModel(name, a, s, r, t,add, this);
	  
  }
  
  
  private ValidationTest createValidationTest(String name, String[] content){
	  RuntimeException e = new RuntimeException("Missing information in test "+name+" defined in validation config file!");
	  String m = null;
	  ValidationModel sm = null;	  
	  ValidationModel tm = null;
	  Boolean s = null;
	  List<String> res = null;
	  
	  for(String c : content){
		  String[]info = c.split(":");		  
		  if(info.length != 2){
			  throw e;
		  }
		  String type = info[0];
		  String val = info[1];
		  if(isMethod(type)){
			  m = val;			 
		  }
		  if(isTagModel(type)){
			 tm = getModel(val); 
		  }
		  
		  if(isSegModel(type)){
				 sm = getModel(val); 
		  }
		  if(isSeg(type)){
				 s = Boolean.parseBoolean(val); 
		  }
		  
		  if(isRes(type)){
			  //TODO
			  res = Utils.formatResourcePaths(getResources(val),"French");
		  }
		  
	  }
	  
	  
	  if(m == null || s == null){
		  throw e;
	  }
	  

	  return new ValidationTest(name, m, sm, tm, s,res);  
  }
  
  
  private void processLine(String type,String name, String[] content) throws IOException{
	  if(isLanguage(type)){
		  this.language = content[0];
		  return;
	  }
	  if(isTagset(type)){
		  this.tagset = new InternalTagset(new Tagset(Utils.getAbsolutePathForLex(content[0], getLanguage())));
		  this.tagsetFilename = content[0]; 
		  return;
	  }
	  if(isTemplates(type)){
		  addContent(this.templates,name,content);
		  return;
	  }
	  
	  if(isResource(type)){
		  addContent(this.resources,name,content);
		  return;
	  }
	  
	  if(isModel(type)){
		  models.put(name, createValidationModel(name,content));
	  }
	  
	  if(isTest(type)){
		 tests.add(createValidationTest(name,content)); 
	  }
	  
  }
  
  
  public LGTaggerValidateConfig(String filename) throws IOException{	  
	  for(String line:FileUtilities.loadList(filename)){
		  //System.err.println(line);
		  String[] tab = line.split("=");
		  String[] tab2 = tab[0].split(":");		  		  
		  if(tab.length != 2 || tab2.length > 2){			  
			  throw new RuntimeException("Wrong format for validation config file!!");
		  }		  		  
		  
		  String type = tab2[0];		  		  
		  String name =""; 
		  if(tab2.length == 2){  
			  name = tab2[1];
		  }
		  String[] content = tab[1].split(",");
		  if(content[0].isEmpty()){
			  throw new RuntimeException("Wrong format for validation config file!!");
		  }
		  processLine(type,name,content);
	  }
  }
  
  private static List<String> getContent(Map<String,List<String>> map,String name){	  
	  List<String> res = new LinkedList<String>(map.get("") == null?new LinkedList<String>():map.get(""));
	  List<String> list = map.get(name) == null?new LinkedList<String>():map.get(name);
	  if(list.isEmpty()){
		  throw new RuntimeException("Name "+name+" does not exist!!");
	  }
	  res.addAll(list);
	  return res;
  }
  
  
  public List<String> getResources(String name){	  
	  return getContent(resources,name);
  }
  
  public List<String> getTemplates(String name){
	  if(getContent(templates,name).isEmpty()){
		  throw new RuntimeException("Validation config file should define templates (at least, default templates)!");
	  }
	  return getContent(templates,name);
  }
    
  public String getTagsetFilename(){
	  if(tagsetFilename == null){
		  throw new RuntimeException("Validation config file should define a tagset!");
	  }
	  return tagsetFilename;
  } 
  
  public InternalTagset getTagset(){
	  if(tagset == null){
		  throw new RuntimeException("Validation config file should define a tagset!");
	  }
	  return tagset;
  }  
  
  public String getLanguage(){
	  if(language == null){
		  throw new RuntimeException("Validation config file should define a language!");
	  }
	  return language;
  }
  
  
  public Collection<ValidationModel> getModels(){
	  return models.values();
  }
  
  public Collection<ValidationTest> getTests(){
	  return tests;
  }
  
  public ValidationModel getModel(String name){
	  return models.get(name);
  }
  
  
}
