/**
 * 
 *
 * Software: LGTools
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.text;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Arrays;

import fr.upemlv.lgtools.tagging.lex.TagParser;
import fr.upemlv.lgtools.tagging.lex.Tagset;

/**
 * Represents a tagged text each word is associated with a tag
 * 
 * @author Matthieu Constant
 *
 */

public class FullTaggedText extends TaggedText {
		
	/**
	 * Constant indicating that there is one word per line in the tagged text file to be loaded
	 */
	public final static int ONE_WORD_PER_LINE = 0;
	
	/**
	 * Constant indicating that there is one word per line in the tagged text file to be loaded
	 */
	public final static int ONE_SENTENCE_PER_LINE = 1;
	
	
	/**
	 * instantiates an empty tagged text
	 */
	
	public FullTaggedText(Tagset tagset){
		super(tagset);
	}
	
	
	
	/**
	 * Constructs a tagged text from a filename
	 * 
	 * @param filename filename
	 * @param encoding encoding
	 * @param parser tag parser
	 * @param format file format (ONE_WORD_PER_LINE or ONE_SENTENCE_PER_LINE)
	 * @throws IOException
	 */
	
	
	public FullTaggedText(String filename,String encoding,Tagset tagset,TagParser parser,int format) throws IOException{
		super(tagset);
		BufferedReader reader = new BufferedReader(new InputStreamReader(
				new FileInputStream(filename), encoding));		
		if(format == ONE_WORD_PER_LINE){
		   loadOneWordPerLineText(filename,reader,tagset,parser,2);
		}
		if(format == ONE_SENTENCE_PER_LINE){
			loadOneSentencePerLineText(filename,reader,tagset,parser);
		}
		reader.close();
	}
	
	/**
	 * constructs a TaggedText from a multi-column formatted text file
	 * 
	 * @param filename filename
	 * @param columnCount number of columns
	 * @param encoding encoding
	 * @param tagset tagset
	 * @param parser tag parser
	 * @throws IOException
	 */
	
	public FullTaggedText(String filename,int columnCount,String encoding,Tagset tagset,TagParser parser) throws IOException{
		super(tagset);
		BufferedReader reader = new BufferedReader(new InputStreamReader(
				new FileInputStream(filename), encoding)); 
		loadOneWordPerLineText(filename,reader,tagset,parser,columnCount);
		reader.close();
	}
	
	
	private void loadOneWordPerLineText(String filename,BufferedReader reader,Tagset tagset,TagParser parser,int columnCount) throws IOException{
		String line;
		Sentence<TaggedElement> sentence = new Sentence<TaggedElement>();
		boolean inSentence = false;
		while((line = reader.readLine())!=null){
			if(line.equals("")){
				// end of sentence
				addSentence(sentence);
				inSentence = false;
				sentence = new Sentence<TaggedElement>();
				continue;
			}
			String[] tab = line.split(TAGGED_TEXT_SEP);
			if(tab.length != columnCount){
				throw new RuntimeException("Error in Tagged Text ("+filename+"): line "+line+" should have two fields: word and tag!"); 
			}
			sentence.add(new TaggedElement(tab[0], parser.parse(tab[columnCount - 1], tagset)));
			inSentence = true;
		}
		if(inSentence){
			addSentence(sentence);
		}
	}
	
	private void loadOneSentencePerLineText(String filename,BufferedReader reader,Tagset tagset,TagParser parser) throws IOException{
		String line;
		while ((line = reader.readLine()) != null) {
			if (line.equals("")) {
				continue;
			}
			Sentence<TaggedElement> sentence = new Sentence<TaggedElement>();
			String[] tab = line.split(" ");
			for (String item : tab) {
				int count = item.length() - item.replaceAll("/", "").length();
				for (int i = 1; i < count; i++) {
					item = item.replaceFirst("/", "#");
				}
				String[] tab2 = item.split("/");			
				if (tab2.length != 2) {
					throw new RuntimeException(filename
							+ " not in good format (line " + sentenceCount() + "):"
							+ Arrays.asList(tab2) + "\n" + line);
				}
				sentence.add(new TaggedElement(tab2[0].trim(),
						parser.parse(tab2[1],tagset)));				
			}
			addSentence(sentence);
			
		}
	}
	
	
}
