/**
 * 
 *
 * Software: LGTagger
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.text;


import java.util.Arrays;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import fr.upemlv.lgtools.tagging.LGTaggerConstants;
import fr.upemlv.lgtools.tagging.impl.Utils;
import fr.upemlv.lgtools.tagging.lex.InternalTagset;
import fr.upemlv.lgtools.tagging.lex.LexicalEntry;
import fr.upemlv.lgtools.tagging.lex.Tag;
import fr.upemlv.lgtools.tagging.lex.TagFilter;
import fr.upemlv.lgtools.tagging.lex.Tagset;
import fr.upemlv.lgtools.wfst.Path;
import fr.upemlv.lgtools.wfst.Transition;

/**
 * 
 * @author Matthieu Constant
 *
 */

public class PropertyExtractorFactory {

	private final static Pattern pattern = Pattern.compile("\\([^)]+\\)$"); 

	public static String getPropertyName(String prop){
		Matcher m = pattern.matcher(prop);
		if(!m.find()){
			throw new RuntimeException("Error: property reference in template declaration of the configuration file is not in the good format!");
		}
		String tmp = m.group();
		return prop.substring(0, prop.length() - tmp.length());
	}


	private static String getBooleanValue(boolean condition){
		return condition?PropertyConstants.TRUE_VALUE:PropertyConstants.FALSE_VALUE;
	}

	/**
	 * 
	 * @return the token value extractor
	 */

	public static PropertyExtractor createTokenPropertyExtractor(){
		return new PropertyExtractor() {
			@Override
			public String perform(SentenceProcessingContext sentence, int pos) {
				return sentence.getSentenceFst().getSentence().get(pos);
			}

			@Override
			public String getPropertyName() {
				return PropertyConstants.TOKEN;
			}
		};
	}

	public static PropertyExtractor createSimplificationPropertyExtractor(){
		return new PropertyExtractor() {
			@Override
			public String perform(SentenceProcessingContext sentence, int pos) {
				Path<LexicalEntry> entries = sentence.getSentenceFst().getPath();
				LexicalEntry e = entries.getTransitions().get(pos).getInput();
				String word = sentence.getSentenceFst().getSentence().get(pos);
				Tag t = e.getTag();
				if(t == null){
					return word;
				}
				if(t.getValue(InternalTagset.INSIDE).isEmpty()){
					return word;
				}
				String struct = t.getValue("STRUCT");
				String cat = t.getValue(InternalTagset.POS);
				//TODO change harded coded simplification into a small language 
				if((struct != null) && struct.startsWith("N") && cat.equals("NC")){
					return word;
				}
				String seq = word;
				pos++;
				while((pos < sentence.getSentenceFst().getSentence().size())){
					e = entries.getTransitions().get(pos).getInput();
					t = e.getTag();
					if(t == null){
						return seq;
					}
					String iob = t.getValue(InternalTagset.IOB);
					if(iob == null){
						return seq;
					}
					if(iob.equals(InternalTagset.I)){
						word = sentence.getSentenceFst().getSentence().get(pos);
						seq += LGTaggerConstants.TOKEN_SEPARATOR+word;
					}
					else{
						return seq;
					}
					pos++;
				}
				return seq;
			}

			@Override
			public String getPropertyName() {
				return PropertyConstants.SIMPLIFICATION;
			}
		};
	}
	
	
	private static PropertyExtractor createLowercasePropertyExtractor(){
		return new PropertyExtractor() {
			@Override
			public String perform(SentenceProcessingContext sentence, int pos) {
				return sentence.getSentenceFst().getSentence().get(pos).toLowerCase();
			}

			@Override
			public String getPropertyName() {
				return PropertyConstants.LOWERCASE;
			}
		};
	}

	private static PropertyExtractor createHasHyphenPropertyExtractor(){
		return new PropertyExtractor() {
			@Override
			public String perform(SentenceProcessingContext sentence, int pos) {
				return getBooleanValue(sentence.getSentenceFst().getSentence().get(pos).contains("-"));
			}

			@Override
			public String getPropertyName() {
				return PropertyConstants.HAS_HYPHEN;
			}
		};
	}

	private static PropertyExtractor createIsMultiwordPropertyExtractor(){
		return new PropertyExtractor() {		
			@Override
			public String perform(SentenceProcessingContext sentence, int pos) {
				return sentence.getSentenceFst().getSentence().get(pos).contains(LGTaggerConstants.TOKEN_SEPARATOR)?PropertyConstants.TRUE_VALUE:PropertyConstants.FALSE_VALUE;
			}

			@Override
			public String getPropertyName() {
				return PropertyConstants.IS_MULTIWORD;
			}
		};
	}

	private static PropertyExtractor createHasDigitPropertyExtractor(){
		return new PropertyExtractor() {
			Pattern digitPattern = Pattern.compile("[0-9]");
			@Override
			public String perform(SentenceProcessingContext sentence, int pos) {
				Matcher m = digitPattern.matcher(sentence.getSentenceFst().getSentence().get(pos));
				return getBooleanValue(m.find());
			}

			@Override
			public String getPropertyName() {
				return PropertyConstants.HAS_DIGIT;
			}
		};
	}

	private static PropertyExtractor createAllUppercasePropertyExtractor(){
		return new PropertyExtractor() {
			@Override
			public String perform(SentenceProcessingContext sentence, int pos) {
				String token = sentence.getSentenceFst().getSentence().get(pos); 
				if(!Utils.containsOnlyLetters(token)){
					return PropertyConstants.FALSE_VALUE;
				}
				return getBooleanValue(token.toUpperCase().equals(token));
			}

			@Override
			public String getPropertyName() {
				return PropertyConstants.ALL_UPPERCASE;
			}
		};
	}

	private static PropertyExtractor createBOSPropertyExtractor(){
		return new PropertyExtractor() {
			@Override
			public String perform(SentenceProcessingContext sentence, int pos) {

				return getBooleanValue(pos == 0);
			}

			@Override
			public String getPropertyName() {
				return PropertyConstants.BOS;
			}
		};
	}

	private static PropertyExtractor createIsCapitalizedPropertyExtractor(){
		return new PropertyExtractor() {
			@Override
			public String perform(SentenceProcessingContext sentence, int pos) {
				char c = sentence.getSentenceFst().getSentence().get(pos).charAt(0);
				return getBooleanValue(Character.isLetter(c) && Character.isUpperCase(c));
			}

			@Override
			public String getPropertyName() {
				return PropertyConstants.IS_CAPITALIZED;
			}
		};
	}


	
	
	private static PropertyExtractor createPrefixPropertyExtractor(final int length){
		return new PropertyExtractor() {
			@Override
			public String perform(SentenceProcessingContext sentence, int pos) {
				String token = sentence.getSentenceFst().getSentence().get(pos);
				if(token.length() < length){
					return PropertyConstants.DEFAULT_VALUE;
				}
				return token.substring(0,length);
			}

			@Override
			public String getPropertyName() {
				return PropertyConstants.PREFIX+"("+length+")";
			}
		};
	}

	
	
	
	private static PropertyExtractor createSuffixPropertyExtractor(final int length){
		return new PropertyExtractor() {
			@Override
			public String perform(SentenceProcessingContext sentence, int pos) {
				String token = sentence.getSentenceFst().getSentence().get(pos);
				if(token.length() < length){
					return PropertyConstants.DEFAULT_VALUE;
				}
				return token.substring(token.length() - length);
			}

			@Override
			public String getPropertyName() {
				return PropertyConstants.SUFFIX+"("+length+")";
			}
		};
	}
	
	public static PropertyExtractor createLemmaPropertyExtractor(){
		return new PropertyExtractor() {
			@Override
			public String perform(SentenceProcessingContext sentence, int pos) {
				Sentence<TaggedElement> tagged = sentence.getTaggedSentence();
				String cat = tagged == null?null:tagged.get(pos).getTag().getValue("POS");
				String word = sentence.getSentenceFst().getSentence().get(pos);
				for(Transition<LexicalEntry> t:sentence.getSentenceFst().getAllTokenTtransitions(pos)){
					Tag tag = t.getInput().getTag();
					//System.err.println(t.getInput());
					if((cat != null) && (tag != null)&&TagFilter.SIMPLE.accept(tag)){
						
						if(cat.equals(tag.getValue("POS"))){
							String lemma = Utils.normalizeWord(t.getInput().getLemma(), sentence.getLang());
							//little heuristics: prefer lemma different from word
							if(!word.equals(lemma)){
								return lemma;
							}
						}
					}
				}
				return word;
			}

			@Override
			public String getPropertyName() {
				return PropertyConstants.LEMMA;
			}
		};
	}
	

	private static PropertyExtractor createACPropertyExtractor(final boolean withSeg,final TagFilter tagFilter,final String name){
		return new PropertyExtractor() {
			@Override
			public String perform(SentenceProcessingContext sentence, int pos) {
				//State state = sentence.getTokenState(pos);
				//String word = sentence.getSentence().get(pos);
				Set<String> concat = new HashSet<String>();
				//for(Transition<LexicalEntry> t:sentence.getTransitions(state)){
				for(Transition<LexicalEntry> t:sentence.getSentenceFst().getAllTokenTtransitions(pos)){
					Tag tag = t.getInput().getTag(); 
					//System.err.println(tag);					
					if((tag != null) && (tagFilter.accept(tag))){
						//System.err.println(tagFilter.getClass());
						//String w = FullRawText.tokens2string(Arrays.asList(t.getInput().getWord()));
						//word = FullRawText.tokens2string(Arrays.asList(word));
						//if(word.equals(w)){
						String tagValue = tag.getValue(Tagset.POS); 
						if(withSeg){
							tagValue += "+"+tag.getValue(InternalTagset.IOB);
						}
						concat.add(tagValue);
						//}
					}
				}
				if(concat.isEmpty()){
					return PropertyConstants.DEFAULT_VALUE;
				}
				//System.out.println(word+"*********"+concat);
				return Utils.listOfTags2String(new LinkedList<String>(concat));
			}

			@Override
			public String getPropertyName() {
				return name;
			}
		};
	}

	
	public static List<Tag> getBPOSTags(boolean withSeg,Tagset tagset){
		List<Tag> tags = new LinkedList<Tag>();
		String[] suffice = new String[]{""};
		if(withSeg){
			suffice = new String[]{InternalTagset.B,InternalTagset.I};
		}
		for(String cat:tagset.getCats(Tagset.POS)){
			for(String suf:suffice){
				Tag tag = new Tag(tagset);
				tag.addCat(cat);
				if(!suf.isEmpty()){
					tag.addCat(suf);
				}
				tags.add(tag);
			}
		}
		return tags;
	}
	
	public static String getBPOSName(Tag tag){
		return PropertyConstants.BPOS+"("+tag.getValue()+")";
	}
	
	private static List<PropertyExtractor> createBPOSPropertyExtractors(boolean withSeg,final TagFilter filter, Tagset tagset){
		List<PropertyExtractor> list = new LinkedList<PropertyExtractor>();
				
		for(final Tag tag:getBPOSTags(withSeg,tagset)){
				list.add(new PropertyExtractor() {

					@Override
					public String perform(SentenceProcessingContext sentence, int pos) {
						for(Transition<LexicalEntry> t:sentence.getSentenceFst().getAllTokenTtransitions(pos)){
							Tag tfstTag = t.getInput().getTag();
							if(tfstTag != null && filter.accept(tfstTag) && tfstTag.contains(tag)){
								return PropertyConstants.TRUE_VALUE;
							}
						}
						
						return PropertyConstants.FALSE_VALUE;
					}

					@Override
					public String getPropertyName() {					
						return getBPOSName(tag);
					}
				});			
		}		
		return list;
	}
	
	/**
	 * 
	 * @return the MWE_IOB property extractor
	 */
	
	public static PropertyExtractor createMWEIOBPropertyExtractor(){
		return new PropertyExtractor() {

			@Override
			public String perform(SentenceProcessingContext sentence, int pos) {
				return getMWEProperty(sentence.getSentenceFst(), pos, InternalTagset.IOB);
			}

			@Override
			public String getPropertyName() {
				return PropertyConstants.MWE_IOB;
			}
		};
	}	

	

	private static String getMWEProperty(SentenceFst sentence,int pos, String arg){
		String type = arg==null?Tagset.POS:arg;				
		Path<LexicalEntry> path = sentence.getPath();
		
		LexicalEntry e = path.getTransitions().get(pos).getInput();
		Tag t = e.getTag();
		if(t == null){
			return PropertyConstants.DEFAULT_VALUE;
		}
		if(t.getValue(InternalTagset.INSIDE).isEmpty()){
			return PropertyConstants.DEFAULT_VALUE;
		}
		if(t.getValue(InternalTagset.INT).isEmpty()){
		  String iob = t.getValue(InternalTagset.IOB);
		  String propValue = t.getValue(type).isEmpty()?PropertyConstants.DEFAULT_VALUE:t.getValue(type);
		  return "".equals(iob)?PropertyConstants.DEFAULT_VALUE:propValue;
		}
		return PropertyConstants.DEFAULT_VALUE;
	}

	/**
	 * 
	 * @param arg argument of MWE property
	 * @return the MWE property extractor with argument arg
	 */
	
	public static PropertyExtractor createMWEPropertyExtractor(final String arg){
		return new PropertyExtractor() {

			@Override
			public String perform(SentenceProcessingContext sentence, int pos) {
				return getMWEProperty(sentence.getSentenceFst(), pos, arg);
			}

			@Override
			public String getPropertyName() {				
				return PropertyConstants.MWE+"("+arg+")";
			}
		};
	}


	private static PropertyExtractor createTagPropertyExtractor(){
		return new PropertyExtractor() {
			
			@Override
			public String perform(SentenceProcessingContext sentence, int pos) {
				Sentence<TaggedElement> taggedSentence = sentence.getTaggedSentence();
				return taggedSentence.getSequence().get(pos).getTag().getValue(Tagset.POS);
			}
			
			@Override
			public String getPropertyName() {
				return PropertyConstants.TAG;
			}
		};
	}
	
	
	/**
	 * Creates a new instance of a given type of PropertyExtractor   
	 * 
	 * @param prop type of Property extractor
	 * @param arguments arguments associated with the type of the extractor
	 * @param withSeg application is in segmentation mode
	 * @return a new instance of property extractor
	 */

	public static List<PropertyExtractor> createInstances(String prop,List<String> arguments,boolean withSeg,Tagset tagset){
		/* properties with 0 argument */

		if(prop.equals(PropertyConstants.TOKEN)){
			return Arrays.asList(createTokenPropertyExtractor());			
		}
		if(prop.equals(PropertyConstants.LOWERCASE)){
			return Arrays.asList(createLowercasePropertyExtractor());
		}
		if(prop.equals(PropertyConstants.HAS_HYPHEN)){
			return Arrays.asList(createHasHyphenPropertyExtractor());
		}
		if(prop.equals(PropertyConstants.HAS_DIGIT)){
			return Arrays.asList(createHasDigitPropertyExtractor());
		}
		if(prop.equals(PropertyConstants.IS_MULTIWORD)){
			return Arrays.asList(createIsMultiwordPropertyExtractor());
		}
		if(prop.equals(PropertyConstants.ALL_UPPERCASE)){
			return Arrays.asList(createAllUppercasePropertyExtractor());
		}
		if(prop.equals(PropertyConstants.BOS)){
			return Arrays.asList(createBOSPropertyExtractor());
		}
		if(prop.equals(PropertyConstants.IS_CAPITALIZED)){
			return Arrays.asList(createIsCapitalizedPropertyExtractor());
		}
        
		if(prop.equals(PropertyConstants.ACZ)){
			return Arrays.asList(createACPropertyExtractor(withSeg,TagFilter.SIMPLE,PropertyConstants.ACZ));
		}
		
		
		
		
		if(prop.equals(PropertyConstants.AC)){
			return Arrays.asList(createACPropertyExtractor(withSeg,TagFilter.DEFAULT,PropertyConstants.AC));
		}
		
		if(prop.equals(PropertyConstants.ACC)){			
			return Arrays.asList(createACPropertyExtractor(withSeg,TagFilter.MW,PropertyConstants.ACC));
		}
		
		if(prop.equals(PropertyConstants.BPOS)){
			return createBPOSPropertyExtractors(withSeg,TagFilter.DEFAULT,tagset);
		}

		if(prop.equals(PropertyConstants.BPOSZ)){
			return createBPOSPropertyExtractors(withSeg,TagFilter.SIMPLE,tagset);
		}
		
		if(prop.equals(PropertyConstants.MWE_IOB)){			
			return Arrays.asList(createMWEIOBPropertyExtractor());
		}

		if(prop.equals(PropertyConstants.TAG)){			
			return Arrays.asList(createTagPropertyExtractor());
		}
		
		if(prop.equals(PropertyConstants.LEMMA)){			
			return Arrays.asList(createLemmaPropertyExtractor());
		}
		

		/* properties with 1 argument */
		if(arguments.size() < 1){
			throw new RuntimeException("Property "+prop+" is not defined or should have at least one argument !");
		}
		String arg = arguments.get(0);
		
		if(prop.equals(PropertyConstants.MWE)){
			//String arg = arguments.size() == 1?arguments.get(0):null;
			return Arrays.asList(createMWEPropertyExtractor(arg));
		}

		if(prop.equals(PropertyConstants.PREFIX)){
			return Arrays.asList(createPrefixPropertyExtractor(Integer.parseInt(arg)));
		}
		if(prop.equals(PropertyConstants.SUFFIX)){
			return Arrays.asList(createSuffixPropertyExtractor(Integer.parseInt(arg)));
		}



		/*
		if(arguments.size() < 2){
			throw new RuntimeException("Property "+prop+" is not defined or should have at least two arguments !");
		}
		String arg2 = arguments.get(1);
		 */

		/*
		if(prop.equals(PropertyConstants.MWE)){
		    return createMWEPropertyExtractors(arg,env);
		}
		if(prop.equals(PropertyConstants.MWE_IOB)){
		    return createMWEIOBPropertyExtractors(arg,env);
		}
		 */
		throw new RuntimeException("Property "+prop+" is not defined in the system !");
	}	
}
