/**
 * 
 *
 * Software: LGTools
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.text;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import fr.upemlv.lgtools.tagging.LGTaggerConstants;
import fr.upemlv.lgtools.tagging.lex.InternalTagset;
import fr.upemlv.lgtools.tagging.lex.LexicalEntry;
import fr.upemlv.lgtools.tagging.lex.LexicalEntryParser;
import fr.upemlv.lgtools.tagging.lex.Tag;
import fr.upemlv.lgtools.tagging.lex.Tagset;
import fr.upemlv.lgtools.wfst.Path;
import fr.upemlv.lgtools.wfst.State;
import fr.upemlv.lgtools.wfst.Transition;
import fr.upemlv.lgtools.wfst.WFST;

public class SentenceFst extends WFST<LexicalEntry>{
	private static final double MW_WEIGHT = 10.0;
	private static final double SIMPLE_WORD_WEIGHT = 100.0;
	//private final boolean withSeg;
	private final LexicalEntryParser parser = LexicalEntryParser.UNITEX_LEXICAL_ENTRY_PARSER;
	private final Path<LexicalEntry> shortestPath;
	private final Tagset augmentedTagset;
	private final List<String> tokens;
	private final Sentence<String> sentence;
	private int tmpCnt = 0;
	private int lastStateId = 0;
	private final List<State> tokenStates;
	private final List<List<State>> allTokenStates;

	/**
	 * Construct sentence fst when no lexical analysis is performed
	 * 
	 * @param tagset
	 * @param sentence
	 * @param taggingWithSeg
	 * 
	 * 
	 */
	
    public SentenceFst(InternalTagset tagset,Sentence<String> sentence){    	
    	//this.withSeg = taggingWithSeg;
		this.augmentedTagset = tagset;
		this.sentence = sentence;
		this.tokens = sentence.getSequence();
		this.tokenStates = new ArrayList<State>(tokens.size());
		this.allTokenStates = new ArrayList<List<State>>(tokens.size());
		int id = 1;
		
		for(String w:tokens){
			addTransition(id - 1, id, new LexicalEntry(w, w, null), new LexicalEntry(w, w, null), SIMPLE_WORD_WEIGHT);
			this.tokenStates.add(getState(id - 1));
			this.allTokenStates.add(new LinkedList<State>(Arrays.asList(getState(id - 1))));
			id++;
		}
		getState(0).setInitial(true);
		if(tokens.size() != 0){
			getState(id - 1).setFinal(true);
		}
		else{
			getState(0).setFinal(true);
		}
		
		
		this.shortestPath = getShortestPath();
    } 
	
	
	public SentenceFst(BufferedReader r,InternalTagset tagset,Sentence<String> sentence) throws IOException{
		//this.withSeg = taggingWithSeg;
		//System.err.println(sentence.getSequence());
		this.augmentedTagset = tagset;
		this.sentence = sentence;
		this.tokens = sentence.getSequence();
		this.tokenStates = new ArrayList<State>(tokens.size());
		this.allTokenStates = new ArrayList<List<State>>(tokens.size());
		WFST<Integer> ifst = readInitialFST(r);
		Map<Integer,Collection<LexicalEntry>> entries = readEntries(r);		
		constructFST(ifst,entries);
		this.shortestPath = getShortestPath();
	}

	private WFST<Integer> readInitialFST(BufferedReader r) throws IOException{
		String line;
		r.readLine();
		r.readLine();
		r.readLine();
		r.readLine();
		WFST<Integer> ifst = new WFST<Integer>();
		int stateCount = 0;
		while((line = r.readLine()) != null && (!line.equals("f"))){
			State state = ifst.addState(stateCount);
			state.setFinal(line.startsWith("t"));
			state.setInitial(stateCount == 0);
			//System.out.println(line);
			parseState(line,stateCount,ifst);
			stateCount++;
		}
		//System.out.println(ifst.getTransitions());
		return ifst;
	}

	private Map<Integer,Collection<LexicalEntry>> readEntries(BufferedReader r) throws IOException{
		Map<Integer,Collection<LexicalEntry>> entries = new HashMap<Integer, Collection<LexicalEntry>>();
		String line;
		int labelId = 1;
		while((line = r.readLine()) != null && (!line.equals("f"))){

			Collection<LexicalEntry> list = parseLexicalEntry(line);
			
			if(list != null){
				//System.out.println(labelId+"="+list);
				entries.put(labelId, list);
				labelId++;
			}

		}
		return entries;
	}






	private Collection<LexicalEntry> parseLexicalEntry(String line){
		if(!line.startsWith("@")){
			return null;
		}
		tmpCnt++;
		if(tmpCnt % 3 != 0){
			return null;
		}
		Collection<LexicalEntry> list;

		if(!line.startsWith("@{") && !line.endsWith("}")){
			list = new LinkedList<LexicalEntry>();
			line = line.substring(1);
			list.add(new LexicalEntry(line, null, null));   
		}
		else{
			list = parser.parse(line.substring(2,line.length() - 1), augmentedTagset);
		}

		return list;
	}

	private static int getTokenLength(String tok){
		return tok.replaceAll(" ","").replaceAll(LGTaggerConstants.TOKEN_SEPARATOR, "").length();
	}



	

	private Map<Integer,State> getReverseStatePositions(Map<State,Integer> statePositions){
		Map<Integer,State> res = new HashMap<Integer, State>();
		for(Entry<State,Integer> e:statePositions.entrySet()){
			res.put(e.getValue(), e.getKey());
		}
		return res;
	}

	private HashMap<State,Integer> constructSimpleTokenTransitions(List<String> tokens, Map<Integer,State> positionStates,Map<State,Integer> statePositions,WFST<Integer> ifst, Map<Integer,Collection<LexicalEntry>> entries){	  
		HashMap<State,Integer> tokenPositions = new HashMap<State, Integer>();
		int cnt = 0;
		int tokenPos = 0;
		//System.err.println(tokens);
		State last = positionStates.get(0);
		//System.err.println(tokens.size());
		for(String tok:tokens){
			int cur = cnt;
			cnt += getTokenLength(tok);
			State s = positionStates.get(cur);
			s.setInitial(cur == 0);		   
			State d = positionStates.get(cnt);
			last = d;
			tokenPositions.put(s,tokenPos);
			
			
			if(s != null && d != null){
				for(Transition<Integer> tr:ifst.getTransitions(s)){
					//State s1 = tr.getSrc();
					State d1 = tr.getDest();
					if(d1 == d){
						int id = tr.getInput();
						for(LexicalEntry e:entries.get(id)){							
							Tag t = e.getTag();
							if(t!=null){
							  t.addCat(InternalTagset.B);
							}
							addTransition(s.getId(), d.getId(), e, e, SIMPLE_WORD_WEIGHT);
						}
					}
				}
				// unknown word
				if(getTransitions(getState(s.getId())).size() == 0){
					LexicalEntry e = new LexicalEntry(tok, null, null);
					addTransition(s.getId(), d.getId(), e, e, SIMPLE_WORD_WEIGHT);
				}
				tokenStates.add(getState(s.getId()));
				List<State> tmp = new LinkedList<State>();
				allTokenStates.add(tmp);
				tmp.add(getState(s.getId()));
				tokenPos++;
			}
			else{				
				throw new RuntimeException("unexpected tokenization for token \""+tok+"\"!");
			}

		}
		tokenPositions.put(last,tokens.size());
		
		return tokenPositions;

	}

	private static double getTagWeight(Tag t){
		//System.err.println(t+"=>"+(MW_WEIGHT - t.getCategories().size()));
		return MW_WEIGHT - t.getCategories().size(); 
	}

	private void createDecomposedTransitions(int src,int dest,LexicalEntry e,List<String> tokens, int spos,int dpos){
		Tag t =new Tag(e.getTag());
		t.addCat(InternalTagset.B);
		t.addCat(InternalTagset.IN);
		LexicalEntry be = new LexicalEntry(tokens.get(spos), e.getLemma(),t);		
		addTransition(src,lastStateId,be,be,getTagWeight(t));
		lastStateId++;
		t = new Tag(e.getTag());
		t.addCat(InternalTagset.I);
		t.addCat(InternalTagset.IN);
		for(int pos = spos + 1 ; pos < dpos ; pos++){			
			String word = tokens.get(pos);
			LexicalEntry ie = new LexicalEntry(word, e.getLemma(), t);
			int src1 = lastStateId - 1;
			int dest1 = lastStateId;
			if(pos != dpos - 1){
				lastStateId++;
			}
			else{
				dest1 = dest;
			}
			allTokenStates.get(pos).add(getState(src1));
			addTransition(src1,dest1,ie,ie,getTagWeight(t));
		}

	}


	private void constructMWTokenTransitions(WFST<Integer> ifst,Map<Integer,Collection<LexicalEntry>> entries,Set<State> validStates,List<String> tokens,HashMap<State,Integer> stateTokenPositions){
		//System.out.println(tokens);
		HashMap<State,Integer> stateWordMinSizes = new HashMap<State, Integer>();	   
		for(State s:getStates()){
			for(Transition<LexicalEntry> tr:getTransitions(s)){
				//State d = tr.getDest();
				LexicalEntry e = tr.getInput();
				stateWordMinSizes.put(s,getTokenLength(e.getWord()));
				break;
			}
		}
		//System.err.println(stateTokenPositions);
		Collection<State> states = new LinkedList<State>(getStates());
		for(State s1:states){
			State s = ifst.getState(s1.getId());
			//System.out.println(s);

			int minSize = s.isFinal()?0:stateWordMinSizes.get(s1);
			for(Transition<Integer> tr:ifst.getTransitions(s)){
				for(LexicalEntry e:entries.get(tr.getInput())){
					int size = getTokenLength(e.getWord());
					//System.out.println(e.getWord());
					State d = tr.getDest();
					State di = ifst.getState(d.getId());
					if(size > minSize && validStates.contains(d)){
						//System.out.println("===="+e.getWord());
						//System.err.println("===>"+e.getWord());
						//System.err.println("***"+e.getTag());
						int spos = stateTokenPositions.get(s);						
						int dpos = stateTokenPositions.get(di);						
						//System.err.println(s);
						//System.err.println(di);
						createDecomposedTransitions(s.getId(), d.getId(), e,tokens,spos,dpos);
						//addTransition(s.getId(), d.getId(), e, e, 0.0);
					}
				}
			}
		}
	}

	private static Set<Integer> getValidPositions(List<String> tokens){
		Set<Integer> validPositions = new HashSet<Integer>();
		validPositions.add(0);
		int cnt = 0;
		for(String tok:tokens){
			cnt += getTokenLength(tok);
			validPositions.add(cnt);
		}
		return validPositions;
	}

	private Map<State,Integer> getStatePositions(List<State> states,WFST<Integer> ifst,Map<Integer,Collection<LexicalEntry>> entries){
		Map<State,Integer> statePositions = new HashMap<State, Integer>();
		statePositions.put(ifst.getInitialState(),0);

		for(State state:states){
			int curr = statePositions.get(state);
			for(Transition<Integer> t:ifst.getTransitions(state)){
				for(LexicalEntry e:entries.get(t.getInput())){				   
					statePositions.put(t.getDest(), curr + getTokenLength(e.getWord()));
				}
			}
		}
		return statePositions;
	}
	
	private static  Set<State> getValidStates(List<State>states, Set<Integer> validPositions,Map<State,Integer> statePositions){
		Set<State> validStates = new HashSet<State>();
		for(State state: states){
			if(validPositions.contains(statePositions.get(state))){
				validStates.add(state);
			}
		}
		return validStates;
	}


	private void constructFST(WFST<Integer> ifst, Map<Integer,Collection<LexicalEntry>> entries){
		lastStateId = ifst.getStates().size();
		Set<Integer> validPositions = getValidPositions(tokens);
		List<State> states = ifst.topologicalOrderSort();
		Map<State,Integer> statePositions = getStatePositions(states,ifst,entries);
		Map<Integer,State> reverseStatePositions = getReverseStatePositions(statePositions);
		Set<State> validStates = getValidStates(states,validPositions,statePositions);
		HashMap<State,Integer> stateTokenPositions = constructSimpleTokenTransitions(tokens,reverseStatePositions,statePositions,ifst,entries);
		constructMWTokenTransitions(ifst,entries,validStates,tokens,stateTokenPositions);
		for(State s:getStates()){
			State r = ifst.getState(s.getId());
			if(r != null){
				s.setFinal(r.isFinal());
				s.setInitial(r.isInitial());
			}
		}
		/*for(Transition<LexicalEntry> t:getTransitions()){
	    	System.out.println(t);
	    }*/

	}

	private void parseState(String line,int stateId,WFST<Integer> ifst){
		if(line.length() < 2){
			return;
		}
		line = line.substring(2);
		String[] tabs = line.split(" ");
		for(int i = 0 ; i < tabs.length ; i+=2){
			ifst.addTransition(stateId, Integer.parseInt(tabs[i+1]), Integer.parseInt(tabs[i]),Integer.parseInt(tabs[i]), 0.0);
		}
	}

	/**
	 * 
	 * @return the projection ofthe sentence fst for lexical analysis
	 */

	public WFST<String> projectForLexicalAnalysis(){
		WFST<String> fst = new WFST<String>();
		for(Transition<LexicalEntry> tr:getTransitions()){
			Tag t = tr.getInput().getTag();
			if(t != null){			   
				String iob = t.getValue(InternalTagset.IOB);
				String tag = t.getValue(Tagset.POS)+(iob == ""?"":"+"+iob);
				fst.addTransition(tr.getSrc().getId(),tr.getDest().getId(),tr.getInput().getWord(),tag,0.0);
			}
			else{
				for(String cat:augmentedTagset.getCats(Tagset.POS)){
					fst.addTransition(tr.getSrc().getId(),tr.getDest().getId(),tr.getInput().getWord(),cat,0.0);
				}			   
			}		  
		}
		for(State state:getStates()){
			State q = fst.getState(state.getId());
			q.setInitial(state.isInitial());
			q.setFinal(state.isFinal());
		}
		return fst;
	}

	/**
	 * 
	 * @param tokenPos token position
	 * @return the state associated with token at the specified position (without considering MW analysis)
	 */

/*	public State getTokenState(int tokenPos){
		return tokenStates.get(tokenPos);
	}*/

	/**
	 * 
	 * @param tokenPos token position
	 * @return the list of transition starting from the state associated with token at the specified position (without considering MW analysis)
	 */
	
	/*public Collection<Transition<LexicalEntry>> getTokenTtransitions(int tokenPos){		
		return getTransitions(getTokenState(tokenPos));
	}*/
	
	/**
	 * 
	 * @param tokenPos token position
	 * @return all transitions starting from states associated with with token at the specified position (by considering MW analysis)
	 */
	
	
	public Collection<Transition<LexicalEntry>> getAllTokenTtransitions(int tokenPos){		
		Collection<Transition<LexicalEntry>> transitions = new ArrayList<Transition<LexicalEntry>>();
		
		for(State s: getAllTokenStates(tokenPos)){
			transitions.addAll(getTransitions(s));
		}
		
		return transitions;
	}
	
	
	
	/**
	 * 
	 * @param index token position
	 * @return all states associated with with token at the specified position (by considering MW analysis)
	 */
	
	public List<State> getAllTokenStates(int index){
		return allTokenStates.get(index);
	}
	
	
	
	
	/**
	 * @return the sentence
	 */
	public Sentence<String> getSentence() {
		return sentence;
	}

	/**
	 * @return the shortestPath
	 */
	public Path<LexicalEntry> getPath() {
		return shortestPath;
	}


	/**
	 * @return the tokens
	 */
	public List<String> getTokens() {
		return tokens;
	}



}
