package fr.upemlv.lgtools.text;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import fr.upemlv.lgtools.tagging.LGTaggerConstants;
import fr.upemlv.lgtools.tagging.lex.InternalTagset;
import fr.upemlv.lgtools.tagging.lex.StandardTagParser;
import fr.upemlv.lgtools.tagging.lex.Tag;
import fr.upemlv.lgtools.tagging.lex.TagParser;
import fr.upemlv.lgtools.tagging.lex.Tagset;

public class TaggedTextFactory {
	private static final TagParser STANDARD_TAG_PARSER = new StandardTagParser();
	private static final Pattern TOKENIZER = Pattern.compile("\\(([^ (]+) ([^()]+)\\)");
	
	private static Sentence<TaggedElement> createTaggedSentenceFromCONLL(String line,Tagset tagset){
		Sentence<TaggedElement> sentence = new Sentence<TaggedElement>();
		Matcher m = TOKENIZER.matcher(line);
		while(m.find()){
			String tag = m.group(1).split("-")[0];
			sentence.add(new TaggedElement(m.group(2), STANDARD_TAG_PARSER.parse(tag, tagset)));			
		}
		return sentence;
	}
	
	public static FullTaggedText createTaggedTextFromCONLL(String filename, String encoding,Tagset tagset) throws IOException{
		BufferedReader reader = new BufferedReader(new InputStreamReader(
				new FileInputStream(filename), encoding));
		FullTaggedText text = new FullTaggedText(tagset);
		String line;
		
		while((line = reader.readLine())!=null){
			if(line.equals("")){
				continue;
			}
			text.addSentence(createTaggedSentenceFromCONLL(line,tagset));
			
		}
		reader.close();
		return text;
	}
	
	
	public static TaggedText createDecomposedTaggedText(TaggedText tt,InternalTagset tagset) throws IOException{		
		TaggedText res = new FullTaggedText(tagset);
		for(Sentence<TaggedElement> sentence:tt){
			Sentence<TaggedElement> newSentence = new Sentence<TaggedElement>();
			for(TaggedElement e:sentence.getSequence()){
				String w = e.getLexicalValue();
				Tag tag = e.getTag();
				int i = 0;
				for(String tok:w.split(LGTaggerConstants.TOKEN_SEPARATOR)){
				   Tag t = new Tag(tag);
				   if(i == 0){
					 t.addCat(InternalTagset.B);  
				   }
				   else{
					   t.addCat(InternalTagset.I);  
				   }
				   newSentence.add(new TaggedElement(tok, t));
				   i++;
				}				
			}
			res.addSentence(newSentence);
		}

		return res;
	}
	
	/**
	 * Transforms an IOB-based tagged sentence into a composed tagged sentence
	 * 
	 * @param s sentences
	 * @return a new sentence
	 */
	
	public static Sentence<TaggedElement> recomposeTaggedTextSentence(Sentence<TaggedElement> s, Tagset tagset,boolean removeMWSeparator,boolean withEmptyUnit){
		Sentence<TaggedElement> sentence = new Sentence<TaggedElement>();
		boolean isInMW = false;
		String currentWord="";
		
		for(int pos= s.getSequence().size() - 1 ; pos >=0  ; pos--){
			Tag tag = s.get(pos).getTag();			
	 		String word = s.get(pos).getLexicalValue();
	 		if(removeMWSeparator){
				 word = word.replaceAll(LGTaggerConstants.TOKEN_SEPARATOR, "");   
			 }
	 		if(tag == null){
				sentence.addFirst(new TaggedElement(word+currentWord, null));	 			
	 			currentWord="";
	 			continue;
			}
	 		String iob = tag.getValue(InternalTagset.IOB);
	 		//System.err.println(iob);
	 			 		
	 		
	 		Tag newTag = new Tag(tagset);
	 		for(String c:tag.getCategories(Tagset.POS)){
	 			newTag.addCat(c);
	 		}
	 		if(iob.equals(InternalTagset.I)){
	 			isInMW = true;
	 			currentWord=LGTaggerConstants.TOKEN_SEPARATOR+word+currentWord;
	 			if(pos == 0){
	 			   currentWord = currentWord.substring(1);
	 			   
	 			   sentence.addFirst(new TaggedElement(currentWord, newTag));
	 			}
	 			//System.out.println("III="+word+","+tag+"==>"+currentWord);
	 		}
	 		else{
	 			if(iob.isEmpty()){	 			
		 			if(!withEmptyUnit){
		 				throw new RuntimeException("Tags should be in the IOB format!!");		 				
		 			}
		 			else{		 				
		 				String emptySymbol = word;//+InternalTagset.EMPTY;
		 				if(isInMW){		 				  	 				
		 				  currentWord = LGTaggerConstants.TOKEN_SEPARATOR+emptySymbol+currentWord;
		 				  continue;
		 				}
		 				newTag = new Tag(InternalTagset.EMPTY, tag.getTagset());
		 				//word = emptySymbol;
		 			}
		 		}
	 			currentWord = word+currentWord;
	 			isInMW = false;
	 			sentence.addFirst(new TaggedElement(currentWord, newTag));
	 			//System.out.println("BBB="+word+","+tag+"==>"+word+currentWord);
	 			currentWord="";
	 		}
		}
		return sentence;
	}
	
	
	
	
	
	public static TaggedText recomposeTaggedText(TaggedText text,boolean removeMWSeparator){
		return recomposeTaggedText(text,removeMWSeparator,false);
	}
	
	/**
	 * Transforms a IOB-based tagged text into a composed tagged text
	 * 
	 */
	
	public static TaggedText recomposeTaggedText(TaggedText text,boolean removeMWSeparator,boolean withEmptyUnit){
    	TaggedText res = new FullTaggedText(text.getTagset());
    	for(Sentence<TaggedElement> sentence:text){
    		res.addSentence(recomposeTaggedTextSentence(sentence, text.getTagset(),removeMWSeparator,withEmptyUnit));
    	}
    	return res;
    }
	
	
	public static List<FullTaggedText> split(FullTaggedText text){
		int N = 10;
		List<FullTaggedText> res = new ArrayList<FullTaggedText>(N);
		int max = text.sentenceCount();
		int freq = (max / N) + 1;
		int cnt = 0;
		FullTaggedText curr = null;
		for(Sentence<TaggedElement> s:text){
			if(cnt % freq == 0){
				curr = new FullTaggedText(text.getTagset());
				res.add(curr);
			}
			curr.addSentence(s);
			cnt++;
		}
		return res;		
	}
	
	public static FullTaggedText constructValidationTrainCorpus(List<FullTaggedText> texts,Tagset tagset,int... removes){
		List<FullTaggedText> copy = new ArrayList<FullTaggedText>(texts);
		Arrays.sort(removes);
		for(int i = removes.length - 1 ; i >=0 ; i--){
			int rem = removes[i]; 
		    copy.remove(rem);
		}
		FullTaggedText res = new FullTaggedText(tagset); 
		for(FullTaggedText tt:copy){
			for(Sentence<TaggedElement> s:tt){
				res.addSentence(s);
			}
		}				
		return res;
	}
	
	
}
