/**
 * 
 *
 * Software: LGTagger
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.tree;

import java.util.LinkedList;
import java.util.List;

import fr.upemlv.lgtools.tagging.lex.LexicalEntry;
import fr.upemlv.lgtools.tagging.lex.LookupMatch;

/**
 * Class representing a lexical tree
 * 
 * @author Matthieu Constant
 *
 */

public class LexicalTree {
	final private char TOKEN_SEPARATOR = '_';
    private LexicalNode root = new LexicalNode('\0');
    
    /**
     * Add a lexical entry in the tree
     * 
     * @param wordTokens lexical entry (i.e. a sequence of String (tokens))
     * @param tags tags of the lexical entries
     */
    
    public void addLexicalEntry(List<String> wordTokens,LexicalEntry tag){
    	String unit = normalize(wordTokens);
    	LexicalNode node = getRoot();
    	for(char c:unit.toCharArray()){
    	  	node = node.findNode(c, true); 
    	}
    	node.addTag(tag);
    }  
    
    
    /**
     * lookup lexical entries in the lexical tree, starting at a given position in a text
     * 
     * @param position starting position in the text of lo
     * @param tokens text
     * @param maxLength maximum length (in terms of words) of lexical units matching; if -1, all lengths are permitted
     * @return list of matches
     */
    
    public List<LookupMatch> lookup(int position, List<String> tokens, int maxLength){
    	LexicalNode node = getRoot();
    	int cnt = 0;
    	List<LookupMatch> results = new LinkedList<LookupMatch>();
    	for(String tok:tokens.subList(position, tokens.size())){
    		for(char c:(tok+TOKEN_SEPARATOR).toCharArray()){ 
    			node = node.findNode(c, false);
    			if(node == null){
    				return results;
    			}
    			if(c == TOKEN_SEPARATOR){
    				if(node.getTags() != null){
    				  results.add(new LookupMatch(position, position+cnt,node.getTags()));
    				  if(maxLength != -1 && cnt >= maxLength){
    					  return results;
    				  }
    				}
    				cnt++;
    			}    			
    		}
    	}
    	return results;
    	
    }
    
    
    
    /*
     * Normalize lexical unit from a list of tokens into a character sequence
     * tokens are separated with TOKEN_SEPARATOR
     * TOKEN_SEPARATOR ends the output sequence
     */
    
    private String normalize(List<String> wordTokens){
      String s = "";
      for(String tok:wordTokens){
    	  s += tok + TOKEN_SEPARATOR;
      }      
      return s;
    } 
    
    /**
     * 
     * @return tree root
     */
    
    public LexicalNode getRoot(){
    	return root;
    }
}
