/**
 * 
 *
 * Software: LGTools
 *
 * Copyright (C) 2010-2011 Université Paris-Est Marne-la-Vallée
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 */

package fr.upemlv.lgtools.wfst;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;


/**
 * @author Matthieu Constant
 *
 */
public class WFST<T> {
	private final Set<State> states = new HashSet<State>(); 
	private final Set<Transition<T>> transitions = new LinkedHashSet<Transition<T>>();
	private final Map<State,List<Transition<T>>> stateIndex = new HashMap<State, List<Transition<T>>>();
	private final Map<Integer,State> stateIds = new HashMap<Integer, State>();
	
	/**
	 * 
	 * @return the initial state
	 */
	public State getInitialState(){
		for(State state:getStates()){
			if(state.isInitial()){
				return state;
			}
		}			
		return null;
	}
	
	/**
	 * 
	 * @return the final states
	 */
	public Collection<State> getFinalStates(){
		LinkedList<State> finals= new LinkedList<State>();
		for(State state:getStates()){
		   if(state.isFinal()){
			   finals.add(state);
		   }	
		}				
		return finals;
		
	}
	
	
	/**
	 * 
	 * @return all states
	 */
	public Collection<State> getStates(){
		return states;
	}
	
	
	/**
	 * 
	 * @param state
	 */
	
	public static void setFinalState(State state){
		state.setFinal(true);
	} 
	
	
	/**
	 * 
	 * @param state
	 */
	
	public static void setInitialState(State state){
		state.setInitial(true);
	} 
	
	
	/**
	 * 
	 * @return all transitions
	 */
	
	public Collection<Transition<T>> getTransitions(){
		return this.transitions;
	}
	
	/**
	 * 
	 * @param q a state
	 * @return all transitions starting from state q
	 */
	
	public Collection<Transition<T>> getTransitions(State q){
		return stateIndex.get(q)==null?new LinkedList<Transition<T>>():stateIndex.get(q);
	}
	
	
	/**
	 * Adds a transition
	 * 
	 * @param src source state id
	 * @param dest destination state id
	 * @param in input label
	 * @param out output label
	 * @param w weight
	 */
		
	public Transition<T> addTransition(int src,int dest, T in, T out, double w){
		State q = addState(src);
		State r = addState(dest);
		Transition<T> trans = new Transition<T>(q,r,in,out,w);
		boolean isCreated = transitions.add(trans);
		if(isCreated){
			List<Transition<T>> list = stateIndex.get(q);
			if(list == null){
				list = new LinkedList<Transition<T>>();
				stateIndex.put(q,list);
			}
			list.add(trans);
		}
		return trans;
	}
	
	/**
	 * adds a state
	 * 
	 * @param id state id
	 * @return the added state
	 */
	
	
	public State addState(int id){
		State q = stateIds.get(id);
		if(q == null){
			q = new State(id);
			states.add(q);
			stateIds.put(id,q);
			stateIndex.put(q, new LinkedList<Transition<T>>());
		}
		
		return q;
	}
	
	
	/**
	 * 
	 * @param id the state id
	 * @return the state with identifier id
	 */
	
	public State getState(int id){
		return stateIds.get(id);
	}
	
	
		
	
	
	private List<Path<T>> explore(State q){
		List<Path<T>> paths = new LinkedList<Path<T>>();
		for(Transition<T> t:getTransitions(q)){
			List<Path<T>> pths = explore(t.getDest());
			if(pths.isEmpty()){
				Path<T> p = new Path<T>();
				p.addFirst(t);
				paths.add(p);				
			}
			for(Path<T> p:pths){
				p.addFirst(t);
				paths.add(p);
			}				
		}
		return paths;
	}
	
	/**
	 * 
	 * @return all paths of WFST
	 */
	
	public List<Path<T>> explorePaths(){
		State q = getInitialState();
		List<Path<T>> list = explore(q);	
		return list;
	}
	
	private static class PairState {
		final private State q1;
		final private State q2;


		public PairState(State q1, State q2) {			
			this.q1 = q1;
			this.q2 = q2;
		}
		@Override
		public int hashCode() {
			return q1.hashCode() + q2.hashCode();
		}
		@Override
		public boolean equals(Object obj) {
		   if(obj instanceof PairState){
			  PairState state = (PairState)obj;
			  return (state.q1 == this.q1) && (state.q2 == this.q2); 
		   }
		   return false;
		}

	}
	
	
	private int getStateId(Map<PairState,Integer> states,PairState q){
		Integer state = states.get(q);
		if(state == null){
			states.put(q, states.size());
		}
		return state;
	}
	
	private HashMap<T,List<Transition<T>>> buildStateTransitionIndex(State q){
		HashMap<T,List<Transition<T>>> map = new HashMap<T,List<Transition<T>>>();
		for(Transition<T> t1:getTransitions(q)){
			T output = t1.getOutput();
			List<Transition<T>> list = map.get(output);
			if(list == null){
				list = new LinkedList<Transition<T>>();
				map.put(output, list);
			}
			list.add(t1);
		}
		return map;
	}
	
	private Map<State,Map<T,List<Transition<T>>>> buildTransitionIndex(){
		Map<State,Map<T,List<Transition<T>>>> index = new HashMap<State,Map<T,List<Transition<T>>>>();
		for(State q: getStates()){
			index.put(q, buildStateTransitionIndex(q));
		}
		return index;
	}
	
	
	/**
	 * Compose this WFST with another one
	 * 
	 * @param wfst wfst to be composed with this
	 * @return the wfst resulting of composition
	 */
	
	public WFST<T> compose(WFST<T> wfst){
		WFST<T> res = new WFST<T>();
		Map<PairState,Integer> states = new HashMap<PairState, Integer>();
		ArrayDeque<PairState> toDo = new ArrayDeque<PairState>();
		HashSet<PairState> visited = new HashSet<PairState>();
		State q1 = this.getInitialState();
		State q2 = wfst.getInitialState();
		PairState q = new PairState(q1, q2);
		int stateId = getStateId(states, q); 
		State state = res.addState(stateId);
		if(q1.isFinal() && q2.isFinal()){
		    setFinalState(state);
		  }
		setInitialState(state);
		Map<State,Map<T,List<Transition<T>>>> index = buildTransitionIndex();
		
		toDo.add(q);		
		while(!toDo.isEmpty()){
			q = toDo.poll();
			if(!visited.contains(q)){				
				visited.add(q);
				q1 = q.q1;
				q2 = q.q2;
				
				
				for(Transition<T> t2:getTransitions(q2)){
					T input = t2.getInput();
					List<Transition<T>> transitions1 = index.get(q1).get(input);
					if(transitions1 != null){
						for(Transition<T> t1:transitions1){
							State r1 = t1.getDest();
							State r2 = t2.getDest();
							PairState r = new PairState(r1, r2);
							toDo.add(r);
							res.addTransition(getStateId(states,q), getStateId(states,r), t1.getInput(), t2.getOutput(), t1.getWeight()+t2.getWeight());						  
							if(r1.isFinal() && r2.isFinal()){
								setFinalState(addState(getStateId(states,r)));
							}
						}
					}
				}
				
			}
		}
		
		return res;
	}
	
	/**
	 * 
	 * @return all states sorted by topological order
	 */
	
	
	public List<State> topologicalOrderSort(){
		List<State> orderedStates = new ArrayList<State>(getStates().size());
		Map<State,Integer> inCounts = new HashMap<State, Integer>();
		ArrayDeque<State> queue = new ArrayDeque<State>();		
		orderedStates.add(getInitialState());
		for(State s:getStates()){
			inCounts.put(s, 0);
		}
		
		for(State s:getStates()){
			//System.out.println(getTransitions(s));
			for(Transition<T> t:getTransitions(s)){
				State q = t.getDest();
				inCounts.put(q, inCounts.get(q) + 1);
			}
		}
		//System.out.println("initial:"+getInitialState());
		queue.add(getInitialState());
		while(!queue.isEmpty()){
			State s = queue.poll();
			for(Transition<T> t:getTransitions(s)){
				State q = t.getDest();
				inCounts.put(q, inCounts.get(q) - 1);
				if(inCounts.get(q) == 0){
					queue.add(q);
					orderedStates.add(q);
				}
			}
		}		
		
		return orderedStates;
	}
	
	
	private static double getScore(Map<State,Double> scores,State q){
		Double sc = scores.get(q);
		if(sc == null){
			sc = Double.MAX_VALUE;
		}
		return sc;
	}	
	
	
	/**
	 * 
	 * @return the shortest path
	 */
	
	public Path<T> getShortestPath(){
		Map<State,Double> scores = new HashMap<State, Double>();
		Map<State,Transition<T>> prev = new HashMap<State, Transition<T>>();		
		scores.put(getInitialState(), 0.0);
		//System.err.println(getStates().size());
		
		for(State s:topologicalOrderSort()){
			double old = getScore(scores,s); 
			for(Transition<T> t:getTransitions(s)){
				double score = old + t.getWeight();
				State r = t.getDest();
				double best = getScore(scores,r);			
				if(score < best){					
					scores.put(r, score);
					prev.put(r, t);
				}
			}
		}				
		
		Path<T> path = new Path<T>();
		double bestScore = Double.MAX_VALUE;
		State bestState = null;
		for(State s:getFinalStates()){
			double score = getScore(scores, s);
			if(score < bestScore){
				bestState = s;
				bestScore = score;
			}
		}
		path.setScore(bestScore);
		//System.out.println("Best score="+bestScore);
		Transition<T> t = prev.get(bestState);
		while(t != null){
		    State s = t.getSrc();
		    path.addFirst(t);
			t = prev.get(s);
		}		
		return path;
	}
	
	
}
